/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "cfgfile.h"
/* 
Parse the file for a line starting with:
varName =
Returns 0 if found, 1 if not in file, 2 if file not found. 
Destination is filled with everything after any white space after 
the equal sign, minus any # comments, and trailing white space, up 
to maxVarLen chars.  It is a null string if varName is not found.
If localOnly == 1, search only current directory for the file 
if parse_file_4_variables finds a line containing exactly:
parse_file_verbose_mode
it will go into verbose mode.
*/
char * verbose_mode = "parse_file_verbose_mode";
int
parse_file_4_variables(char *fileName,char *varName,int maxVarLen,
        char *destination,int localOnly)
{
	char file[256];
    char lineBuf[512];
	FILE *fd;
	int f;
	char *p;
	char *p2;
	char *homename;
	int verbose;
	int last;

	*destination = 0; /* start as a null string */
	verbose = 0;

	if (localOnly) 
		last = 1;
	else
		last = 2;

	for(f=0;f < last;f++)
	{
		switch(f)
		{
			case 0:
				sprintf(file,"%s",fileName);
				break;
			case 1:
				if((homename = getenv("HOME")) != NULL)
				{
				sprintf(file,"%s/.%s",fileName);
				}
				else
				{
					sprintf(file,"%s",fileName);
				}
				break;
		}
		if((fd = fopen(file,"r")) == NULL)
			continue;

		/* read lines */
		while(fgets(lineBuf,sizeof(lineBuf),fd) != 0)
		{
			lineBuf[strlen(lineBuf) -1] = 0; /* get rid of slash n */
			p = strstr(&lineBuf[0],verbose_mode);
			if(p == &lineBuf[0])
			{
				verbose = 1;
				printf("Parcing file %s for variable %s\n",file,varName);
			}
			/* now look for varName starting at start of line */
			p = strstr(&lineBuf[0],varName);
			/* if found at start of line */
			if(p == &lineBuf[0])
			{
				p += strlen(varName); /* first char after varName */
				while(isspace(*p)) p++; /* skip spaces, tabs, etc */
				if(*p != '=')
					continue; /* try next line of file */
				p++; /* skip over equal */

				/* OK, got a match, now clean it up */
				while(isspace(*p)) p++; /* skip spaces, tabs, etc */

				/* get rid of comments */
				p2 = strchr(p,'#');
				if(p2 != NULL)
				  *p2 = 0; /* get rid of comments */

				/* get rid of trailing spaces */
				p2 = p + strlen(p) -1;
				while(isspace(*p2)) *p2-- = 0;

				/* ok, give it back to the caller */
				strncpy(destination,p,maxVarLen);
				if(verbose)
					printf("Found in %s! %s = %s\n",file,varName,p);
				fclose(fd);
				return (0); /* success */
			}
		}
		fclose(fd);
				if(verbose)
					printf("Variable %s Not found\n",varName);
		return 1; /* didn't find the variable name */
	}
				if(verbose)
					printf("File %s Not found\n",fileName);
	return 2; /* didn't find the file */
}

