/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
/*****************************************************************************/
/* Wild card expander for DOS.  Hacked from an ls.c program.                 */
/* Any code that resembles the original authors code, is probably the part   */
/* that still works. :-)                                                     */
/* wild card entry point.  The main program would call this code like this */
/*
        main(argc, argv)
        int  argc;
        char *argv[];
        {
                dosWildCard(&argc, &argv);
        }

*/
/* Here's the orginal original author's description box.                     */
/* 1994 John Smith (johns@FileNet.com)                                       */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* ls - a Unix-like directory listing program for MS-DOS 2.x +               */
/*                                                                           */
/*               1984  R. Edward Nather                                      */
/*               1985  Larry A. Shurr                                        */
/*               1989  R. Ray Hensel  Comverted to Microsoft V.5             */
/*                                    cleaned up code                        */
/*                                    added many coments                     */
/*                                    replaced heap sort with quick sort     */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>

#ifndef unix
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <direct.h>
#include <dos.h>
#include "doswld.h"

static char *REVISION = "2.0.2";

#if 0
#define USE_QSORT
#endif

static unsigned long string_sizes;              /* size of strings */
static unsigned long added_string_sizes;              /* size of strings */
static unsigned int nbr_of_strings;
static unsigned int pass2_string_nbr;
int debug_doswld = 0;
#if 0
#define DEBUG_DOSWLD
#endif

typedef int bool;                       /*Define "boolean" type*/
typedef int mchar;                      /*Define "metachar" type*/

struct sysreg {                    /*Generic register file*/
    int ax;
    int bx;
    int cx;
    int dx;
    int si;
    int di;
    int ds;
    int es; 
};
typedef struct sysreg SYSREG;

#ifdef __POWERC
#define FILEDATA ffblk
#else
#define FILEDATA find_t
#endif


#define FALSE 0                         /*Define "false" value*/
#define TRUE  1                         /*Define "true"  value*/
#define is ==                           /*Define some readable stuff*/
#define isnot !=
#define and &&
#define or ||
/*****************************************************************************/
/* customizing constants */

#define NAMESIZ 13                      /*12 character name + NULL*/
/*****************************************************************************/
/*****************************************************************************/
#define MAXSPATH 80
static char spath[MAXSPATH + 1];              /*holds current pathname string*/
static char curDir[MAXSPATH + 1];              /*holds current pathname string*/

static mchar qs       = '\\';                  /*filename separator character*/
static char  dqs[]    = "\\";                  /*filename separator (string)*/

static bool
df_all   = FALSE,                 /*1 => show hidden files by default*/
df_colm  = FALSE,                 /*1 => 1-column listing by default*/
df_du    = FALSE,                 /*1 => include disk use by default*/
df_id    = FALSE,                 /*1 => always identify directory*/
df_long  = FALSE,                 /*1 => long listing by default*/
df_more  = FALSE,                 /*1 => pause for more*/
df_rsort = FALSE,                 /*1 => reverse sort by default*/
df_tsort = FALSE,                 /*1 => time sort by default*/
df_exitNF = FALSE,                /*1 => exit if file not found */
f_help   = FALSE                  /*Help screen requested*/
;
/* global variables and flags */

int   f_all,                            /*Include hidden & system files*/
f_colm,                           /*1-column format*/
f_du,                             /*Print disk usage*/
f_long,                           /*Long listing*/
f_rsort,                          /*Reverse sort*/
f_tsort,                          /*Timesort the listing*/
f_more  = FALSE,                  /*Pause for more*/
f_recd  = FALSE,                  /*Recursive descent requested*/
f_so    = FALSE,                  /*Print sizes only*/
f_tsc   = FALSE,                  /*Output is to console screen*/
np,                               /*number of groups printed*/
nargs,                            /*number of non-option arguments*/
clsize  = 0,                      /*size of a cluster, in bytes*/
clmask,                           /*clsize-1 for rounding & chopping*/
drive                             /*code number for drive requested*/
;

long  left,                             /*unused space left on disk*/
total                             /*total of sizes encountered*/
;
void search(char *path,int size_em);
void parse_names(int argc,char *argv[],int size_em);
void curdrv(char *sp);
void getpath(char *sp);
int _showds(void);
int sysint(int interupt,SYSREG* r_in,SYSREG* r_out);
int disk_drive(void);
int find_first(char  *path,struct FILEDATA *dta,int  mask);
int find_next(struct FILEDATA *dta);
/*****************************************************************************/


/* If dosWildCard finds any wildcards in the args, it will allocate the */
/* needed buffer space, create a new argc and argv val and array, and */
/* modify the originals to reflect this. */


/* returns 1 if any wild cards in string, 0 if not. */

int
any_wildcards(char *string)
{
    if((strchr(string,'*') == NULL) && (strchr(string,'?') == NULL))
    {
        return 0;
    }
#if 1
    return 1;
#else
    /* Work on this later.  Bigger problem that I thought at first. */
    /* To act like unix, have to detect the wild, then strip out the */
    /* escape chars. Have to be careful when I strip, cause this */
    /* routine is called more than once */
    /* Got a wild card. Let's see if it is escaped */
    if((strchr(string,'\\') == NULL) && (strchr(string,'\'') == NULL))
    return 1;
    /* Theres an escape char in there.  Let's see if our wild card is escaped */
    lenStr = strlen(string);
    inquote = 0;
    for(i=0;i < lenStr;i++)
    {
        if((string[i] == '*') || (string[i] == '?'))
        {
            if(!inquote)
            /* Got a wild.  Not within quotes, Maybe escaped with \ */
        }
    }
#endif
}

static char **strPtrs; /* buf to put string pointers in */
static char *stringBuf; /* buf to put strings in */

static void
setCurDir(char *work)
{
    char *cptr;

    strcpy(curDir,work);
    if((cptr = strrchr(curDir,'\\')) != NULL)
    {
      cptr[1] = 0; /* strip off the file name */
    }
    else
    {
      if(work[1] != ':') /* If it is not a drive name */
        curDir[0] = 0;
      else
        curDir[2] = 0; /* the rest of it was a file name */
    }
}

static void
addStr2Size(char *string,int plusDir)
{
                if(plusDir)
                string_sizes += strlen(curDir);
                string_sizes += strlen(string) + 4;
                nbr_of_strings++;
                if(string_sizes > 64000)
                {
                    fprintf(stderr,"Too many files, aborting\n");
                    exit(-1);
                }
#ifdef DEBUG_DOSWLD
                printf("Sizing string %s\n",string);
                printf("string_sizes = %u\n",string_sizes);
                printf("curDir = %s\n",curDir);
                printf("nbr_of_strings = %d\n",nbr_of_strings);
#endif
}

static void
StoreString(char *string,int plusDir)
{

                if(plusDir)
                {
                    strcpy(strPtrs[pass2_string_nbr],curDir);
                    strcat(strPtrs[pass2_string_nbr],string);
                }
                else
                {
                    strcpy(strPtrs[pass2_string_nbr],string);
                }
				pass2_string_nbr++;
				strPtrs[pass2_string_nbr] = 
                strPtrs[pass2_string_nbr - 1] +
                strlen(strPtrs[pass2_string_nbr - 1]) + 2;
                if(plusDir)
                    added_string_sizes += strlen(curDir);
                added_string_sizes += strlen(string) + 4;
#ifdef DEBUG_DOSWLD
                printf("Storing string %s\n",string);
                printf("added_string_sizes = %u\n",added_string_sizes);
                printf("curDir = %s\n",curDir);
                printf("pass2_string_nbr = %d\n",pass2_string_nbr);
#endif
}

#endif /* unix */

void
dosWildCard(int *cargc, char **cargv[])
{
#ifdef unix
    return;
#else
    char *charPtr;
    int status;
    int dargc;
    char **dargv;
    unsigned int needStrPtrSize;
    unsigned int needStringBufSize;
    int i;
    string_sizes = 0;              /* size of strings */
    nbr_of_strings = 0;
	pass2_string_nbr = 0;
    dargc = *cargc;
    dargv = *cargv;
    spath[MAXSPATH + 1] = 0;
    curDir[0] = 0;


    for(i=0;i < dargc;i++)
    {
        if(any_wildcards(dargv[i]))
        {
            break;
        }
    }
    if(i == dargc)
      return;  /* No need to mess around */

    /* Got some wild cards, have to expand them */
    /* pass 1, see how much buffer space we need */
    parse_names(dargc,dargv,1);
    if(debug_doswld)
        printf("String sizes = %u\n",string_sizes);
    needStrPtrSize = (unsigned int)((nbr_of_strings + 4) * sizeof(charPtr)) + 4;
    needStringBufSize = (unsigned int)string_sizes + 4;
    strPtrs = (char **)malloc(needStrPtrSize + 1);
    if(strPtrs == NULL)
    {
        fprintf(stderr,"Unable to allocate wild card buf for %u files",
            nbr_of_strings);
            exit(-1);
    }
    stringBuf = malloc(needStringBufSize + 1);
    if(stringBuf == NULL)
    {
        fprintf(stderr,"Unable to allocate wild card buf of %u bytes",
            string_sizes);
            exit(-1);
    }

    charPtr = (char *)strPtrs;
    charPtr[needStrPtrSize] = (char)0x55; /* sanity check */
    stringBuf[needStringBufSize] = 0x55; /* sanity check */
    strPtrs[0] = stringBuf;
    /* pass 2, store file names in allocated buffer */
    parse_names(dargc,dargv,0);
    if(debug_doswld)
        printf("Added String sizes = %u\n",added_string_sizes);
    if((status = stringBuf[needStringBufSize]) != 0x55)
    {
        fprintf(stderr,"dosWld failed sanity check 1\n");
        exit(-1);
    }


    charPtr = (char *)strPtrs;
    if((status = charPtr[needStrPtrSize]) != (char)0x55)
    {
        fprintf(stderr,"dosWld failed sanity check 2\n");
        exit(-1);
    }

    if(nbr_of_strings != pass2_string_nbr)
    {
        fprintf(stderr,"DosWld failed sanity check 3\n");
        exit(-1);
    }
#if 1
    *cargc = nbr_of_strings;
    *cargv = (char **)strPtrs;
#endif
    return;
#endif /* unix */
}


#ifndef unix
static void
parse_names(argc, argv,size_em)
int  argc;
char *argv[];
int size_em;
{
    int i;
    int l;
    int  c = 0;
    int nt = 0;

    /* process input options */

    nargs = argc;

    if(argc is 0) 
    {
        argc++;
        curdrv(spath);                  /*default to current drive*/
        printf("Called curdrv\n");
    }
    else
        strncpy(spath, *argv,MAXSPATH);

    for(;;) 
    {                           /*cycle through args present*/
        i = 0;                      /*spath editing index*/
        l = strlen(spath);          /*spath string length*/
        curDir[0] = 0;

        /*the following while loop should be looked at carefully if things*/
        /*don't look right. compilers don't process the increment the same*/
        if(any_wildcards(spath))
        {
        while (i < l)                   /*Scan spath string...*/
        {
            spath[i] = (char)tolower(spath[i]); /*Convert to lower case*/
            i++;
        }
        }
#if 0
        if (spath[1] is ':' and     spath[2] is '\0') /*If path is drive only */
        {
            getpath(spath);             /*get path*/
            printf("Called getpath\n");
        }
#endif
        if (c is 0)
        {
            if(any_wildcards(spath))
            {
            search(spath,size_em);              /*go do the hard work*/
            }
            else
            {
            if(size_em) 
			{
                addStr2Size(spath,0);
            }
            else
            {
                StoreString(spath,0);
            }
            }
        }

        if (--argc > 0)
        {
            strncpy(spath, *++argv,MAXSPATH);
        }
        else
        {
            break;
        }
    }
}
#ifdef USE_QSORT
/* for qsort */
static int
cmpFileNames(char **s1,char **s2)
{
    return(strcmp((char *)*s1,(char *)*s2));
}
#endif /* USE_QSORT */

static int compare(unsigned int index1, unsigned int index2, char **array)
{
    return (strcmp(array[index1],array[index2]));
}

static void swapit(unsigned int index1, unsigned int index2, char **list)
{
    char *ptrtmp;
    ptrtmp = list[index1];
    list[index1] = list[index2];
    list[index2] = ptrtmp;
}

/* BubbleSort - BubbleSort cycles through the elements, comparing
 * adjacent elements and swapping pairs that are out of order. It
 * continues to do this until no out-of-order pairs are found.
 * pass in the array, its size, and pointers to two functions, one to
 * compare elements, one to swap elements.
 */

static void BubbleSort(char **list,unsigned int arraySize)
{
    unsigned int curString, lastSwap;

    arraySize--;
    if(arraySize < 1)
        return;
    /* Move the highest string down to the bottom until all are in order. */
    do
        {
        lastSwap = 0;
        for( curString = 0; curString < arraySize; curString++ )
        {
            /* If two adjacent elements are out of order, swap their values
                                                 * and redraw those two bars.
                                                 */
            if((compare)(curString,curString + 1, list) > 0)
            {
                swapit(curString, curString + 1, list);
                lastSwap = curString;
            }
        }

        /* Sort on next pass only to where the last switch was made. */
        arraySize = lastSwap;
    } 
    while( lastSwap );
}

/*****************************************************************************/
/* Expands one wild card string.  If size_em = 1, only bumps the size of */
/* the strings needed to store the expanded file names.  If size_em = 0 */
/* Stores the files in the allocated arrays */

/* Some variables for search.  Staticly defined here so recursive calls */
/* don't eat up stack space */

static char     work[MAXSPATH + 1];  /*working path string for search*/
static unsigned int start_string_nbr;
static int     mask = 0x0010;                /*attribute mask*/

static void
search(path,size_em)              /*search for filename or directory*/
char *path;
int size_em;
{
    struct FILEDATA dta;                          /*DOS file data table*/
    int     path_len;                     /*length of initial path*/
    int     z;                            /*char counter*/
    int     k;                            /*counts number of entries found*/
    long     bytes = 0;                   /*count of disk usage this directory*/
#ifdef USE_QSORT
    char *cptr;
#endif
    /* Setup to sort file names */
    k = 0;
    start_string_nbr = pass2_string_nbr;

    if (f_all) mask = 0x001F;
    strcpy(work,path);
    setCurDir(work);
    path_len = strlen(work);            /*save original path length*/

    if (!find_first(work, &dta, 0) or work[path_len-1] is qs) 
    {
        if(work[path_len-1] isnot qs) 
        {
            strcat(work, dqs);          /*if path is to a directory*/
            path_len++;
            setCurDir(work);
        }
        strcat(work,"*.*");             /*list everything in it*/
    }

    if(find_first(work, &dta, mask)) 
    {
        do 
        {
#ifndef __POWERC
            if (dta.attrib & 0x08)        /*ignore volume label*/
                continue;
            if (dta.name[0] is '.' and !f_all)  /*unless -a option*/
                continue;               /*ignore "." and ".."*/
            if(size_em) 
			{
                addStr2Size(dta.name,1);
            }
            else
            {
                StoreString(dta.name,1);
            }
            k++;
#else
            if (dta.ff_attrib & 0x08)        /*ignore volume label*/
                continue;
            if (dta.ff_name[0] is '.' and !f_all)  /*unless -a option*/
                continue;               /*ignore "." and ".."*/
            if(size_em) 
			{
                addStr2Size(dta.ff_name,1);
            }
            else
            {
                StoreString(dta.ff_name,1);
            }
            k++;
#endif /* __POWERC */
        } 
        while(!find_next(&dta));
    }
    else 
    {
        if (!f_recd)
        {
        work[path_len-1] = 0;
        fprintf(stderr, "Can't find a file or directory named \"%s\"\n", work);
        if(df_exitNF)
            exit(-1);
        }
    }

    work[path_len] = 0;              /*restore directory pathname*/

    /* sort files here */
    if(size_em == 0)
    {
#ifdef USE_QSORT
    /* Qsort works, but can't get rid of compiler warnings.  Some day */
    /* I'll figure it out, and use it.  It's a touch faster */
    /* Actually, I know now, just have to get to it. */
/* void __cdecl qsort(void *, size_t, size_t, int (__cdecl *)
    (const void *, const void *)); */
    cptr = (char *)&strPtrs[start_string_nbr];
        qsort((void *)cptr,
            (size_t)k,(size_t)sizeof(strPtrs[0]),cmpFileNames);
#else
    if(k)
    BubbleSort(&strPtrs[start_string_nbr],k);
#endif
    }
    if (!f_recd) 
        return;                         /*quit if not -R*/

    strcat(work, "*.*");
    if(find_first(work, &dta, mask))    /*else find all sub-dirs*/
        do     
        {
#ifndef __POWERC
        if(dta.attrib & 0x10 and dta.name[0] isnot '.')
        {
            work[path_len] = 0;     /*discard old name              */
            for(z=0; dta.name[z] isnot 0; z++)
                dta.name[z] = tolower(dta.name[z]);
            strcat(work, dta.name);/*install a new one*/
            strcat(work, dqs);
#ifdef DEBUG_DOSWLD
            printf("Recursing into %s\n",work);
#endif
            search(work,size_em);           /*and recurse*/
        }
#else
        if(dta.ff_attrib & 0x10 and dta.ff_name[0] isnot '.')
        {
            work[path_len] = 0;     /*discard old name              */
            for(z=0; dta.ff_name[z] isnot 0; z++)
                dta.ff_name[z] = (char)tolower(dta.ff_name[z]);
            strcat(work, dta.ff_name);/*install a new one*/
            strcat(work, dqs);
#ifdef DEBUG_DOSWLD
            printf("Recursing into %s\n",work);
#endif
            search(work,size_em);           /*and recurse*/
        }
#endif /* __POWERC */
    } 
    while(!find_next(&dta));
    return;
}
/******************************************************************************/
static void
getpath(sp)                    /*get drive's current directory path*/
char     *sp;
{
#if 1
    struct {                    /*Register file for GET PATH*/
        int   ax;
        int   bx;
        int   cx;
        int   dx;
        char* si;
        int   di;
        int   ds;
        int   es;
    } r;
#else
    SYSREG  r;
#endif

    strcat(sp, dqs);                      /*append root symbol to drive name*/

    /*Prepare to get current path...*/
    r.ax = 0x4700;                        /*Set 'get current directory' code*/
    r.dx = *sp - '`';                     /*Set drive number*/
    r.ds = _showds();                     /*Set pathname buffer segment*/
    r.si = sp + 3;                        /*Set pathname buffer offset*/
    sysint(0x21, (SYSREG *)&r, (SYSREG *)&r);                 /*Call DOS*/
}

/*****************************************************************************/
/*                                                                           */
/*                Compiler-specific code for                                 */
/*                                                                           */
/*                 Microsoft version 5.0 and later                           */
/*                                                                           */
/*****************************************************************************/

/*********************************/

/*********************************/
static int
find_first(path, dta, mask)          /*find first file in directory*/
char     *path;
struct FILEDATA *dta;
int     mask;
{
#ifndef __POWERC
    return(!(_dos_findfirst(path, mask , dta)));
#else
    return(!(findfirst(path, dta,mask)));
#endif
}

/*********************************/

static int
find_next(dta)                    /*find next file in same directory*/
struct FILEDATA *dta;
{
#ifndef __POWERC
    return(_dos_findnext(dta));
#else
    return(findnext(dta));
#endif
}

/*********************************/
static void
curdrv(sp)                    /*get current default drive*/
char     *sp;
{
    *sp++ = disk_drive() + 'a';
    *sp++ = ':';
}

/*********************************/

static int _showds()                    /*Obtain DS register value*/
{
    struct SREGS sfile;

    segread(&sfile);

    return(sfile.ds);
}


/*********************************/

static int
sysint(interupt,r_in,r_out)          /*Interface sysint() to int86x()*/
int     interupt;
SYSREG* r_in;
SYSREG* r_out;
{
    int     result;

    union  REGS  rfile;
    struct SREGS sfile;

    segread(&sfile);

    rfile.x.ax = r_in->ax;
    rfile.x.bx = r_in->bx;
    rfile.x.cx = r_in->cx;
    rfile.x.dx = r_in->dx;
    rfile.x.si = r_in->si;
    rfile.x.di = r_in->di;

    sfile.ds = r_in->ds;
    sfile.es = r_in->es;

    result = int86x(interupt,&rfile,&rfile,&sfile);

    r_out->ax = rfile.x.ax;
    r_out->bx = rfile.x.bx;
    r_out->cx = rfile.x.cx;
    r_out->dx = rfile.x.dx;
    r_out->si = rfile.x.si;
    r_out->di = rfile.x.di;

    r_out->ds = sfile.ds;
    r_out->es = sfile.es;

    return(result);
}

/*********************************/

static int
disk_drive()
{
    unsigned drive;
#ifndef __POWERC
    _dos_getdrive(&drive);
#else
    drive = getdisk();
#endif
    return(drive - 1);
}

#endif /* unix */

#ifdef DOSWLD_MAIN
void
main(argc, argv)
int  argc;
char *argv[];
{
    int i;

    printf("Doswld rev %s\n",REVISION);
    printf("Args before doswld expansion\n");
    for(i=0;i < argc;i++)
    {
        printf("Printing argv[%d] >%s<\n",i,argv[i]);
    }

    printf("\nArgs after doswld expansion\n");
/*    f_recd = TRUE; */
    dosWildCard(&argc,&argv);
    for(i=0;i < argc;i++)
    {
        printf("Printing argv[%d] >%s<\n",i,argv[i]);
    }
}
#endif

