/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
/* See fsort.doc*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef SUNOS4_1_2
#include <unistd.h>
#endif
#ifndef unix
#include <string.h>
#include <ctype.h>
#include <direct.h>
#include <dos.h>
#include <io.h>
#endif
#include "fsort.h"

static char *REVISION = "2.0.2";

#if 0
#define DEBUG_FSORT
#define DEBUG_FP stdout
#endif

#ifndef sys_errlist
extern char *sys_errlist[];
#endif

#define TEMP_FILE_END   ".TMP"
#ifndef unix
#define TEMP_FILE_DIR_START "c:\\tmp"
#define TEMP_FILE_START "\S"
#else
#define TEMP_FILE_DIR_START "/tmp"
#define TEMP_FILE_START "/S"
#endif /* unix */

#ifndef unix
#ifdef USE_FAR
#define MAX_NUM_LINES   10000
#else
#define MAX_NUM_LINES   200
#endif
#else
#define MAX_NUM_LINES   10000
#endif /* unix */
#define MIN_NUM_LINES 200

unsigned int max_num_lines = MAX_NUM_LINES;
unsigned int lines_read_in;
unsigned int tmp_files_created;
int show_files_records = 0;  /* calling program can set this for debugging */
int delete_tmp_files = 1;    /* calling program can clear this for debugging*/
static size_t needStrPtrSize;
static unsigned long totalRecordBufSize;
static unsigned long storedSize;
static unsigned int thisFarBufSize;
static char FAR *charPtr;

static char FAR *readBuf1;
static char FAR *readBuf2;
static char temp_file_start[FILENAME_MAX];
char temp_directory[FILENAME_MAX]; /* used by other programs */

static char FAR **strPtrs = NULL; /* buf to put string pointers in */

/* Possibly used by applications with binaryMemorySearch */
int farBufInUse = 0;
int nbrFarBuffersAllocated = 0;
char FAR *farBuffers[MAX_FAR_BUFFERS];
unsigned int farBufferSizes[MAX_FAR_BUFFERS];

int max_far_buffers = MAX_FAR_BUFFERS;

static char openFileWriteMode[5];
static char openFileReadMode[5];
 
#ifdef NEED_MEMMOVE
/* for machines that don't have memmove */
void memmove (char *to,char  *from, int  length)
{
    char *T, *F;

    if (to < from) 
	{
    /* left to right */
	T = to;
	F = from;
	while (length--)
	    *(T++) = *(F++);
    }
    else 
	{
    /* right to left */
	T = to + length - 1;
	F = from + length - 1;
	while (length--)
	    *(T--) = *(F--);
    }
}
#endif

/* for qsort */
static int
cmpStrings(char **s1,char **s2)
{
    char FAR * ptr1;
    char FAR * ptr2;

#ifdef USE_FAR
    /* Qsort sorts an array in near memory.  But the array is one of */
    /* Far pointers.  Haven't found anyother way using "c" to get */
    /* A far pointer through a near pointer.  This one works */
    /* even when not using far, when FAR == nothing */
    memcpy(&ptr1,s1,sizeof(char FAR *));
    memcpy(&ptr2,s2,sizeof(char FAR *));
    return(STRCMP((char FAR *)ptr1,(char FAR *)ptr2));
#else
    return(STRCMP(*s1,*s2));
#endif /* DEBUG_FSORT */
}

static int
readLine(FILE *fp_src,char FAR *buf,int maxlen)
{
    int size;
    char FAR * ptr;
#ifdef USE_FAR
    char tmpBuf[1024];
    ptr = (char FAR *)buf;
    ptr = (char FAR *)*buf;

    /* I'm not aware of any DOS read that will read directly into a */
    /* far buffer, so this reads into the near tmpBuf, then MEMMOVE */
    /* copies it to the far buffer */
    if(fgets(tmpBuf, maxlen, fp_src) == 0) return 0;
    tmpBuf[(size = strlen(tmpBuf)) -1] = '\0';
    MEMMOVE((char FAR *)buf,(char FAR *)&tmpBuf[0],(size_t) maxlen);
#else
    if(fgets(buf, maxlen, fp_src) == 0) return 0;
    buf[(size = strlen(buf)) -1] = '\0';
#endif
    return size;
}

static int
writeLine(FILE *fp_out,char FAR *buf)
{
    char FAR * ptr;
#ifdef USE_FAR
    ptr = (char FAR *)buf;
    ptr = (char FAR *)*buf;
    /* If you are writing a binary record, you must do the opposite */
    /* of the readLine.  MEMMOVE the far data into a near tmp buf, then */
    /* fwrite the data.  See readLine comments. */
    fprintf(fp_out,"%Fs\n",(char FAR *)buf);
#else
    fprintf(fp_out,"%s\n",buf);
#endif

    return 1;
}

/*  MAKE_FILE_NAME */
 
static void
make_file_name(int filenum, char *destination)
/*int   filenum; */
/*char  destination[];*/ 
{
    char  numbuffer[80]; 
 
    destination[0] = 0;
#if 0
    sprintf(numbuffer, "t%x", filenum);
#else
    strcat(destination, temp_file_start);
#ifdef unix
    sprintf(numbuffer, "%x_%x", getpid(),filenum);
#else
    sprintf(numbuffer, "%x", filenum);
#endif   
#ifndef unix
    if (strlen(numbuffer) > 8 )
    {
        fprintf(stderr, "\nToo many merge files. ");
#if 0
          exitStats();
#endif
        exit(1);
    }
#endif
#endif
    strcat(destination, numbuffer);
    strcat(destination, TEMP_FILE_END);
}

void
get_tmp_directory(int show_warnings)
{
    char *t;
	struct stat buffer;

    strcpy(temp_directory,TEMP_FILE_DIR_START);

    if((t = getenv("TMP")) ||
        (t = getenv("tmp")) ||
        (t = getenv("TEMP")) ||
        (t = getenv("temp")))
    {
     strcpy(temp_directory,t);
    }
    /* got a directory name.  Is it reachable */

    if(stat(temp_directory,&buffer) == -1)
	{
        if(show_warnings)               /* 1 = return status */
        {
            fprintf(stderr,"Can't access tmp directory \"%s\"\n",
             temp_directory);
            fprintf(stderr,"Tmp files normally go in %s\n", TEMP_FILE_DIR_START);
            fprintf(stderr,"or as specified by TMP or TEMP environmental variables\n");
            fprintf(stderr,"Defaulting to current directory\n");
        }
        /* We will put tmp files in the current directory */
        temp_directory[0] = 0;
	}
	else
	{
		/* if we could stat it, it exists. */
        /* Possible flaw in this logic.  It could be a file */
#ifdef unix
        strcat(temp_directory,"/");
#else
        strcat(temp_directory,"\\");
#endif /* unix */
	}
}

static void
init_tmp_file_start(fsort_args *sparms)
{
     get_tmp_directory(!sparms->PassBackErrorStatus);
     strcpy(temp_file_start,temp_directory);
     strcat(temp_file_start,"S");
}

static void
SetupNextRecPtr(fsort_args *sparms,int string_nbr_2_store,int size)
{
    char FAR *cptr;
    /* set up pointer to next string, if we get one */
    cptr = strPtrs[string_nbr_2_store];
    cptr += size + 2;
    storedSize += size + 2;
    thisFarBufSize += size + 2;
    if((thisFarBufSize + sparms->maxRecSize + 2) >= farBufferSizes[farBufInUse])
    {
        farBufInUse++;
        thisFarBufSize = size;
        cptr = farBuffers[farBufInUse];
#if 0
       fprintf(stderr,"Working on far buf %d\n",farBufInUse);
#endif
    }
    strPtrs[string_nbr_2_store + 1] = cptr;
    /* Next string is null terminated */
    /* unless we bumped up past allocated buffers */
    if((farBufInUse + 1) <nbrFarBuffersAllocated)
        *cptr = 0;
}


/* Works best for fixed length records */
/* Returns 0 if record found and is in destBuf */
int debug_binary_sort = 0;
int
binaryFileSearch(fsort_args *sparms, FILE *fp,char FAR *source,char FAR *destBuf)
{
    long start,mid,end,old;
    int pos;
    int status;
    long unit;
    long where;

    unit = 1;

    if(sparms->fixedLength)
      unit = sparms->maxRecSize;
    start = 0;
#ifdef unix
    fseek(fp,0L,SEEK_END);
	end = ftell(fp);
#else
    end = filelength(fileno(fp));
#endif
    end = end / unit;

    old = 0;
    old--;  /* turns unsigned into -1 */

    while(1)
    {
        mid = ((end - start + 1) / 2) + start;
        if(mid == old)
        return(1); /* not found */
        old = mid;
		if(debug_binary_sort)
		printf("start 0x%lx, mid 0x%lx, end 0x%lx\n",
			start * unit,mid * unit,end * unit);
        where = mid * unit;
        pos = fseek(fp,(long)where,SEEK_SET);
        if(pos)
        {
            if(sparms->PassBackErrorStatus)
            {
                return(errno);
            }
            else
            {
                fprintf(stderr,"Error seeking to 0x%lx\n",(long)mid * unit);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
                return (-1);
            }
        }
        /* if random length text file, first read is just to find eol */
        if((sparms->fileType == 0) && (sparms->fixedLength == 0))
        {
            (*sparms->readRecord)(fp,(char FAR *)destBuf, sparms->maxRecSize);
            mid = ftell(fp) / unit; /* mid is always lined up at start of line */
        }
        if(!(*sparms->readRecord)(fp,(char FAR *)destBuf, sparms->maxRecSize))
		  return 1; /* zero read, not found */
		
        status = (*sparms->compareRecord)((char *)&source,
            (char *)&destBuf);
        if(status == 0)
          return 0;  /* FOUND IT */
        if(status < 0)
        {
			if(mid == 0)
		  		return 1; /* Not found */
			if((end == 1) && (mid == 1))
			  end = 0;
			else
            	end = mid;
        }
        else
        {
            start = mid;
        }
    }
    return(-1); /* Just to shut up compiler */
}

#if 1

/* Search an array in FAR memory. Typical use would be a file that */
/* was sorted with sort_file, then read into FAR memory.  "last" would */
/* not be necessary if this would only be searched for fixed length */
/* records, but since this can be used for both, it is a neccesary evil. */
/* source is what we are searching for, destBuf is left pointing to */
/* what was found if it was found. */
int
binaryMemorySearch(fsort_args *sparms, unsigned int nbrElements,
    char FAR *base, char FAR *last, char FAR *source,char FAR *destBuf)
{
    unsigned long start,mid,end,where,old;
    char FAR *mid_ptr;
    int status;
    unsigned long unit;

    if(sparms->fixedLength)
      unit = sparms->maxRecSize;
    else
      unit = 1;
    /* use mid, start as tmps.  pointer arithmatic doesn't work right */
#if 1
    mid = (unsigned long) base;
    start = (unsigned long) last;
    end = (start - mid) / unit; /* (last - base) / unit */
#else
    end = (last - base) / unit; /* what I tried to do */
#endif
    start = 0;

    old = 0;
    old--;  /* turns unsigned into -1 */

    while(1)
    {
        mid = ((end - start + 1) / 2) + start;
        if(mid == old)
         return(1); /* not found */
#if 1
        if((old == 0) && (mid == 0))
            return(1); /* not found */
        if((mid - 1) == (unsigned long)nbrElements)
            return(1); /* not found */
#endif
        old = mid;
        where = mid * unit;
        mid_ptr = base + where;
#if 0
        /* if random length text file, first read is just to find eol */
        if((sparms->fileType == 0) && (sparms->fixedLength == 0))
        {
            (*sparms->readRecord)(fp,(char FAR *)destBuf, sparms->maxRecSize);
            mid = ftell(fp) / unit; /* mid is always lined up at start of line */
        }
        if(!(*sparms->readRecord)(fp,(char FAR *)destBuf, sparms->maxRecSize))
		  return 1; /* zero read, not found */
#endif
        status = (*sparms->compareRecord)((char *)&source,
                (char *)&mid_ptr);
        if(status == 0)
        {
          /* This copies a far mid_ptr to a far pointer pointed to by */
          /* a near pointer.  Confusing isn't it */
/*          memcpy(destBuf,&mid_ptr,sizeof(mid_ptr)); */
          MEMMOVE(destBuf,mid_ptr,sparms->maxRecSize);
          return 0;  /* FOUND IT */
        }
        if(status < 0)
        {
#if 1
			if(mid == 0)
		  		return 1; /* Not found */
			if((end == 1) && (mid == 1))
			  end = 0;
			else
            	end = mid;
#endif
        }
        else
        {
            start = mid;
        }
    }
    return(-1); /* Just to shut up compiler */
}
#endif

#if 1
/* Just to debug far allocation problems */
void heapdump(char *file,int line)
{
#ifndef __POWERC
#ifndef unix
    _HEAPINFO hinfo;
    int heapstatus;
    printf("Heap dump from %s line %d\n",file,line);
    hinfo._pentry = NULL;
    printf("Near heap %u\n",_memmax());
    while ((heapstatus = _fheapwalk(&hinfo)) == _HEAPOK)
    {
        printf("%6s block at %Fp of size 0X%4.4X\n",
          (hinfo._useflag == _USEDENTRY ? "USED" : "FREE" ),
          hinfo._pentry,hinfo._size);
    }
    switch(heapstatus)
    {
        case _HEAPEMPTY:
           printf("OK - empty heap\n");
           break;
        case _HEAPEND:
           printf("OK - end of heap\n");
           break;
        case _HEAPBADPTR:
           printf("ERROR - bad pointer to heap\n");
           break;
        case _HEAPBADBEGIN:
           printf("Error - bad start of heap\n");
           break;
        case _HEAPBADNODE:
           printf("ERROR - bad node in heap\n");
           break;
        default:
           printf("ERROR - Unknown error type %d\n",heapstatus);
           break;
    }
    fflush(stdout);
#endif /* unix */
#endif /* __POWERC */
}
#endif
static int
writeRecords(fsort_args *sparms,char *outfile)
{
    unsigned int i;
	FILE *fp_out;
    char FAR *cptr1;
    char FAR *cptr2;


    if(outfile[0] != 0)
    {
        fp_out = fopen(outfile,openFileWriteMode);
    }
    else
        fp_out = stdout;

    if(fp_out == NULL)
    {
            if(sparms->PassBackErrorStatus)
                return(errno);
            else
            {
                fprintf(stderr,"Unable to open %s\n",outfile);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
                exit(-1);
            }
    }
	for(i=0;i < lines_read_in;i++)
	{
        /* Gets rid of some dups, but not all. */
        if(sparms->deleteDuplicateRecords)
        {
            if(i != 0)
            {
                cptr1 = strPtrs[i];
                cptr2 = strPtrs[i -1];
                if((*sparms->compareDupRecord)((char *)&cptr1,
                        (char *)&cptr2))
                    (*sparms->writeRecord)(fp_out,(char FAR *)strPtrs[i]);
            }
            else
                (*sparms->writeRecord)(fp_out,(char FAR *)strPtrs[i]);
        }
        else
        (*sparms->writeRecord)(fp_out,(char FAR *)strPtrs[i]);
	}
	fclose(fp_out);
    if(show_files_records)
      fprintf(stderr,"Sort pass with file %s %d records\n",outfile,lines_read_in);
    return 0;
}

/* tmp_files_created ends up pointing at the last tmp file name used, which*/
/* for small files may be 0 */
static int
do_sort_pass(fsort_args *sparms)
{
	FILE *fp_in;
    int size;
	char outfile[80];
    void (*sortit)() = qsort;

    fp_in = fopen(sparms->inFileName,openFileReadMode);
    if(fp_in == NULL)
    {
            if(sparms->PassBackErrorStatus)
                return(errno);
            else
            {
                fprintf(stderr,"Unable to open %s\n",sparms->inFileName);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
                exit(-1);
            }
    }
	lines_read_in = 0;
	tmp_files_created = 0;
	make_file_name(tmp_files_created,&outfile[0]);
    while(size = (*sparms->readRecord)(fp_in,(char FAR *)strPtrs[lines_read_in],
    sparms->maxRecSize))
    {
        SetupNextRecPtr(sparms,lines_read_in,size);
		lines_read_in++;
        if(((lines_read_in + 1) >=  max_num_lines) ||
           (farBufInUse == nbrFarBuffersAllocated) ||
           ((farBufInUse == nbrFarBuffersAllocated -1) &&
           ((thisFarBufSize + sparms->maxRecSize + 10) >= farBufferSizes[farBufInUse])))
		{
        (*sortit)(strPtrs,
            (size_t)lines_read_in,
            (size_t)sizeof(charPtr),
            sparms->compareRecord);
            writeRecords(sparms,outfile);
			tmp_files_created++;
			lines_read_in = 0;
            farBufInUse = 0;
            storedSize = 0;
            thisFarBufSize = 0;
            strPtrs[0] = farBuffers[0];
            strPtrs[1] = NULL;
			make_file_name(tmp_files_created,&outfile[0]);
		}
	}
	if(lines_read_in)
	{
        (*sortit)(strPtrs,
            (size_t)lines_read_in,
            (size_t)sizeof(charPtr),
            sparms->compareRecord);
            writeRecords(sparms,outfile);
            farBufInUse = 0;
            thisFarBufSize = 0;
	}
	else
		if(tmp_files_created)
			tmp_files_created--;  /* Didn't use that one */
	fclose(fp_in);
    return 0;
}


/* returns 0 if ok, -1 if couldn't delete */
static int 
deletefile(char *name)
{
    int status;

#ifdef MAIN_DEMO
    fprintf(stderr,"Deleting %s\n",name);
#endif
#ifndef DOS
    status = unlink(name);
#else
    status = remove(name);
#endif /* DOS */
    if(status)
    {
        fprintf(stderr,"Unable to delete %s\n",name);
        fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
    }
    return status;
}

static int
merge_2_files(fsort_args *sparms,char *in1, char *in2, char *outFile)
{
	FILE *fp_in1;
	FILE *fp_in2;
	FILE *fp_out;
    int status1,status2;
    int lines;
    char FAR *cptr1;
    char FAR *cptr2;

    cptr1 = readBuf1;
    cptr2 = readBuf2;

    lines = 0;
    fp_in1 = fopen(in1,openFileReadMode);
    if(fp_in1 == NULL)
    {
            if(sparms->PassBackErrorStatus)
                return(errno);
            else
            {
                fprintf(stderr,"Unable to open %s\n",in1);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
                exit(-1);
            }
	}

    fp_in2 = fopen(in2,openFileReadMode);
    if(fp_in2 == NULL)
    {
            if(sparms->PassBackErrorStatus)
                return(errno);
            else
            {
                fprintf(stderr,"Unable to open %s\n",in2);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
                exit(-1);
            }
	}
    fp_out = fopen(outFile,openFileWriteMode);
    if(fp_out == NULL)
    {
            if(sparms->PassBackErrorStatus)
                return(errno);
            else
            {
                fprintf(stderr,"Unable to open %s\n",outFile);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
                exit(-1);
            }
	}
    
    status1 = (*sparms->readRecord)(fp_in1,(char FAR *)readBuf1,sparms->maxRecSize);
    status2 = (*sparms->readRecord)(fp_in2,(char FAR *)readBuf2,sparms->maxRecSize);
	while(status1 && status2)
	{
            if((*sparms->compareRecord)((char *)&cptr1,(char *)&cptr2) > 0)
			{
              (*sparms->writeRecord)(fp_out,(char FAR *)readBuf2);
              lines++;
              status2 = (*sparms->readRecord)(fp_in2,(char FAR *)readBuf2,
                sparms->maxRecSize);
			}
			else
			{
              (*sparms->writeRecord)(fp_out,(char FAR *)readBuf1);
              lines++;
              status1 = (*sparms->readRecord)(fp_in1,(char FAR *)readBuf1,
                sparms->maxRecSize);
			}
	}

	/* Flush the one that's not empty */
	if(status1)
	{
        (*sparms->writeRecord)(fp_out,(char FAR *)readBuf1);
        lines++;
        while((*sparms->readRecord)(fp_in1,(char FAR *)readBuf1,
                sparms->maxRecSize))
        {
              (*sparms->writeRecord)(fp_out,(char FAR *)readBuf1);
              lines++;
        }
	}

	if(status2)
	{
        (*sparms->writeRecord)(fp_out,(char FAR *)readBuf2);
        while((*sparms->readRecord)(fp_in2,(char FAR *)readBuf2,
                sparms->maxRecSize))
        {
              (*sparms->writeRecord)(fp_out,(char FAR *)readBuf2);
              lines++;
        }
	}

	fclose(fp_in1);
	fclose(fp_in2);
	fclose(fp_out);
    if(show_files_records)
      fprintf(stderr,"Merge pass with file %s %d records\n",outFile,lines);
    return 0;
}

static int
do_merge_pass(fsort_args *sparms)
{
    unsigned int infilenbr;
	char inFile1[80];
	char inFile2[80];
	char outfile[80];
    char FAR *cptr1;
    char FAR *cptr2;
    char FAR *cptr3;
    int i;
	FILE *fp_in;
    FILE *fp_out;

#ifdef DEBUG_FSORT
    fprintf(DEBUG_FP,"IN do_merge_pass\n");
#endif
	infilenbr = 0;
	tmp_files_created++;
	make_file_name(tmp_files_created,&outfile[0]);
	make_file_name(infilenbr,&inFile1[0]);
	make_file_name(infilenbr + 1,&inFile2[0]);
	while((infilenbr + 1) <  tmp_files_created)
	{
#ifdef DEBUG_FSORT
    fprintf(DEBUG_FP,"Merging %s and %s\n",inFile1, inFile2);
#endif
        merge_2_files(sparms,inFile1, inFile2, outfile);
        if(delete_tmp_files)
        {
            deletefile(inFile1);
            deletefile(inFile2);
        }
		infilenbr += 2;
		tmp_files_created++;
		make_file_name(infilenbr,&inFile1[0]);
		make_file_name(infilenbr + 1,&inFile2[0]);
		make_file_name(tmp_files_created,&outfile[0]);
	}

    fp_in = fopen(inFile1,openFileReadMode);
    if(fp_in == NULL)
    {
            if(sparms->PassBackErrorStatus)
                return(errno);
            else
            {
                fprintf(stderr,"Unable to open %s\n",inFile1);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
                exit(-1);
            }

	}
    if(sparms->outFileName[0] != 0)
    {
        fp_out = fopen(sparms->outFileName,openFileWriteMode);
        if(fp_in == NULL)
        {
            if(sparms->PassBackErrorStatus)
                return(errno);
            else
            {
                fprintf(stderr,"Unable to open %s\n",sparms->outFileName);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
                exit(-1);
            }
        }
    }
    else
    {
      fp_out = stdout;
    }
    i = 0;

    /* Final pass, read tmp file and copy it to output file name, deleting */
    /* dup records if desired */
    cptr1 = readBuf1;
    cptr2 = readBuf2;
    cptr3 = cptr1;
    while((*sparms->readRecord)(fp_in,(char FAR *)cptr3, sparms->maxRecSize))
	{
        if(sparms->deleteDuplicateRecords)
        {
            if(i != 0)
            {
                if((*sparms->compareDupRecord)((char *)&cptr1,
                        (char *)&cptr2))
                    (*sparms->writeRecord)(fp_out,(char FAR *)cptr3);
            }
            else
                (*sparms->writeRecord)(fp_out,(char FAR *)cptr3);

            /* alternate reading into buf 1 and 2 so we can compare for */
            /* duplicates, to delete the dups */
            if(cptr3 == cptr1)
                cptr3 = cptr2;
            else
                cptr3 = cptr1;
        }
        else
        {
            (*sparms->writeRecord)(fp_out,(char FAR *)cptr3);
        }

        i++;
	}
    if(delete_tmp_files)
        deletefile(inFile1);
	fclose(fp_in);
    if(fp_out != stdout)
      fclose(fp_out);
    return 0;
}


int
init_fsort(fsort_args *sparms)
{
    int i;
    unsigned int tryAlloc;
    unsigned int TryFarBufferSize;
    unsigned int maxNumLinesTries;

    if(sparms->maxRecSize == 0)
        sparms->maxRecSize = MAX_STR_LEN;
    if(sparms->compareRecord == NULL)
        sparms->compareRecord = cmpStrings;
    if(sparms->compareDupRecord == NULL)
        sparms->compareDupRecord = cmpStrings;
    if(sparms->readRecord == NULL)
        sparms->readRecord = readLine;
    if(sparms->writeRecord == NULL)
        sparms->writeRecord = writeLine;


    if(sparms->fileType == 0)
    {
        strcpy(openFileWriteMode, "w");
        strcpy(openFileReadMode,"r");
    }
    else
    {
#ifndef unix
        strcpy(openFileWriteMode, "wb");
        strcpy(openFileReadMode,"rb");
#else
        strcpy(openFileWriteMode, "w");
        strcpy(openFileReadMode,"r");
#endif
    }

    lines_read_in = 0;
    tmp_files_created = 0;
    totalRecordBufSize = 0;
    storedSize = 0;
    thisFarBufSize = 0;
    farBufInUse = 0;
    nbrFarBuffersAllocated = 0;

    for(max_num_lines = MAX_NUM_LINES; max_num_lines > MIN_NUM_LINES;
        max_num_lines--)
    {
    /* max lines + 2.  One extra for null terminated string, last for sanity */
    needStrPtrSize = (sizeof(charPtr)) * max_num_lines + 2;
#ifndef unix
    TryFarBufferSize = 0;
    TryFarBufferSize--; /* 0 - 1 == max number in unsigned math */
    TryFarBufferSize--; /* 0 - 1 == max number in unsigned math */
#else
    TryFarBufferSize = 0xfffe;  /* about 64k */
#endif
    TryFarBufferSize -= (2 * sparms->maxRecSize); /* so we don't overflow */
    thisFarBufSize = 0;
    strPtrs = NULL; /* for debug */
    strPtrs = (char FAR**)malloc((size_t)needStrPtrSize + 1);
    if(strPtrs == NULL)
    {
#if 0
            if(sparms->PassBackErrorStatus)
                return(errno);
            else
            {
                fprintf(stderr,
                "Unable to allocate pointer buf for %u bytes, %u lines\n",
                    needStrPtrSize,max_num_lines);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
                exit(-1);
            }
#else
    continue;
#endif
    }
    else
      break; /* out of while loop */
    }
    if(max_num_lines < MIN_NUM_LINES)
    {
            if(sparms->PassBackErrorStatus)
                return(errno);
            else
            {
                fprintf(stderr,
                "Unable to allocate pointer buf for %u bytes, %u lines\n",
                    needStrPtrSize,max_num_lines);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
#if 1
                heapdump(__FILE__,__LINE__);
#endif
                exit(-1);
            }
    }
    readBuf1 = (char FAR *)MALLOC(sparms->maxRecSize + 1);
    if(readBuf1 == NULL)
    {
            if(sparms->PassBackErrorStatus)
                return(errno);
            else
            {
                fprintf(stderr,"Unable to allocate readBuf1 of %u bytes",
                    sparms->maxRecSize);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
                exit(-1);
            }
    }
    readBuf2 = (char FAR *)MALLOC(sparms->maxRecSize + 1);
    if(readBuf2 == NULL)
    {
            if(sparms->PassBackErrorStatus)
                return(errno);
            else
            {
                fprintf(stderr,"Unable to allocate readBuf2 of %u bytes",
                    sparms->maxRecSize);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
                exit(-1);
            }
    }

    nbrFarBuffersAllocated = 0;
    farBufInUse = 0;
    totalRecordBufSize = 0;
    for(i=0;i < max_far_buffers;i++)
    {
        tryAlloc = TryFarBufferSize + 1;
        /* If we can't get a huge buffer, try smaller and smaller till we */
        /* get one */
        while(((farBuffers[i] = (char FAR *)MALLOC((size_t)tryAlloc + 2 +
            sparms->maxRecSize)) == NULL) && (tryAlloc > 32000))
            {
                tryAlloc -= sparms->maxRecSize;
            }
        if(farBuffers[i] == NULL)
        {
            if(nbrFarBuffersAllocated == 0)
            {
            if(sparms->PassBackErrorStatus)
                return(errno);
            else
            {
                fprintf(stderr,"Unable to allocate string buf of %u bytes",
                    tryAlloc);
                fprintf(stderr,"Error = %d  %s\n",errno,sys_errlist[errno]);
#if 0
                heapdump(__FILE__,__LINE__);
#endif
                exit(-1);
            }
            break; /* out of for loop */
            }
        }
        else
        {
            farBufferSizes[i] = tryAlloc;
            totalRecordBufSize += (tryAlloc - sparms->maxRecSize);
            charPtr = (char FAR *)farBuffers[i];
            charPtr[farBufferSizes[i] ] = 0x55; /* sanity check */
            nbrFarBuffersAllocated++;
        }
    }
#ifdef MAIN_DEMO
	fprintf(stderr,"%d Far buffers allocated\n",nbrFarBuffersAllocated);
#endif

    charPtr = (char FAR *)strPtrs;
    charPtr[needStrPtrSize - 1] = (char)0x55; /* sanity check */
    strPtrs[0] = farBuffers[0];
    strPtrs[1] = NULL;
    storedSize = 0;
    return 0;
}

int
close_fsort(fsort_args *sparms)
{
    int i;
    if(strPtrs != NULL)
    {
        if(charPtr != (char *)strPtrs)
            fprintf(stderr,"strPtrs modified 2\n");
        if(charPtr[needStrPtrSize - 1] != (char)0x55)
            fprintf(stderr,"failed sanity test 3\n");
    }

    FREE(readBuf1);
    FREE(readBuf2);
    if(strPtrs != NULL)
        free(strPtrs);
    for(i=0;i < nbrFarBuffersAllocated;i++)
    {
        charPtr = (char FAR *)farBuffers[i];
        if(charPtr[farBufferSizes[i] ] != 0x55)
        {
            fprintf(stderr,"failed sanity test 4 in buf %d\n",i);
            exit(-1);
        }
        FREE(farBuffers[i]);
    }
    return(0);
}

/* Some programs use fsorts far buffers and have no need for the local */
/* near buffers.  This frees them up.  NULL is used detecte it has been done */
void
free_fsort_local_buffers()
{
    if(strPtrs != NULL)
        free(strPtrs);
    strPtrs = NULL;
}

int
sort_file(fsort_args *sparms)
{

    init_tmp_file_start(sparms);

    init_fsort(sparms);

    do_sort_pass(sparms);
    if(charPtr != (char *)strPtrs)
      fprintf(stderr,"strPtrs modified 1\n");
    if(charPtr[needStrPtrSize - 1] != (char)0x55)
      fprintf(stderr,"failed sanity test 1\n");
    do_merge_pass(sparms);
    close_fsort(sparms);
    return 0;
}

void
set_sort_file_defaults(fsort_args *sparms)
{
    memset((char *)sparms,0,sizeof(sparms));
    strcpy(sparms->inFileName,"");     /* File to sort */
    strcpy(sparms->outFileName,"");    /* Destination file name */
    sparms->maxRecSize = 0;            /* record size */
    sparms->fileType = 0;              /* text type */
    sparms->deleteDuplicateRecords = 0; /* Delete Duplicate Records */
    sparms->fixedLength = 0;           /* Not fixed length records */
    sparms->PassBackErrorStatus = 0;   /* abort on error */
    sparms->compareRecord = NULL;      /* use default compare routine*/
    sparms->compareDupRecord = NULL;   /* use default compare routine */
    sparms->readRecord = NULL;         /* use default read routine */
    sparms->writeRecord = NULL;        /* use default write routine */
}

#ifdef MAIN_DEMO
void
main(argc, argv)
int  argc;
char *argv[];
{
    fsort_args sparms;
    int status;

    set_sort_file_defaults(&sparms);
    strcpy(sparms.inFileName,"test");  /* File to sort */
    sparms.deleteDuplicateRecords = 1; /* Delete Duplicate Records */
    delete_tmp_files = 1;              /* 0 = Leave tmp files for debugging */
    show_files_records = 1;

    if(status = sort_file(&sparms))   /* do sort */
    {
        fprintf(stderr,"Sort failed.\n");
        fprintf(stderr,"error %d = %s\n",status,sys_errlist[status]);
    }
}
#endif /* MAIN_DEMO */
