#ifndef FSORT_H
#define FSORT_H

#define MAX_STR_LEN     134
#ifndef FILENAME_MAX
#define FILENAME_MAX 512
#endif

#if !defined(FAR)
#ifdef unix
#define FAR
#else /* unix */
/* Haven't gotten far to work yet */
#if 1
#define USE_FAR
#endif

#ifdef USE_FAR
#define FAR  far
#else /* USE_FAR */
#define FAR
#endif /* USE_FAR */
#endif /* unix */
#endif /* defined (FAR) */

#ifdef USE_FAR
#define STRCMP _fstrcmp
#define STRICMP _fstricmp
#define STRNICMP _fstrnicmp
#define MALLOC _fmalloc
#define MEMMOVE _fmemmove
#define FREE _ffree
#else
#define MEMMOVE memmove
#define STRCMP strcmp
#ifdef unix
#define STRICMP strcasecmp
#define STRNICMP mstrincmp /* in decode.c */
#else
#define STRICMP stricmp
#define STRNICMP strnicmp
#endif
#define MALLOC malloc
#define FREE free
#endif

#define MAX_FAR_BUFFERS 10
/* if calling program uses far memory, it can limit how much */
/* fsort tries to allocate by shorting max_far_buffers before */
/* calling its init routines.  It defaults to MAX_FAR_BUFFERS */

extern int max_far_buffers;

typedef struct {
        /* File name to sort */
        char inFileName[FILENAME_MAX];     /* if "" stdin */

        /* Output file name */
        char outFileName[FILENAME_MAX];    /* if "" stdout */

        /* Address of routine to do record compares */
        int (*compareRecord)();                 /* If NULL, strcmp */

        /* Address of routine to read records */
        int (*readRecord)();                    /* If NULL, readLine */

        /* Address of routine to write records */
        int (*writeRecord)();                   /* If NULL, writeLine */

        /* Address of routine to compare records to delete duplicates */
        int (*compareDupRecord)();              /* If NULL, strcmp */

        /* File type.  Binary or text.  Required for fopen */
        int fileType;                           /* 0 = text, 1 = binary */

        /* If fixed length records.  Important NOTICE.  For binaryFileSearch. */
        /* for non fixed length.  Your read routine will be used to line up */
        /* the next record.  A seek will be preformed to the calculated */
        /* middle of the file.  It can be assumed that the seek put us in */
        /* the middle of a record.  One readRecord will be done to find the */
        /* begining of the next record */
        int fixedLength;                        /* 1 = fixed length, 0 = not */

        /* if fixedLength, maxRecSize must show size */
        /* if random, must show max size */
        unsigned int maxRecSize;                /* if 0, MAX_STR_LEN */

        /* if fsort should display errors, or pass errors back to the */
        /* calling routine */
        int PassBackErrorStatus;                /* 1 = return status */
                                                /* 0 = abort */

        /* if deleteDuplicateRecords selected, compareDupRecord is called */
        /* instead of compareRecords.  Allows different compare decision */
        int deleteDuplicateRecords;             /* 1 = delete dup records*/

        /* The following are used internally by msort */
        FILE *fp_in;
        FILE *fp_out;
} fsort_args;

int sort_file(fsort_args *sparms);
extern int debug_binary_sort;
/* Show temp file info as sort is in progress */
extern int show_files_records;
/* set to 0 to leave tmp files, for debugging */
extern int delete_tmp_files;
extern int farBufInUse;
extern int nbrFarBuffersAllocated;
extern char FAR *farBuffers[MAX_FAR_BUFFERS];
extern unsigned int farBufferSizes[MAX_FAR_BUFFERS];
extern char temp_directory[FILENAME_MAX]; /* used by other programs */

int binaryFileSearch(fsort_args *sparms, FILE *fp,char FAR *source,
        char FAR *destBuf);
int binaryMemorySearch(fsort_args *sparms, unsigned int nbrElements,
    char FAR *base, char FAR *last, char FAR *source,char FAR *destBuf);
int init_fsort(fsort_args *sparms);
int close_fsort(fsort_args *sparms);
void set_sort_file_defaults(fsort_args *sparms);
void free_fsort_local_buffers(void);
void heapdump(char *file,int line); /* call with __FILE__, __LINE__ */
void get_tmp_directory(int show_errors);

#endif /* FSORT_H*/

