/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include "doswld.h"
#include "fsort.h"
#ifdef USE_GEDCOM
#include "gedsubs.h"
#include "gedutils.h"
#include "gedhash.h"
#else
#include "pafsubs.h"
#include "pafutils.h"
#endif
#include "printer.h"

static char *revision = "REV 2.01";
/*--------------------------------------------------------------------------
   12/03/93 Working.
   12/03/93 Added getopt
   04/04/95 Lots more.  html, gedcom, etc.
---------------------------------------------------------------------------*/

/* Options to add */
/* Number of generations to print */
/* Chart # of first chart */
/* Number 1 on first chart is the same as number ? on chart ? */
/* char # for continuation charts */
/* Force surnames to upper case */
/* Use formfeeds */
/* Show LDS ordinances */
/* Selection of RIN & MRIN, or ID, or NONE */
/* Print notes */
/* Left margin */

#if 0
#define DEBUG_PROG
#endif

#if USE_GEDCOM
RIN      rin_max;
#else
extern RECORD_PTR rin_max;
#endif

#ifndef sys_errlist
extern char *sys_errlist[];
#endif

#define EOS             '\0'

#define MAX_LINE_SIZE 255

#ifdef NEED_GETOPT
extern getopt(int nargc,char **nargv,char *ostr);
extern char *optarg;
extern int optind;
#endif

extern char *monthnames[];
/* date to go on chart.  Format like : 16 Nov 1993  */
char chart_date[80];

void printnotes(int position,RECORD_PTR rin,FILE *fp_out);
void strnOvr(char *dest, char *src, int max_size);
char *readLine(char *buf,int maxlen,FILE *fp_in);

int read_from_cont_file(CHART_START *cont,int chart_nbr);
void add_to_cont_file(CHART_START *cont);
void erase_cont_file(void);

void init_used_rin_file(void);
int read_from_used_rin_file(USED_RIN *rin,RECORD_PTR person);
int write_to_used_rin_file(USED_RIN *rin,RECORD_PTR person);

int current_chart_nbr = 1; /* chart being printed */

/* command line options */
int start_persons_rin = 1;
int start_persons_spouse_rin = 0;
int chart_nbr_of_first_chart = 1; /* First chart to be printed */
int next_chart_nbr = 2; /* init with first continuation number */
int charts_printed = 0; /* increments with each chart printed */
int max_generations = 32767;  /* Max depth we will follow */
#ifdef HTML
int show_rins = 0;  /* put rin numbers in charts */
int show_mrins = 0; /* put mrin numbers in charts */
int show_lds_ords = 0; /* Put LDS ordinances in charts */
int show_notes = 0;  /* show notes after each chart */
int charts_per_file = 1;
char chart_base_name[80];
char chart_file_name[80];
char chart_ndx_file_name[80];
#else
int show_rins = 1;  /* put rin numbers in charts */
int show_mrins = 1; /* put mrin numbers in charts */
int show_lds_ords = 0; /* Put LDS ordinances in charts */
int show_notes = 1;  /* show notes after each chart */
#endif
int show_names_progress = 1;

int noteLinesPerPage = 55;
int linesPerPage = 55;

int startGeneration = 1;

char *tmp_note_file = "notes.$$$";
char *tmp_index_file = "index.$$$";
FILE *fp_tmpNdx;

#include "pedfrm.c"
char ped_chart[NBR_PED_LINES] [170];

typedef struct 
{
    int line;
    int column;
} 
CHART_LOCATIONS;

/* When we put a person in a chart, the names and places may be truncated. */
/* If so, the full text is put here to be put in the notes. */
typedef struct
{
    char person[80];
    char spouse[80];
    char bplace[80];
    char mplace[80];
    char dplace[80];
}
FULL_NAMES;

#ifdef HTML
CHART_LOCATIONS chart_locs[] =
{
    51,2,
    23,8,
    83,8,
    9,14,
    39,14,
    69,14,
    97,14,
    1,20,
    17,20,
    31,20,
    45,20,
    61,20,
    77,20,
    89,20,
    103,20
  };
#else

CHART_LOCATIONS chart_locs[] = 
{
    31,3,
    19,3,
    47,3,
    12,45,
    26,45,
    40,45,
    54,45,
    8,87,
    16,87,
    22,87,
    30,87,
    36,87,
    44,87,
    50,87,
    58,87,
};
#endif /* HTML */

#ifndef HTML
int gotSubmitter = 0;
char submitter [5][40];
#endif

void
init_ped_chart()
{
    int i;
    for(i = 0; i < NBR_PED_LINES -1 ;i++)
        strcpy(ped_chart[i],ped_chart2[i]);
#ifndef HTML
    if(gotSubmitter)
    {
    }
#endif
}

void
open_ped_index_file()
{
    fp_tmpNdx = fopen (tmp_index_file,"w");
    if(!fp_tmpNdx)
    {
        fprintf(stderr,"Unable to open temporary index file. Aborting\n");
        exit(-1);
    }
}
void
append2Ped_index_file(char *string)
{
    fprintf(fp_tmpNdx,"%s\n",string);
}
void
close_ped_index_file()
{
    fclose(fp_tmpNdx);
}

void
delete_ped_index_file()
{
#ifdef unix
    unlink(tmp_index_file);
#else
    remove(tmp_index_file);
#endif /* DOS */
}
void
print_notes_file(FILE *fp_out)
{
    FILE *fp_notes;
    char s1[MAX_LINE_SIZE];
    int page;
    int line;
    char *chart_header =
"                                  Page %d of notes for chart no. %d\n\n";

    if(show_notes)
    {
        page = 1;
        line = 0;
        fp_notes = fopen (tmp_note_file,"r");
        if(readLine(s1,sizeof(s1),fp_notes) != NULL)
        {
            do
            {

            if(line > noteLinesPerPage)
            {
                line = 0;
                form_feed(fp_out);
                page++;
            }

            if(line == 0)
            {
                fprintf(fp_out,chart_header,page,current_chart_nbr);
                line += 2;
            }

                fprintf(fp_out,"%s\n",delete_trailing_blanks(s1));
                line++;
            }
            while(readLine(s1,sizeof(s1),fp_notes) != NULL);
            if(line != 0)
               form_feed(fp_out);
        }
        fclose(fp_notes);
    }
}

#define MAX_NAME_LEN 36
#define MAX_PLACE_LEN 36

/* returns 1 if successful, 0 if person already used */
int
put_person_in_chart(INDIV_REC *indiv,INDIV_REC *spouse,
	int position,RECORD_PTR rin_nbr)
{
    char s1[100];
    char s2[100];
    char s3[100];
    char s_ndx[256];
    int extra_married_lines;
    USED_RIN rin;
    MARR_REC marriage_record;
    RECORD_PTR spouse2check;
    RECORD_PTR mrin2check;
    int is_male;
    int dataOffset;
    int nameOffset;
    int spouseOffset;
    char indiv_type;
	int status;
	char *this_person = "This person is the same person";
	char *this_person2 = "as no. %d on chart no. %d";

    if(spouse); /* Shut up compile, I will use it eventually*/

#ifdef HTML
    dataOffset = 6;
    nameOffset = 0;
    spouseOffset = 9;
#else
    dataOffset = 10;
    nameOffset = 3;
    spouseOffset = 10;
#endif
	status = 1;
    is_male = (position & 1) || (position == 0);

    if(is_male)
        extra_married_lines = 2;
    else
        extra_married_lines = 0;
#ifdef USE_GEDCOM
#ifdef HTML
    convert_name(s1, &indiv->names,sizeof(s1));
#else
    convert_name(s1, &indiv->names,31);
#endif
    convert_name2(s3, &indiv->names,sizeof(s3));
#else
    convert_name(s1, indiv->names);
    convert_name2(s3, indiv->names);
#endif
    if(show_rins)
    {
#ifdef USE_GEDCOM
      sprintf(s2,"-%ld",rin_nbr);
#else
      sprintf(s2,"-%d",rin_nbr);
#endif
      strcat(s1,s2);
    }
#ifdef HTML
   /* create link in index file to this person */
   sprintf(s_ndx,
   "<A HREF = \"%s_%3.3d.html#p%2.2dn%2.2d\">%-32.32s</A>",
       chart_base_name, current_chart_nbr,1,position + 1,s3);
#ifdef USE_GEDCOM
   sprintf(s2," %5ld",rin_nbr);
#else
   sprintf(s2," %5d",rin_nbr);
#endif
   strcat(s_ndx,s2);
#else /* HTML */
   sprintf(s_ndx,"%-32.32s",s3);
#ifdef USE_GEDCOM
   sprintf(s2," %8ld %4d",rin_nbr,current_chart_nbr);
#else
   sprintf(s2," %8d %4d",rin_nbr,current_chart_nbr);
#endif
   strcat(s_ndx,s2);
#endif /* HTML */
    if(show_names_progress)
    {
    printf("Chart %3.3d Position %2.2d %32.32s\r",
        current_chart_nbr,position + 1,s1);
    }
#ifdef DEBUG_PROG
    printf("Putting %s in chart %d\n", s1,current_chart_nbr);
    printf("Position %d, is_male %d %s passed in rin %d struct rin %d\n",
    position, is_male, s1, rin_nbr,indiv->rin);
#endif /* DEBUG_PROG */
    strnOvr(
    &ped_chart[chart_locs[position].line] 
        [chart_locs[position].column + nameOffset],s1,MAX_NAME_LEN);

    if((position == 0) && (spouse->rin != 0))
    {
#ifdef USE_GEDCOM
#ifdef HTML
    convert_name(s1, &spouse->names,sizeof(s1));
#else
    convert_name(s1, &spouse->names,31);
#endif

#else
    convert_name(s1, spouse->names);
#endif
    if(show_rins)
    {
      sprintf(s2,"-%d",spouse->rin);
      strcat(s1,s2);
#ifdef DEBUG_PROG
    printf("Putting spouse %s in chart %d\n", s1,current_chart_nbr);
#endif /* DEBUG_PROG */
    }
    strnOvr(
    &ped_chart[chart_locs[position].line + 8]
        [chart_locs[position].column + spouseOffset],s1,MAX_NAME_LEN);
    }

	read_from_used_rin_file(&rin,rin_nbr);
    if(rin.used == 0 || position == 0)
	{
	status = 1;
#ifdef USE_GEDCOM
    convert_date(s1,&indiv->birth);
#else
    convert_date(s1,indiv->birthdate);
#endif
    strnOvr(
    &ped_chart[chart_locs[position].line + 1] 
        [chart_locs[position].column + dataOffset],s1,36);
    sprintf(s2," %14s",s1);
    strcat(s_ndx,s2);
#ifdef USE_GEDCOM
#ifdef HTML
    convert_place(s1,&indiv->birth,sizeof(s1));
#else
    convert_place(s1,&indiv->birth,31);
#endif
#else
    convert_place(s1,indiv->birthplace);
#endif
    strnOvr(
    &ped_chart[chart_locs[position].line + 2]
        [chart_locs[position].column + dataOffset],s1,MAX_PLACE_LEN);

#if 1
    if(is_male && (spouse != 0))
    {
    /* Marriage stuff here */
    /* Walk through marriages looking for the spouse passed in to us */
    indiv_type = indiv->sex;
    if(indiv->ownmarriage)
    {
        mrin2check = indiv->ownmarriage;
        do
        {
                if(mrin2check)
                    get_marr_rec(mrin2check,&marriage_record);
                else
                    break;  /* Out of loop. Spouse not found */

                if(indiv_type == 'M')
                    spouse2check = marriage_record.wife;
                else
                    spouse2check = marriage_record.husband;
#ifdef DEBUG_PROG
                printf("comparing spouse2check %d to spouse->rin %d\n",
                spouse2check,spouse->rin);
                printf("marriage_record.wife %d, husband %d",
                marriage_record.wife,marriage_record.husband);
#endif /* DEBUG_PROG */

                if(spouse->rin == spouse2check)
                {

                    /* Found spouse.  Put data in chart */
#ifdef USE_GEDCOM
                    convert_date(s1,&marriage_record.married);
#else
                    convert_date(s1,marriage_record.marriagedate);
#endif

                    strnOvr(
                        &ped_chart[chart_locs[position].line + 3]
                        [chart_locs[position].column + dataOffset],s1,36);
#ifdef USE_GEDCOM
#ifdef HTML
                    convert_place(s1,&marriage_record.married,sizeof(s1));
#else
                    convert_place(s1,&marriage_record.married,31);
#endif
#else
                    convert_place(s1,marriage_record.marriageplace);
#endif
                    strnOvr(
                        &ped_chart[chart_locs[position].line + 4]
                        [chart_locs[position].column + dataOffset],s1,MAX_PLACE_LEN);

                    break;  /* out of while loop, found spouse. */
                }
                /* setup to read in next marriage record to check */
                if(indiv_type == 'M')
                    mrin2check = marriage_record.mnext_wife;
                else
                    mrin2check = marriage_record.mnext_husband;
        }while(1);
    }
    }
#endif
#ifdef USE_GEDCOM
    convert_date(s1,&indiv->death);
#else
    convert_date(s1,indiv->deathdate);
#endif
    strnOvr(
    &ped_chart[chart_locs[position].line + 3 + extra_married_lines]
        [chart_locs[position].column + dataOffset],s1,36);

    sprintf(s2," %14s",s1);
    strcat(s_ndx,s2);
    append2Ped_index_file(s_ndx);
#ifdef USE_GEDCOM
#ifdef HTML
    convert_place(s1,&indiv->death,sizeof(s1));
#else
    convert_place(s1,&indiv->death,31);
#endif
#else
    convert_place(s1,indiv->deathplace);
#endif
    strnOvr(
    &ped_chart[chart_locs[position].line + 4 + extra_married_lines]
        [chart_locs[position].column + dataOffset],s1,MAX_PLACE_LEN);
	}
	else
	{
		status = 0;
#ifdef DEBUG_PROG
		printf("%s already in chart.\n",s1);
#endif /* DEBUG_PROG */
    	strnOvr(
    		&ped_chart[chart_locs[position].line + 2]
            [chart_locs[position].column + dataOffset],this_person,MAX_NAME_LEN);
#ifdef HTML
        sprintf(s2,this_person2,rin.from_position + 1,rin.chart_nbr);
        sprintf(s1,"<A HREF = \"%s_%3.3d.html#p01n%2.2d\">%s</A>",
               chart_base_name, rin.chart_nbr,rin.from_position + 1,
               s2);
#else
		sprintf(s1,this_person2,rin.from_position + 1,rin.chart_nbr);
#endif
    	strnOvr(
    		&ped_chart[chart_locs[position].line + 3]
            [chart_locs[position].column + dataOffset],s1,strlen(s1));
	}

	return status;
}
#if 0
<A HREF = "ped_001.html#p01n01">Martha CORLISS                  ></A>     1  0     2 JUN 1652
#endif

int
cmpNdxStrings(char FAR **s1,char FAR **s2)
{
    int status;
    char FAR * ptr1;
    char FAR * ptr2;

    memcpy(&ptr1,s1,sizeof(char FAR *));
    memcpy(&ptr2,s2,sizeof(char FAR *));
#ifdef HTML
    status = STRNICMP((char FAR *)&ptr1[32],
        (char FAR *)&ptr2[32],100);
    if(status)
        return(status);
#endif
    /* check everything else */
    return(STRICMP((char FAR *)ptr1,(char FAR *)ptr2));
}
/* returns 1 if person found */
/* The search path variable contains a binary number with a high order */
/* start bit.  A binary 0 causes us to follow the father's path, and a */
/* binary 1 causes us to follow the mother's path.  For example a */
/* binary 1 0 1 0 parsing from left to right: Start bit, father, mother, */
/* father. */
/* A start bit alone as the left most bit means only find the start_person*/ 
int
get_one_person(
RECORD_PTR start_person,  /* Start search rin number */
long search_path,         /* search path */
RECORD_PTR *end_person,   /* rin found */
INDIV_REC *person,        /* record for person found */
int ignore_type)          /* 0 no ignores.
                             1 ignore used rin file if start person.
                             2 ignore used rin for all persons. */
{
    long parent_mask;
    int valid;
    int person_nbr;
    RECORD_PTR      current_person;
    INDIV_REC       indiv;
    MARR_REC        parents;
    USED_RIN used_rin;


    parent_mask = 1;

    person->sex = 'D'; /* Assume not found, or deleted */

    while(search_path >= parent_mask )
    {
        parent_mask = parent_mask << 1;
    } 
    parent_mask = parent_mask >> 1;

#ifdef DEBUG_PROG2
    printf("Using start_person 0x%x searching for 0x%lx parrent_mask 0x%lx\n",
    start_person, search_path,parent_mask);
#endif /* DEBUG_PROG2 */
    valid = 1; /* set to zero if we don't find someone */
    if(start_person > rin_max)
        return 0;

    current_person = start_person;
    get_indiv_rec(current_person, &indiv);
    if (indiv.sex == 'D')
    {
        valid = 0;
    }
    else
    {
#ifdef DEBUG_PROG2
        printf("%%%% [%d]", start_person);
        convert_name(s1, indiv.names);
        printf(" %s ", s1);
        printf(" %%%%");
#endif /* DEBUG_PROG2 */
    }
    person_nbr = -1;
    while(valid && (parent_mask > 1))
    {
        person_nbr++;
        parent_mask = parent_mask >> 1;
#if 1
        if((ignore_type == 0) || ((person_nbr != 0) && (ignore_type == 1)))
        {
        read_from_used_rin_file(&used_rin,current_person);
        if(used_rin.used == 1)
        {
            valid = 0;  /* Already in a chart, fake it as not in. */
            break;
        }
        }
#endif
        if(indiv.parentmarriage)
        {
#if 0
            get_paf_rec(indiv.parentmarriage,get_paf_filed(MARR),&parents);
#else
            get_marr_rec(indiv.parentmarriage,&parents);
#endif
            if(search_path & parent_mask)
            {
#ifdef DEBUG_PROG2
                printf("mother ");
#endif /* DEBUG_PROG2 */
                if(parents.wife)
                {
                    current_person = parents.wife;
#if 0
                    get_paf_rec(current_person,get_paf_filed(INDIV),&indiv);
#else
                    get_indiv_rec(current_person,&indiv);
#endif
#ifdef DEBUG_PROG2
                    convert_name(s1, indiv.names);
                    printf(" %s ", s1);
#endif /* DEBUG_PROG2 */
                }
                else
                {
                    valid = 0;
                }
            }
            else
            {
#ifdef DEBUG_PROG2
                printf("father ");
#endif /* DEBUG_PROG2 */
                if(parents.husband)
                {
                    current_person = parents.husband;
                    get_indiv_rec(current_person,&indiv);
#ifdef DEBUG_PROG2
                    convert_name(s1, indiv.names);
                    printf(" %s ", s1);
#endif /* DEBUG_PROG2 */
                }
                else
                {
                    valid = 0;
                }
            }
        }
        else
            valid = 0;
    }
#ifdef DEBUG_PROG2
    printf("\n");
#endif /* DEBUG_PROG2 */
    if(valid)
    {
        memcpy(person,&indiv,sizeof(indiv));
        memcpy (end_person,&current_person,sizeof(current_person));
    }
    else
    {
    	memset(person,0,sizeof(&indiv));
        memset(end_person,0,sizeof(current_person));
    }
    return (valid);
}

int
make_one_chart(FILE *fp_out,int depth,RECORD_PTR start_person,
RECORD_PTR start_spouse)
{
    long search_path;
    FILE *fp_notes;
    RECORD_PTR person_found;
    RECORD_PTR spouse_found;
    RECORD_PTR father_found;
    RECORD_PTR mother_found;
    USED_RIN used_rin;
    struct {
        RECORD_PTR used_rin;
        unsigned char used_position;
        }used_rins_table[16];
    int used_rins_counter;
    CHART_START cont;
	USED_RIN rin;
    long done;
    unsigned char position;
    int i;
    INDIV_REC       indiv;
    INDIV_REC       indiv2;
    INDIV_REC       spouse;
    int is_male;
    int ignore_type;
    char            s1[100];
    char            s2[100];
    char *chart_no = "Chart no. %d";
    char *nbr_1 = "Number 1 on this chart is the same as no. %-2d on chart no. %d";
    char *ordinances[] =
        {
        "Ordinance Codes:",
        "B=Baptized",
        "E=Endowed",
        "P=Sealed to parents",
        "S=Sealed to spouse",
        "C=Children's ordinances",
        NULL
    };

    if(ordinances[0]) ; /* Shut up compiler, I will use it eventually */
    if(start_person == start_spouse) ; /* same thing */
    init_ped_chart();
#ifndef HTML
    strnOvr(&ped_chart[3] [59],chart_date,strlen(chart_date));

    sprintf(s1,chart_no,current_chart_nbr);

    strnOvr(&ped_chart[3] [116],s1,strlen(s1));

    if(show_lds_ords)
    {
        for(i=0;ordinances[i] != NULL;i++)
            strnOvr(&ped_chart[7+i] [3],ordinances[i],strlen(ordinances[i]));
    }
#endif
    if(current_chart_nbr != 1)
    {
        read_from_used_rin_file(&used_rin,start_person);
        if(used_rin.used == 1)
        {
#ifndef HTML
            sprintf(s1,nbr_1,used_rin.from_position + 1,used_rin.chart_nbr);
            strnOvr(&ped_chart[5] [0],s1,strlen(s1));
#endif /* HTML */
        }
    }

    used_rins_counter = 0;
    search_path = 0;
    done = 1;
    fp_notes = fopen(tmp_note_file,"w");
    if(!fp_notes)
    {
        fprintf(stderr,"Unable to open temporary note file. Aborting\n");
        exit(-1);
    }
    for(i=0;i < depth;i++)
        done = (done << 1 ) + 1;
#ifdef DEBUG_PROG
    printf("Making chart %d, rin %d spouse rin %d\n",
    current_chart_nbr,start_person,start_spouse);
    printf("depth = 0x%x done = 0x%x\n",depth, done);
#endif /* DEBUG_PROG */
    position= 0;  position--; /* kind of like -1. Remember UNSIGNED char */
    for(search_path = 1;search_path <= (done >> 1);search_path++)
    {
        position++;
#ifdef DEBUG_PROG2
        printf("search_path = 0x%x\n",search_path);
#endif /* DEBUG_PROG2 */

        is_male = (position & 1) || (position == 0);
#if 0
        if(position == 0) ignore_type = 1; else ignore_type = 0;
#else
        ignore_type = 1;
#endif

        if( get_one_person(start_person,search_path,&person_found,
            &indiv,ignore_type))
        {
#ifdef DEBUG_PROG2
            convert_name(s1, indiv.names);
            printf("Found %s\n", s1);
#endif /* DEBUG_PROG2 */
            /* find spouse */
            if(position == 0)
            {
                get_indiv_rec(start_spouse,&spouse);
            }
            else
            {
            if(indiv.sex == 'M')
            {
                get_one_person(start_person,search_path + 1,
                &spouse_found,&spouse,
                ignore_type);
            }
            else
            {
                get_one_person(start_person,search_path - 1,
                &spouse_found,&spouse,
                ignore_type);
            }
            }
            if(put_person_in_chart(&indiv,&spouse,position,person_found))
			{
                if(show_notes)
                {
                    read_from_used_rin_file(&used_rin,person_found);
                    printnotes(position,person_found,fp_notes);
                }
            /* Check if we got to test for generations in next page */
            if((search_path) > (done >> 2))
            {
                /* check if we can find a father or mother */
#ifdef DEBUG_PROG2
                printf("Check for continue on another page\n");
#endif /* DEBUG_PROG2 */
                if((get_one_person(start_person, (search_path << 1) + 0,
                &father_found,&indiv2,1)) ||
                    (get_one_person(start_person, (search_path << 1) + 1,
                &mother_found,&indiv2,1)))
                {
#ifdef HTML
                /* create the link */
#ifdef USE_GEDCOM
                convert_name(s2,&indiv.names,sizeof(s2));
#else
                convert_name(s2,indiv.names);
#endif
                sprintf(s1,"<A HREF = \"%s_%3.3d.html#p01n01\">%s -></A>",
                    chart_base_name, next_chart_nbr,s2);
                /* overlay the one already there */
                strnOvr(
                    &ped_chart[chart_locs[position].line]
                    [chart_locs[position].column],s1,strlen(s1));
#else
                    sprintf(s1,"%4d",next_chart_nbr);
#endif
#ifdef HTML
#if 0
                    strnOvr(
                    &ped_chart[chart_locs[position].line + 1]
                        [chart_locs[position].column + 31],s1,98);
#endif
#else
                    strnOvr(
                    &ped_chart[chart_locs[position].line + 1]
                        [chart_locs[position].column + 41],s1,98);
#endif
                    cont.indiv = person_found;
                    cont.spouse = spouse_found;
                    cont.chart_nbr = next_chart_nbr;
                    cont.from_position = position + 1;
                    cont.continued_from_chart = current_chart_nbr;
                    cont.generation = startGeneration + 4;
                    if((startGeneration + 4) < max_generations)
                        add_to_cont_file(&cont);
#ifdef DEBUG_PROG2
                    printf("Found a continuation\n");
#endif /* DEBUG_PROG2 */
                    next_chart_nbr++;
                }
            }
			}
            used_rins_table[used_rins_counter].used_rin = person_found;
            used_rins_table[used_rins_counter].used_position = position;
            used_rins_counter++;
        }
    }
    fclose(fp_notes);
    condensed_print(fp_out);
    for(i = 0; i < NBR_PED_LINES -1 ;i++)
        fprintf(fp_out,"%s\n",delete_trailing_blanks(ped_chart[i]));
    standard_print(fp_out);
    form_feed(fp_out);
    print_notes_file(fp_out);
    for(i=0; i < used_rins_counter;i++)
    {
       rin.chart_nbr = current_chart_nbr ;
       rin.from_position = used_rins_table[i].used_position;
       rin.used = 1;
       write_to_used_rin_file(&rin,used_rins_table[i].used_rin);
    }
    charts_printed++;
    return 1;
}

void
show_selected_options()
{
    printf("\"r\" Start rin %d\n",start_persons_rin);
    printf("\"s\" Start spouse rin %d\n",start_persons_spouse_rin);
    printf("\"S\" Starting chart number %d\n",chart_nbr_of_first_chart);
    printf("\"C\" Contining chart number %d\n",next_chart_nbr);
    printf("\"m\" Maximum generations to follow %d\n",max_generations);
#ifndef HTML
    printf("\"R\" Show rin numbers %d\n",show_rins);
    printf("\"M\" Show mrin numbers %d\n",show_mrins);
#endif
#if 0
    printf("\"L\" Show LDS ordinances %d\n",show_lds_ords);
#endif
#ifndef USE_GEDCOM
    printf("\"n\" Show notes %d\n",show_notes);
    printf(" \"d\" Deletes index files created for the gedcom file\n");
#endif
#ifdef HTML
    printf("\"f\" html file names start with %s\n",chart_base_name);
#endif
#ifndef HTML
    printf("\"p\" Select printer type %d\n",printerType);
    show_printer_type();
#endif
}

void
usage(char *progname)
{
#ifdef USE_GEDCOM
    printf("%s -[frsSCmRMn]<value> gedcomFileName\n",progname);
#else
#ifdef HTML
    printf("%s -[frsSCmn]<value>\n",progname);
#else
    printf("%s -[rsSCmRMn]<value>\n",progname);
#endif
#endif /* USE_GEDCOM */
    fprintf(stderr,"%s\n",revision);
    printf("Generate cascading pedigree charts from PAF data.\n");
    printf("Charts are identical to PAF generated charts with\n");
    printf("the addition of all notes displayed on following pages.\n");
    printf("Selectable options are:\n");
    show_selected_options();
}
void
append_file(char *fileName,FILE *fp_out)
{
    FILE *fp_in;
    char s1[255];
        fp_in = fopen(fileName,"r");
        if(!fp_in)
        {
            fprintf(stderr,"Unable to open input file %s. Aborting\n",
                fileName);
            exit(-1);
        }
        while(readLine(s1,sizeof(s1),fp_in) != NULL)
            fprintf(fp_out,"%s\n",s1);
        fclose(fp_in);
}

int
ndx_header_lines(FILE *fp_out)
{
fprintf(fp_out,
"                        INDEX OF NAMES ON PEDIGREE CHART\n");

fprintf(fp_out,
"================================================================================\n");
fprintf(fp_out,
"Name                                  RIN Chrt           Born           Died\n");
fprintf(fp_out,
"--------------------------------------------------------------------------------\n");
return 4;
}

void
main(int argc,char  *argv[])
{
#ifndef USE_GEDCOM
    NAMADD_REC      namadd;
    PAF_FILE       *indiv_file;
#endif
    FILE *fp_out;
#ifndef HTML
    FILE *fp_in;
    int lineCount;
	char readBuf[256];
#endif
    CHART_START cont;
    int parmv;
    fsort_args sparms;
    int status;
#ifdef USE_GEDCOM
    int arg;
    char *source;
    int deleteIndexFiles;
#endif
    char *getopt_str;
    int c;
    char *output_file = "pedigree.rpt";

#ifdef USE_GEDCOM
    disableRamFile = 1;
    deleteIndexFiles = 0;
    dosWildCard(&argc,&argv);  /* DOS wild card expander. Just returns if unix */
#endif

#ifdef HTML
    allowPrintFormatting = 0;
    strcpy(chart_base_name,"ped");
    getopt_str = "f:r:s:n:S:C:R:M:L:m:p:d";
#else
    getopt_str = "r:s:n:S:C:R:M:L:m:p:d";
#endif

    while ((c = getopt(argc, argv, getopt_str)) != EOF)
        switch(c)
        {
                /* start rin value */
                case 'r':
                     parmv = atoi(optarg);
                     start_persons_rin = parmv;
                break;
                /* start spouse rin */
                case 's':
                     parmv = atoi(optarg);
                     start_persons_spouse_rin = parmv;
                break;
                case 'n':
                     parmv = atoi(optarg);
                     show_notes = parmv;
                break;
                case 'S':
                     parmv = atoi(optarg);
                     chart_nbr_of_first_chart = parmv;
                break;
                case 'C':
                     parmv = atoi(optarg);
                     next_chart_nbr = parmv;
                break;
                case 'R':
                     parmv = atoi(optarg);
                     show_rins = parmv;
                break;
                case 'M':
                     parmv = atoi(optarg);
                     show_mrins = parmv;
                break;
                case 'L':
                     parmv = atoi(optarg);
                     show_lds_ords = parmv;
                break;
                case 'm':
                     parmv = atoi(optarg);
                     max_generations = parmv;
                break;
                case 'p':
                     parmv = atoi(optarg);
                     if(parmv > MAX_PRINTER_TYPE)
                     {
                        printf("Illegal printer type\n");
                        show_allowed_printer_types();
                        exit(-1);
                     }
                     printerType = parmv;
                break;
#ifdef HTML
                case 'f':
                     strcpy(chart_base_name,optarg);
                break;
#endif
                case 'd':
#ifdef USE_GEDCOM
                deleteIndexFiles = 1;
                break;
#endif
                default:
                     usage(argv[0]);
                     exit(-1);
                     break;
        }

#ifdef USE_GEDCOM
    if((optind + 1) > argc)
    {
        usage(argv[0]);
        exit(-1);
    }
#endif

    printf("Generating chart with options:\n");
    show_selected_options();

#ifdef HTML
    sprintf(chart_file_name,"%s_%3.3d.html",chart_base_name,
        chart_nbr_of_first_chart);
    output_file = chart_file_name;
#endif

#ifdef USE_GEDCOM
    showHashDetails = 0;
    alloc_hash_buffers();
    for(arg = optind;arg < argc;arg++)
    {
    source = argv[arg];
    printf("Indexing %s\n",source);
    hash_gedcom_file(source);
    init_use_hash(source);
    rin_max = max_rin; /* init my stuff from hash_gedcom_file */
#endif

    fp_out = fopen(output_file,"w");
    if(!fp_out)
    {
        fprintf(stderr,"Unable to open output file %s. Aborting\n",output_file);
        exit(-1);
    }

#ifdef HTML
    append_file("ped_hdr.txt",fp_out);
#endif

#ifndef USE_GEDCOM
    paf_open_name('r');
    paf_open_indiv('r');
    paf_open_marr('r');
    paf_open_notes('r');
    paf_open_namadd('r');
    indiv_file = get_paf_filed(INDIV);
    rin_max = indiv_file->nrec;
    get_namadd_rec((RECORD_PTR) 1, &namadd);
#endif /* USE_GEDCOM */

    erase_cont_file();
    init_used_rin_file();
    strcpy(chart_date,get_cur_date());
#ifdef DEBUG_PROG
    printf("%% %s\n", namadd.name);
    printf("%% %s\n", namadd.addr1);
    printf("%% %s\n", namadd.addr2);
    printf("%% %s\n", namadd.addr3);
    printf("%% %s\n", namadd.phone);
    printf("%%\n");
#endif
    current_chart_nbr = chart_nbr_of_first_chart;
    startGeneration = 1;
    if(!show_names_progress)
        fprintf(stderr,"\nMaking chart %d",charts_printed + 1);
    if(show_names_progress)
      printf("\n");
    open_ped_index_file();
    make_one_chart(fp_out,4,start_persons_rin,start_persons_spouse_rin);
    while(read_from_cont_file(&cont,charts_printed -1))
    {

        current_chart_nbr = cont.chart_nbr;
        startGeneration =  cont.generation;
        if(!show_names_progress)
            fprintf(stderr,"\nMaking chart %d for rin %d",
        charts_printed + 1,cont.indiv);
#ifdef HTML
        append_file("ped_tail.txt",fp_out);
        fclose(fp_out);
        sprintf(chart_file_name,"%s_%3.3d.html",chart_base_name,
            current_chart_nbr);
        output_file = chart_file_name;
        fp_out = fopen(output_file,"w");
        if(!fp_out)
        {
            fprintf(stderr,"Unable to open output file %s. Aborting\n",
                output_file);
            exit(-1);
        }
        append_file("ped_hdr.txt",fp_out);
#endif
        make_one_chart(fp_out,4,cont.indiv,cont.spouse);
    }
    if(show_names_progress)
      printf("\n");
#ifdef HTML
    append_file("ped_tail.txt",fp_out);
#endif
    fclose(fp_out);
    delete_cont_file();
    erase_used_rin_file();
#ifdef unix
    unlink(tmp_note_file);
#else
    remove(tmp_note_file);
#endif /* DOS */
    close_ped_index_file();
#ifdef USE_GEDCOM
    close_use_hash(source,deleteIndexFiles);
    }
    free_hash_buffers();
    deleteOverFlowFile();
#endif
    printf("Sorting index file\n");
    set_sort_file_defaults(&sparms);
    strcpy(sparms.inFileName,tmp_index_file);  /* File to sort */
    sprintf(sparms.outFileName,"temp2.$$$");
    sparms.deleteDuplicateRecords = 1; /* Delete Duplicate Records */
    sparms.compareRecord = cmpNdxStrings;
    delete_tmp_files = 1;              /* 0 = Leave tmp files for debugging */

    if(status = sort_file(&sparms))   /* do sort */
    {
        fprintf(stderr,"Sort failed.\n");
        fprintf(stderr,"error %d = %s\n",status,sys_errlist[status]);
    }
    delete_ped_index_file();
#ifndef HTML
    fp_out = fopen("pedigree.ndx","w");
    lineCount = 0;
    fp_in = fopen(sparms.outFileName,"r");
	lineCount += ndx_header_lines(fp_out);
    while(readLine(readBuf,sizeof(readBuf),fp_in))
    {
        lineCount++;
        if(lineCount >= linesPerPage)
        {
            form_feed(fp_out);
            lineCount = 0;
			lineCount += ndx_header_lines(fp_out);
        }
        fprintf(fp_out,"%s\n",readBuf);
    }
    form_feed(fp_out);
	fclose(fp_out);
	fclose(fp_in);
    delete_file(sparms.outFileName);
#else
    sprintf(chart_ndx_file_name,"%s_ndx.html",chart_base_name);
    fp_out = fopen(chart_ndx_file_name,"w");
    if(!fp_out)
    {
        fprintf(stderr,"Unable to open output file %s. Aborting\n",
            chart_ndx_file_name);
        exit(-1);
    }
    append_file("ndx_hdr.txt",fp_out);
    append_file("temp2.$$$",fp_out);
    append_file("ndx_tail.txt",fp_out);
    fclose(fp_out);
    delete_file("temp2.$$$");
#endif

}

