/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#ifdef SUNOS4_1_2
#include <unistd.h>
#endif
#include "doswld.h"
#include "fsort.h"
#include "gedrtns.h"
#include "gedindex.h"
#include "gedhash.h"
#include "ramfile.h"

#define STDERROR stdout

#ifdef unix
#define FOPEN_WB "w"
#define FOPEN_WBP "w+"
#define FOPEN_RB "r"
#else
#define FOPEN_WB "wb"
#define FOPEN_WBP "wb+"
#define FOPEN_RB "rb"
#endif

#ifndef sys_errlist
extern char *sys_errlist[];
#endif

RIN max_rin;
RIN max_mrin;
hash_loc hashTagRecord;
hash_rin rinLocRecord;
hash_rin mRinLocRecord;
char hashedGedcomFileName[FILENAME_MAX];
FILE *fp_ReadHashedGedcomFile;
FILE *fp_HashedRinFile;
FILE *fp_HashedMRinFile;
int ignore_ndx_dates = 0;
int always_make_ndx = 0;
/* A hash function */

/* 
   Set this to the something several times larger than the maximum # of 
   lines in a history file.  It should be a prime number.
*/
#define INDEX_SIZE 99991 

#ifndef max
#define	max(a, b) 		((a) < (b) ? (b) : (a))
#endif
#ifndef min
#define	min(a, b) 		((a) > (b) ? (b) : (a))
#endif
#ifndef abs
#define	abs(x)			((x) >= 0 ? (x) : -(x))
#endif

/* This is a simplified version of the pathalias hashing function.  */
/* Thanks to Steve Belovin and Peter Honeyman */

/*
 * fold a string into a long int.  31 bit crc (from andrew appel).
 * the crc table is computed at run time by crcinit() -- we could
 * precompute, but it takes 1 clock tick on a 750.
 *
 * This fast table calculation works only if POLY is a prime polynomial
 * in the field of integers modulo 2.  Since the coefficients of a
 * 32-bit polynomial won't fit in a 32-bit word, the high-order bit is
 * implicit.  IT MUST ALSO BE THE CASE that the coefficients of orders
 * 31 down to 25 are zero.  Happily, we have candidates, from
 * E. J.  Watson, "Primitive Polynomials (Mod 2)", Math. Comp. 16 (1962):
 *	x^32 + x^7 + x^5 + x^3 + x^2 + x^1 + x^0
 *	x^31 + x^3 + x^0
 *
 * We reverse the bits to get:
 *	111101010000000000000000000000001 but drop the last 1
 *         f   5   0   0   0   0   0   0
 *	010010000000000000000000000000001 ditto, for 31-bit crc
 *	   4   8   0   0   0   0   0   0
 */

#define POLY32 0xf5000000	/* 32-bit polynomial */
#define POLY31 0x48000000	/* 31-bit polynomial */
#define POLY POLY31	/* use 31-bit to avoid sign problems */

static long CrcTable[128];

static void crcinit()
{	register int i,j;
	register long sum;

	for (i = 0; i < 128; i++) {
		sum = 0;
		for (j = 7-1; j >= 0; --j)
			if (i & (1 << j))
				sum ^= POLY >> j;
		CrcTable[i] = sum;
	}
}

static long
fold(s,size)
	register char *s;
	int size;
{	register long sum = 0;

	while (size--) {
		sum = (sum >> 7) ^ CrcTable[(sum ^ (*s++)) & 0x7f];
	}

	return sum;
}

#define HASH1(n) ((n) % INDEX_SIZE);

long
hash(name,size)
	char *name;
	int  size;
{
	register long probe;

	static int init = 1;

	if (init) {
           crcinit();
	   init = 0;
	}
        
	probe = fold(name,size);
	probe = HASH1(probe);

	return probe;
}

int showHashDetails = 1;
#define SPECIAL_SPOT 0
#if SPECIAL_SPOT
char *special = "I60";
#endif

int getstamp(char *path,struct stat *sbuf)   /* get a file's date/time stamp */
/* char *path; file to get stamp from */
/* struct stat *sbuf;  storage for data */
{
    if(stat(path, sbuf)) 
    {
        /* must not exist or some problem */
        return(-1);
    }
    return 0;
}

void
make_ndx_name(char *gedcom_file,char *buffer)
{
        char name_base[FILENAME_MAX];

        cleanup_file_name(gedcom_file,name_base,1);
        strcat(name_base,".gdi");
        strcpy(buffer,name_base);
}

void
make_rinFile_name(char *gedcom_file,char *buffer)
{
        char name_base[FILENAME_MAX];

        cleanup_file_name(gedcom_file,name_base,1);
        strcat(name_base,".gdr");
        strcpy(buffer,name_base);
}
void
make_marriageRinFile_name(char *gedcom_file,char *buffer)
{
        char name_base[FILENAME_MAX];

        cleanup_file_name(gedcom_file,name_base,1);
        strcat(name_base,".gdm");
        strcpy(buffer,name_base);
}

/* with a MAX_GED_HASH of 16k, DOS has problems if we keep this all in */
/* one table.  It would be 64k in size, and the far malloc would often */
/* fail.  Using two tables breakes it up into to 32k requests. */
static unsigned long FAR *hashPointers1;
static unsigned long FAR *hashPointers2;
static FILE *fp_index;
static char index_file[FILENAME_MAX];
static char rin_file[FILENAME_MAX];
static char mRin_file[FILENAME_MAX];

static fsort_args sparms;


int
alloc_hash_buffers()
{
    size_t size;
    int i;
    char FAR *sanity;

    size = ((MAX_GED_HASH/2) + 4) * sizeof(long);

    hashPointers1 = (unsigned long  FAR *)MALLOC(size);
    hashPointers2 = (unsigned long  FAR *)MALLOC(size);
    if((hashPointers1 == NULL) || (hashPointers2 == NULL))
    {
            fprintf(STDERROR,"Unable to allocate hashPointers\n;");
#if 1
                heapdump(__FILE__,__LINE__);
#endif
            exit(-1);
    }
    sanity = (char FAR *)hashPointers1;
    sanity[size -1] = 0x55;
    sanity = (char FAR *)hashPointers2;
    sanity[size -1] = 0x55;
    sparms.maxRecSize = sizeof(struct HASH_LOC);            /* record size */
/*    max_far_buffers = MAX_FAR_BUFFERS - 4;  */
    set_sort_file_defaults(&sparms);
    init_fsort(&sparms);
    free_fsort_local_buffers();
#if 1
    nbrRamFarBuffers = 2;
    ramFarBuffers[0]= (char FAR *)hashPointers1;
    ramFarBuffers[1]= (char FAR *)hashPointers2;
    ramFarBufferSizes[0] = MAX_GED_HASH/2 * sizeof(long);
    ramFarBufferSizes[1] = MAX_GED_HASH/2 * sizeof(long);
    /* ones allocated by fsort */
    for(i=0;i < nbrFarBuffersAllocated;i++)
    {
        ramFarBuffers[i + 2] = farBuffers[i];
        ramFarBufferSizes[i + 2] = farBufferSizes[i];
        nbrRamFarBuffers++;
    }
#endif
    return(1);
}

long
getHashPointer(long index)
{
    long val;

    if(index >= (MAX_GED_HASH / 2))
    val = hashPointers2[index % (MAX_GED_HASH / 2)];
    else
    val = hashPointers1[index];
    return(val);
}

void
setHashPointer(long index, long val)
{
    if(index >= (MAX_GED_HASH / 2))
        hashPointers2[index % (MAX_GED_HASH / 2)] = val;
    else
        hashPointers1[index] = val;
}

int
init_use_hash(char *gedcom_file)
{
        long Index;
        struct stat sbuf;
        long fileSize;
        int i;
        make_ndx_name(gedcom_file,index_file);
        registerNameWithRamfile(index_file);
        fp_index = fopen(index_file,FOPEN_RB);
        if(fp_index == NULL)
        {
            fprintf(STDERROR,"Unable to open %s\n to read",index_file);
            exit(-1);
        }
        if(stat(index_file, &sbuf) == 0)
        {
            fileSize = sbuf.st_size;
        }
        else
        {
            fprintf(STDERROR,"Unable to stat %s\n to read",index_file);
            exit(-1);
        }
        fp_HashedRinFile = fopen(rin_file,FOPEN_RB);
        if(fp_HashedRinFile == NULL)
        {
            fprintf(STDERROR,"Unable to open %s to read\n",rin_file);
            exit(-1);
        }
        if(stat(rin_file, &sbuf) == 0)
        {
            fileSize = sbuf.st_size;
            max_rin = (int)(fileSize / (long)sizeof(rinLocRecord));
        }
        else
        {
            fprintf(STDERROR,"Unable to stat %s\n to read",rin_file);
            exit(-1);
        }
        fp_HashedMRinFile = fopen(mRin_file,FOPEN_RB);
        if(fp_HashedRinFile == NULL)
        {
            fprintf(STDERROR,"Unable to open %s to read\n",mRin_file);
            exit(-1);
        }
        if(stat(mRin_file, &sbuf) == 0)
        {
            fileSize = sbuf.st_size;
            max_mrin = (int)(fileSize / (long)sizeof(rinLocRecord));
        }
        else
        {
            fprintf(STDERROR,"Unable to stat %s\n to read",mRin_file);
            exit(-1);
        }
        fp_ReadHashedGedcomFile = fopen(gedcom_file,"r");
        if(fp_ReadHashedGedcomFile == NULL)
        {
            fprintf(STDERROR,"Unable to open %s to read\n",gedcom_file);
            exit(-1);
        }

        for(i=0;i < MAX_GED_HASH;i++)
        {
          if(!fread(&Index,sizeof(Index),1,fp_index))
          {
            fprintf(STDERROR,"Error reading index file %s\n",index_file);
            exit(0);
          }
          setHashPointer(i,Index);
        }
        return(1);
}

/* returns 0 if ok, -1 if couldn't delete */
int deletefile(char *name)
{
    int status;

#ifndef DOS
    status = unlink(name);
#else
    status = remove(name);
#endif /* DOS */
    return status;
}

void
free_hash_buffers()
{
        size_t size;
        char FAR *sanity;

        size = ((MAX_GED_HASH/2) + 4) * sizeof(long);
        sanity = (char FAR *)hashPointers1;
        if(sanity[size -1] != 0x55)
        {
            fprintf(stderr,"hashPointers1 corrupted\n");
            exit(-1);
        }
        sanity = (char FAR *)hashPointers2;
        if(sanity[size -1] != 0x55)
        {
            fprintf(stderr,"hashPointers2 corrupted\n");
            exit(-1);
        }
        FREE(hashPointers1);
        FREE(hashPointers2);
        close_fsort(&sparms);
}

void
close_use_hash(char *gedcom_file,int deleteHashFile)
{
        fclose(fp_index);
        fclose(fp_HashedRinFile);
        fclose(fp_HashedMRinFile);
        fclose(fp_ReadHashedGedcomFile);
        if(deleteHashFile)
        {
            deletefile(index_file);
            deletefile(rin_file);
            deletefile(mRin_file);
        }
}

/*
Pass in rin number which is actually the order in which an INDI is
encountered in the gedcom file.  Returns 1 for success, and location
returned in location.
Also, global runLocRecord contains successful search results.
*/
int
find_rin_entry(RIN rin, long *location)
{
    if(rin > max_rin)
    {
        return(0); /* error, not legal rin, not found */
    }

    fseek(fp_HashedRinFile,(long)((long)(rin - 1) * (long)sizeof(rinLocRecord)),
        SEEK_SET);
    if(!fread(&rinLocRecord,sizeof(rinLocRecord),1,fp_HashedRinFile))
    {
          fprintf(STDERROR,
               "Unable to read to rin file %s\n",rin_file);
          return(0);
    }
    *location = rinLocRecord.loc;
    return(1);
}

/* same as find_rin_entry but it finds a Married rin number */
int
find_mRin_entry(RIN mrin, long *location)
{
    if(mrin > max_mrin)
    {
        return(0); /* error, not legal mrin, not found */
    }

    fseek(fp_HashedMRinFile,(long)((long)(mrin - 1) * (long)sizeof(rinLocRecord)),
        SEEK_SET);
    if(!fread(&rinLocRecord,sizeof(rinLocRecord),1,fp_HashedMRinFile))
    {
          fprintf(STDERROR,
               "Unable to read to mrin file %s\n",mRin_file);
          return(0);
    }
    *location = rinLocRecord.loc;
    return(1);
}

/*
Pass in gedcom file @nbr@ value.  Returns 1 for success, and location
returned in location.
Also, global hashTagRecord contains successful search results.
*/
int
find_hash_entry(char *string,long *location)
{
    long results;
    unsigned long val;
    long loc;
#if SPECIAL_SPOT
     if(!strcmp(string,special))
     {
        fprintf(STDERROR,"Looking for %s\n",special);
              val++;
              val--;
     }
#endif
    results = (hash(string,strlen(string)) % MAX_GED_HASH);

    /* seek to first tag pointed to by hashPointers */
    val = getHashPointer(results) -1;

    loc = (long)((long)(MAX_GED_HASH * (long)sizeof(long)) +
               ((long)(val) * (long)sizeof(hashTagRecord)));
    fseek(fp_index,(long)(loc),SEEK_SET);
    if(!fread(&hashTagRecord,sizeof(hashTagRecord),1,fp_index))
    {
        fprintf(STDERROR,"Unable to read hashTagRecord record from %s\n",index_file);
        exit(0);
    }
    if(!strcmp(string,hashTagRecord.name))
    {
        *location = hashTagRecord.loc;
        return(1);
    }
    while(1)
    {
        if(!hashTagRecord.nextTag)
        {
            return(0); /* not found */
        }
        else
        {
            fseek(fp_index,(long)(hashTagRecord.nextTag),SEEK_SET);
            if(!fread(&hashTagRecord,sizeof(hashTagRecord),1,fp_index))
            {
                fprintf(STDERROR,"Unable to read hashTagRecord record from %s\n",index_file);
                exit(0);
            }
            if(!strcmp(string,hashTagRecord.name))
            {
                *location = hashTagRecord.loc;
                return(1);
            }
        }
    }
}

/* Call with the tag string, and it returns the rin number */

RIN
indi2rin(char *indi)
{
    long loc;
    if(!find_hash_entry(indi,&loc))
        return(0);
    return(hashTagRecord.rin);

}

/* Generates three files with the same name as the gedcom file, with*/
/* extensions.  gdi, gdm and gdr.  The gedcom.gdi is a hash table for */
/* all tags in the gedcom file, of all types.  The gdm is a table of */
/* marriage tags, and the gdr is for rins (INDI). */

void
hash_gedcom_file(char *gedcom_file)
{
    FILE *fp_gedcom;
    char readBuf[256];
    unsigned int i;
	char *cptr1;
	char *cptr2;
    unsigned long nextIndex;
    unsigned long nextIndex2;
    hash_rin rin_loc;
    hash_loc hashTagRecordTmp;
    struct stat ndate;
    struct stat idate;
    long val;
    long tmp;
    long tmp2;
    long location;
    long results;
    long results2;
    long endOfFile;
    long collisions;
    long depth;
    long maxDepth;
    long iAmHere;
    int makeit;
    unsigned int rin;
    unsigned int mrin;

    depth = collisions = maxDepth = 0;
    rin = mrin = 0;
    max_rin = 0;
    max_mrin = 0;

    strcpy(hashedGedcomFileName,gedcom_file);
        make_ndx_name(gedcom_file,index_file);
        make_rinFile_name(gedcom_file,rin_file);
        make_marriageRinFile_name(gedcom_file,mRin_file);
#if 1
        /* Code to check date on index files */
        /* no need to build them if they are newer than gedcom file */
        getstamp(gedcom_file,&ndate);  /* How old is ndx file */

        makeit = 0;
        if(!always_make_ndx)
        {
        /* if the ndx file is older than the ndi file, recreate ndi */
        /* by not executing the read and return */
        if(!getstamp(index_file,&idate)) /* How old is ndi file */
        {
            if((ndate.st_mtime > idate.st_mtime) || (ignore_ndx_dates))
            {
                makeit = 1;
            }
        }
        else
            makeit = 1;
        if(!getstamp(rin_file,&idate)) /* How old is ndi file */
        {
            if((ndate.st_mtime > idate.st_mtime) || (ignore_ndx_dates))
            {
                makeit = 1;
            }
        }
        else
            makeit = 1;
        if(!getstamp(mRin_file,&idate)) /* How old is ndi file */
        {
            if((ndate.st_mtime > idate.st_mtime) || (ignore_ndx_dates))
            {
                makeit = 1;
            }
        }
        else
            makeit = 1;
        }
        else
           makeit = 1;
        if(makeit == 0)
            return;
#endif
        for(i=0;i < MAX_GED_HASH;i++)
        {
            setHashPointer(i,0);
        }
        fp_gedcom = fopen(gedcom_file,"r");
        if(fp_gedcom == NULL)
        {
            fprintf(STDERROR,"Unable to open %s\n",gedcom_file);
            exit(-1);
        }
        registerNameWithRamfile(index_file);
        fp_index = fopen(index_file,FOPEN_WBP);
        if(fp_index == NULL)
        {
            fprintf(STDERROR,"Unable to open %s for writting\n",index_file);
            exit(-1);
        }
        fp_HashedRinFile = fopen(rin_file,FOPEN_WB);
        if(fp_HashedRinFile == NULL)
        {
            fprintf(STDERROR,"Unable to open %s for writting\n",rin_file);
            exit(-1);
        }
        fp_HashedMRinFile = fopen(mRin_file,FOPEN_WB);
        if(fp_HashedMRinFile == NULL)
        {
            fprintf(STDERROR,"Unable to open %s for writting\n",mRin_file);
            exit(-1);
        }

        if(showHashDetails)
        {
            fprintf(STDERROR,"Indexing %s\n",gedcom_file);
        }
        nextIndex = 0;
        for(i=0;i < MAX_GED_HASH;i++)
        {
          if(!fwrite(&nextIndex,sizeof(nextIndex),1,fp_index))
          {
            fprintf(STDERROR,"Unable to write to index file %s\n",index_file);
          }
        }
        endOfFile = ftell(fp_index);
        nextIndex = 1;
        fseek(fp_gedcom,0L,SEEK_SET);
        memset(&hashTagRecord,0,sizeof(hashTagRecord));
        while(find_next_level_zero(fp_gedcom, &readBuf[0]))
        {
            /* Not necessary, just looks better with hex dump */
            /* of index file */
            memset(hashTagRecord.name,0,sizeof(hashTagRecord.name));
            hashTagRecord.rin = 0;
            cptr1 = strchr(readBuf,'@');
            if(cptr1 == NULL)
            {
                if(strstr(readBuf,"TRLR") == NULL)
                    continue; /* not a valid level zero, try again */
                else
                    break;  /* End of gedcom file data, quit while */
            }
      		cptr1++;
    		cptr2 = strchr(cptr1,'@');
            *cptr2 = 0;
            cptr2++;
            if(strstr(cptr2,"INDI") != NULL)
            {
                /* Put it in rin file */
					 memset(&rin_loc,0,sizeof(rin_loc));
                rin++;
                hashTagRecord.rin = rin;
                max_rin++;
                strcpy(rin_loc.name,cptr1);
                rin_loc.loc = cur_ged_line_start;
                if(!fwrite(&rin_loc,sizeof(rin_loc),1,fp_HashedRinFile))
                {
                    fprintf(STDERROR,
                        "Unable to write to rin file %s\n",rin_file);
                }
            }

            if(strstr(cptr2,"FAM") != NULL)
            {
                /* Put it in Mrin file */
					 memset(&rin_loc,0,sizeof(rin_loc));
                mrin++;
                hashTagRecord.rin = mrin;
                max_mrin++;
                strcpy(rin_loc.name,cptr1);
                rin_loc.loc = cur_ged_line_start;
                if(!fwrite(&rin_loc,sizeof(rin_loc),1,fp_HashedMRinFile))
                {
                    fprintf(STDERROR,
                        "Unable to write to rin file %s\n",mRin_file);
                }
            }

            /* debug stuff */
#if SPECIAL_SPOT
            if(!strcmp(cptr1,special))
            {
                fprintf(STDERROR,"Placing %s in hash table\n",special);
                fprintf(STDERROR,"%s\n",readBuf);
                fprintf(STDERROR,"Line found at 0x%lx\n",cur_ged_line_start);
              i++;
              i--;
              iAmHere = ftell(fp_index); /* debug */
            }
#endif
            depth = 0;
            results = (hash(cptr1,strlen(cptr1)) % MAX_GED_HASH);

            /* Have we ever written an entry under this hash value? */
            if(getHashPointer(results) == 0)
            {
                /* first time this hash encountered.  set it up, and */
                /* write hashTagRecord pointer */
                setHashPointer(results,nextIndex);
                strcpy(hashTagRecord.name,cptr1);
                hashTagRecord.loc = cur_ged_line_start;
                hashTagRecord.nextTag = 0;
#if PAD_HASH_LOC
                hashTagRecord.pad = 0;
#endif
                fseek(fp_index,0L,SEEK_END);
#if SPECIAL_SPOT
            if(!strcmp(cptr1,special))
            {
              i++;
              i--;
              iAmHere = ftell(fp_index); /* debug */
              fprintf(STDERROR,"special %s goes at 0x%lx\n",special,iAmHere);
              fprintf(STDERROR,"Results %ld\n", results);
              fprintf(STDERROR,"hashPointers[results] %ld\n",
                            getHashPointer(results));
              val = getHashPointer(results) -1;
              location = (long)((long)(MAX_GED_HASH * (long)sizeof(nextIndex)) +
                    ((long)(val) * (long)sizeof(hashTagRecord)));
              fprintf(STDERROR,"Calculated loc = 0x%lx\n",location);
              fprintf(STDERROR,"hashTagRecord.loc = 0x%lx\n",hashTagRecord.loc);
            }
#endif
                if(!fwrite(&hashTagRecord,sizeof(hashTagRecord),1,fp_index))
                {
                    fprintf(STDERROR,"Unable to write to index file %s\n",
                        index_file);
                }
                fseek(fp_index,0L,SEEK_END); /* get location of eof */
                endOfFile = ftell(fp_index);
                nextIndex++;
            }
            else
            {
                /* used this hash before.  Find last hashTagRecord under this */
                /* hash, update its pointer to this new value */

                collisions++;
                /* seek to first hashTagRecord pointed to by hashPointers */
                val = getHashPointer(results) -1;
                tmp = (long)(MAX_GED_HASH * (long)sizeof(long));
                tmp2 = (val) * (long)sizeof(hashTagRecord);
                location = (long)((long)(MAX_GED_HASH * (long)sizeof(long)) +
                    ((long)(val) * (long)sizeof(hashTagRecord)));
                if(fseek(fp_index,(long)(location),SEEK_SET))
                {
                    fprintf(STDERROR,"Error seeking to %ld\n",location);
                    fprintf(STDERROR,"Error = %d  %s\n",
                    errno,sys_errlist[errno]);
                    exit(0);
                }

                while(1)
                {
                    depth++;
                    maxDepth = max(depth,maxDepth);
                    if(!fread(&hashTagRecordTmp,sizeof(hashTagRecord),1,
                        fp_index))
                    {
                        fprintf(STDERROR,"Error reading index file %s\n",
                            index_file);
                        exit(0);
                    }

                    if(!strcmp(hashTagRecordTmp.name,cptr1))
                    {
                        fprintf(STDERROR,
                            "Error, %s has two tags of the same name.\n",
                            gedcom_file);
                        fprintf(STDERROR,
                            "%s found twice.  Ignoring second one.\n",cptr1);
                        break; /* out of while */
                    }

                    results2 = (hash(hashTagRecordTmp.name,
                        strlen(hashTagRecordTmp.name)) % MAX_GED_HASH);
                    if(results2 != results)
                    {
                        fprintf(STDERROR, "Internal error, miss possitioned\n");
                        iAmHere = ftell(fp_index); /* debug */
                        fprintf(STDERROR,"Actual loc 0x%lx, calculated loc 0x%lx\n",
                        iAmHere,location);
                        fprintf(STDERROR,"nextIndex %d\n",nextIndex);
                        fprintf(STDERROR,"cptr1 %s\n",cptr1);
                        fprintf(STDERROR,"hashTagRecordTmp.name %s\n",
                            hashTagRecordTmp.name);
                        fprintf(STDERROR,"Results %ld Results2 %ld\n",
                            results,results2);
                        fprintf(STDERROR,"hashPointers[results] %ld\n",
                            getHashPointer(results));
                        fprintf(STDERROR,"location 0x%lx\n",location);
                        fseek(fp_index,0L,SEEK_SET);
                        for(i=0;i < MAX_GED_HASH;i++)
                        {
                            nextIndex2 = getHashPointer(i);
                            if(!fwrite(&nextIndex2,sizeof(nextIndex2),1,fp_index))
                            {
                                fprintf(STDERROR,
                                "Unable to write to index file %s\n",
                                index_file);
                                exit(0);
                            }
                        }
                        fclose(fp_index);
                        fclose(fp_HashedRinFile);
                        exit(0);
                    }
                    if(hashTagRecordTmp.nextTag == 0)
                    {
                        hashTagRecordTmp.nextTag = endOfFile;
                        /* position to rewrite this hashTagRecord */
                        fseek(fp_index,-(long)sizeof(hashTagRecord),SEEK_CUR);
                        if(!fwrite(&hashTagRecordTmp,sizeof(hashTagRecord),1,
                            fp_index))
                        {
                            fprintf(STDERROR,
                            "Unable to write to index file %s\n",index_file);
                            exit(0);
                        }
                        /* now write the new hashTagRecord at eof */
#if 0
                        memset(hashTagRecord.name,0,sizeof(hashTagRecord.name));
#endif
                        strcpy(hashTagRecord.name,cptr1);
                        hashTagRecord.loc = cur_ged_line_start;
                        hashTagRecord.nextTag = 0;
#if PAD_HASH_LOC
                        hashTagRecord.pad = 0;
#endif
                        fseek(fp_index,0L,SEEK_END); /* get location of eof */
#if SPECIAL_SPOT
                        if(!strcmp(cptr1,special))
                        {
                            iAmHere = ftell(fp_index); /* debug */
                            fprintf(STDERROR,"special %s goes at 0x%lx\n",
                                special,iAmHere);
                            fprintf(STDERROR,"DEPTH %d\n",depth);
                            fprintf(STDERROR,"Results %ld\n", results);
                            fprintf(STDERROR,"hashPointers[results] %ld\n",
                                getHashPointer(results));
                            val = getHashPointer(results) -1;
                            location = (long)((long)(MAX_GED_HASH * (long)sizeof(nextIndex)) +
                                ((long)(val) * (long)sizeof(hashTagRecord)));
                            fprintf(STDERROR,"Calculated loc for first hashTagRecord = 0x%lx\n",location);
                            fprintf(STDERROR,"hashTagRecord.loc = 0x%lx\n",hashTagRecord.loc);
                        }
#endif
                        if(!fwrite(&hashTagRecord,sizeof(hashTagRecord),1,fp_index))
                        {
                            fprintf(STDERROR,
                            "Unable to write to index file %s\n",
                            index_file);
                            exit(0);
                        }
                        fseek(fp_index,0L,SEEK_END); /* get location of eof */
                        endOfFile = ftell(fp_index);
                        nextIndex++;
                        break; /* out of while */
                    }
                    else
                    {
                        /* position at next hashTagRecord */
                        fseek(fp_index,hashTagRecordTmp.nextTag,SEEK_SET);
                    }
                }

            }

        }
        fclose(fp_gedcom);
        fclose(fp_HashedRinFile);
        fclose(fp_HashedMRinFile);
        if(showHashDetails)
        {
            fprintf(stderr,"Number of tags %d\n",nextIndex -1);
            fprintf(stderr,"Number of collisions %ld\n",collisions);
            fprintf(stderr,"Maximum depth of collisions %ld\n",maxDepth);
        }
        fseek(fp_index,0L,SEEK_SET);
        for(i=0;i < MAX_GED_HASH;i++)
        {
          nextIndex2 = getHashPointer(i);
          if(!fwrite(&nextIndex2,sizeof(nextIndex2),1,fp_index))
          {
            fprintf(STDERROR,"Unable to write to index file %s\n",index_file);
            exit(0);
          }
        }
        dumpRamFile2Disk(index_file);
        fclose(fp_index);
}

