#ifndef GEDHASH_H
#define GEDHASH_H
#include "gedutils.h"
#include "gedrtns.h"
#include "gedindex.h"
#ifndef RAMFILE_H
#include "ramfile.h"
#endif

#define MAX_GED_HASH 16384

/* just pads the hash file to make it easier to look at with a hex */
/* dump.  Set to zero for released version */

#define PAD_HASH_LOC 0

typedef struct HASH_LOC{
    char name[MAX_TAG_NAME_LEN];
    long loc;
    long nextTag;
    RIN rin;
#if PAD_HASH_LOC
    int pad; /* to make it line up in hex dumps */
#endif
}hash_loc;

typedef struct HASH_RIN{
        char name[MAX_TAG_NAME_LEN];
        long loc;
#if PAD_HASH_LOC
        char pad[8]; /* to make it line up in hex dumps */
#endif
}hash_rin;

/* Typical gedcom hash usage. */
/* 1. Alloc */
int alloc_hash_buffers(void);
/* 2. Start gedcom files loop. Hash the gedcom file */
void hash_gedcom_file(char *gedcomFile);
/* 3. init to use the hash and index values with the find routines. */
int init_use_hash(char *gedcom_file);
/* 4. Close the hash and index files for this gedcom file */
/*    Loop back to step 2 for next gedcom file. */
void close_use_hash(char *gedcom_file,int deleteHashFile);
/* 5. All done with hash stuff, free buffers. */
void free_hash_buffers(void);
/* 6. Delete the ramfile.c's overflow file */
/*    deleteOverFlowFile(); */

int find_hash_entry(char *string,long *location);
int find_rin_entry(RIN rin, long *location);
int find_mRin_entry(RIN mrin, long *location);
long hash(char *name, int  size);
RIN indi2rin(char *indi);

extern RIN max_rin;
extern RIN max_mrin;
extern hash_loc hashTagRecord;
extern hash_rin rinLocRecord;
extern int showHashDetails;
extern char hashedGedcomFileName[FILENAME_MAX];
extern FILE *fp_ReadHashedGedcomFile;

#endif /* GEDHASH_H */
