/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#ifdef SUNOS4_1_2
#include <unistd.h>
#endif
/* #include "fsort.h" */
#include "gedhash.h"
#include "gedrtnsh.h"
#include "gedindex.h"

/* hash based gedrtns.c */

#ifdef unix
#define DIR_SEP '/'
#else
#define DIR_SEP '\\'
#endif

/* Skip over tag */
char *
skip_over_tag(char *cptr)
{
	/* Skip leading spaces */
	while(isspace(*cptr)) cptr++;
	/* Skip over level number */
	while(isdigit(*cptr)) cptr++;
	/* Skip over spaces */
	while(isspace(*cptr)) cptr++;
	/* Skip over tag */
	while(isalnum(*cptr)) cptr++;
	/* Skip over spaces */
	while(isspace(*cptr)) cptr++;
	return (cptr);
}

/* Scans the gedcom line passed in via str.  Extracts out the level number */
/* 0 @I81@ INDI */
/* If invalid line, returns -1 */
/* Pass in the address of the pointer to the string.  If modit is 1, it */
/* will bump the pointer to after the level number. */
/* @I81@ INDI */

int
get_level_nbr(char **str, int modit)
{
	char *ptr;
    char *cptr;
	int level;

	ptr = *str;
	/* Skip white space */
	while(isspace(*ptr)) ptr++;
	/* Get level number */
    if(isdigit(*ptr))
        level = atoi(ptr);
    else
        return -1;

    if(level == 0)
    {
    /* sanity should be two @ after number */
    cptr = ptr;
    if((cptr = strchr(cptr,'@')) == NULL)
      return -1;
      cptr++;
    if((cptr = strchr(cptr,'@')) == NULL)
      return -1;
    }

	/* If they want us to modify the pointer str to point after the level */
	if(modit)
	{
		while(isdigit(*ptr)) ptr++;
		while(isspace(*ptr)) ptr++;

		*str = ptr;
	}

	return level;

}

char *
remove_trailing_blanks(char *string)
{
	int i;
    
	for(i = strlen(string) -1;i > 0;i--)
    {
          if(string[i] == ' ')
          string[i] = 0;
          else
          break;
    }
    return(string);
}


int
is_this_the_right_type(char *readBuf,char *id, char *type)
{
	char *ptr;
	char *ptr2;
	int level;

    if(strstr(readBuf,type) != NULL)
    {
        ptr = readBuf;
        level = get_level_nbr(&ptr, 1);
        if((level == 0) && (*ptr == '@'))
        {
            ptr2 = ptr;
            while((!isspace(*ptr2)) && (*ptr2 != 0)) ptr2++;
            *ptr2 = 0;
            ptr2++;
            remove_trailing_blanks(ptr2);
            if(!strcmp(ptr2,type))
            {
                strcpy(id,ptr);
                return(1);
            }
        }
    }
    return(0);
}

char *
find_next_of_this_type(FILE *fp_gedcom,char *type)
{
	static char id[40];
    char readBuf[READ_GED_LINE_SIZE];

    while(readGedLine(fp_gedcom,readBuf,READ_GED_LINE_SIZE - 1))
	{

        if(is_this_the_right_type(readBuf,id,type))
            return(&id[0]);
	}
	return NULL;
}


/* Scans through the gedcom file searching for the next individual record */
/* 0 @I81@ INDI */
/* returns NULL if not found */
/* returns the address of the string @I81@ if found.  Note, this string is */
/* modified each time this routine is called */
char *
find_next_individual(FILE *fp_gedcom)
{
	return(find_next_of_this_type(fp_gedcom,"INDI"));
}

char *
find_next_family(FILE *fp_gedcom)
{
	return(find_next_of_this_type(fp_gedcom,"FAM"));
}


/* Reads lines looking for tagName. example: DATE */
/* reads into callers buffer.  Returns 1 if found.  Monitors */
/* level number of lines scanned, and if they are < than */
/* curLevel, then returns 0. */
/* Example: We found individual @i231@  His name is level 1.  Level is the */
/* first number in the gedcom line.  If we search for BIRT (birth info) */
/* and there is none, we will hit another level 0 INDI, and fail */ 
int
find_tag(FILE *fp_gedcom,char *tagName, int curLevel, char *buf)
{
	char *cptr;
	int level;
	int tagLen;


    while(readGedLine(fp_gedcom,buf,READ_GED_LINE_SIZE - 1))
	{
	remove_trailing_blanks(buf);
	cptr = buf;

	while(isspace(*cptr)) cptr++;
	level = atoi(cptr);
	
	while(isdigit(*cptr)) cptr++;
	while(isspace(*cptr)) cptr++;


	if(level < curLevel)
	{
	  return 0;
	}

	tagLen = strlen(tagName);
	if(!strncmp(tagName,cptr,tagLen))
	{
		if(isspace(cptr[tagLen]) || (cptr[tagLen] == 0))
		{
		/* Found it */
#if DEBUG_IT 
		printf("Found tag %s in %s\n",tagName,buf);
#endif /* DEBUG_IT */
		return 1;
		}
	}
	}
	return 0;
}

int
find_next_level_zero(FILE *fp_gedcom, char *readBuf)
{
	static char id[40];
	int level;

    while(readGedLine(fp_gedcom,readBuf,READ_GED_LINE_SIZE - 1))
	{
        level = get_level_nbr(&readBuf, 0);
        if(level == 0)
		{
            return 1;
		}
	}
    return 0;
}

long cur_ged_line_start = 0; /* readGedLine keeps track of start of line here */
unsigned int gedLineLength = 0;
int
readGedLine(FILE *fp_src,char *buf,int maxlen)
{
    cur_ged_line_start = ftell(fp_src);
    if(fgets(buf, maxlen, fp_src) == 0) return 0;
    gedLineLength = strlen(buf);
    if(buf[gedLineLength -1] == '\n')
        buf[gedLineLength -1] = '\0';
    /* if we encounter a "0 TRLR" return 0 to ignore garbage after */
    if(gedLineLength < 8)
    {
        if((strstr(buf,"0 ") != NULL) && (strstr(buf,"TRLR") != NULL))
            return 0;
    }

    return 1;
}

