/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#ifndef unix
#include <process.h>
#endif
#ifdef SUNOS4_1_2
#include <unistd.h>
#endif
#include "gedutils.h"
#include "gedsubs.h"
#include "gedhash.h"
#include "gedrtnsh.h"

RIN rin_max = 0; /* must be filled in by application */

char *tagTypes =
"NAME SEX BIRT DATE PLAC DEAT MARR HUSB WIFE CHIL FAMC FAMS BAPL ENDL TEMP REFN BURI CHR";

#define TAGTYPE_NAME 0
#define TAGTYPE_SEX 1
#define TAGTYPE_BIRT 2
#define TAGTYPE_DATE 3
#define TAGTYPE_PLAC 4
#define TAGTYPE_DEAT 5
#define TAGTYPE_MARR 6
#define TAGTYPE_HUSB 7
#define TAGTYPE_WIFE 8
#define TAGTYPE_CHIL 9
#define TAGTYPE_FAMC 10
#define TAGTYPE_FAMS 11
#define TAGTYPE_BAPL 12
#define TAGTYPE_TEMP 13
#define TAGTYPE_REFN 14
#define TAGTYPE_BURI 15
#define TAGTYPE_CHR 16

#define TAGTYPE_UNDEFINED -1

int
name2INDIV_REC(char *source, INDIV_REC *irec)
{
    char lastName[50];
    char firstName[50];
    char middleName[50];
	char *cptr;
	char *cptr2;
	char *cptr3;
    char *cptr4;

    lastName[0] = 0;
    middleName[0] = 0;
    firstName[0] = 0;
    cptr = source;
	cptr = skip_over_tag(cptr);
	
	if((cptr2 = strchr(cptr,'/')) != NULL)
	{
	cptr2++; /* skip over slash */
	if((cptr3 = strchr(cptr2,'/')) != NULL)
	{
		*cptr3 = 0;
        cptr4 = cptr2;
        while(isspace(*cptr4)) cptr4++;
        strncpy(lastName,cptr4,sizeof(lastName));
	}
    else
    {
        /* Some programs don't put the terminating slash, */
        /* To make things worse, they sometimes add a comma, title after.*/
        cptr3 = cptr2;
        while(*cptr3 != 0)
        {
            if(!isalpha(*cptr3))
            {
                *cptr3 = 0;
                break;
            }
            cptr3++;
        }
        cptr4 = cptr2;
        while(isspace(*cptr4)) cptr4++;
        strncpy(lastName,cptr4,sizeof(lastName));
    }
    lastName[sizeof(lastName) - 1] = 0;
	cptr2--; /* Back to the first slash */
	*cptr2 = 0; /* Dump the last name so we can work on the first */
	}
	if(((cptr2 = strchr(cptr,' ')) != NULL) || (*cptr != 0))
	{
		/* Got a first name */
		if(cptr2 == 0)
		{
		/* No middle name */
        cptr4 = cptr;
        while(isspace(*cptr4)) cptr4++;
        strncpy(firstName,cptr4,sizeof(firstName));
		}
		else
		{
		/* first name and middle name */
		*cptr2 = 0;
        cptr4 = cptr;
        while(isspace(*cptr4)) cptr4++;
        strncpy(firstName,cptr4,sizeof(firstName));
		cptr2++;
		while(isspace(*cptr2)) cptr2++;
		if(*cptr2 != 0)
            strncpy(middleName,cptr2,sizeof(middleName));
		
		}
	}
    lastName[sizeof(lastName) - 1] = 0;
    middleName[sizeof(middleName) - 1] = 0;
    firstName[sizeof(firstName) - 1] = 0;
    delete_trailing_blanks(lastName);
    delete_trailing_blanks(firstName);
    delete_trailing_blanks(middleName);
#if 1
    strcpy(irec->names.lastName,lastName);
    strcpy(irec->names.middleName,middleName);
    strcpy(irec->names.firstName,firstName);
#else
    strcpy(irec[0],lastName);
    strcpy(irec[1],middleName);
    strcpy(irec[2],firstName);
#endif
	return 1;
}

int
getTagType(char *type)
{
    char *loc;
    int addr;
    char *cptr;
    char name[25];

    /* extract out the tag */
    cptr = name;
    while(isalnum(*type)) *cptr++ = *type++;
    *cptr = 0;

    if((loc = strstr(tagTypes,name)) != NULL)
    {
        addr = (int)(loc -  tagTypes);
        switch(addr)
        {
            case 0:
                return TAGTYPE_NAME;
            case 5:
                return TAGTYPE_SEX;
            case 9:
                return TAGTYPE_BIRT;
            case 14:
                return TAGTYPE_DATE;
            case 19:
                return TAGTYPE_PLAC;
            case 24:
                return TAGTYPE_DEAT;
            case 29:
                return TAGTYPE_MARR;
            case 34:
                return TAGTYPE_HUSB;
            case 39:
                return TAGTYPE_WIFE;
            case 44:
                return TAGTYPE_CHIL;
            case 49:
                return TAGTYPE_FAMC;
            case 54:
                return TAGTYPE_FAMS;
            case 79:
                return TAGTYPE_BURI;
            case 84:
                return TAGTYPE_CHR;
        }

    }
    return TAGTYPE_UNDEFINED;
}

/* Takes this
0 @I5@ INDI
or this
1 FAMS @F1@
and returns I5 or F1.
and 1 for success
*/

int
extract_tag_value(char *source, char *destination)
{
    char *cptr1;
    char *cptr2;

    *destination = 0;
    cptr1 = strchr(source,'@');
    if(cptr1 == NULL)
    {
       return 0;
    }
    cptr1++;
    cptr2 = strchr(cptr1,'@');
    if(cptr2 == NULL)
    {
        strcpy(destination,cptr1);
    }
    else
    {
        strncpy(destination,cptr1,(int)(cptr2-cptr1));
        destination[(int)(cptr2-cptr1)] = 0;
    }
    return 1;
}

void
getPlaceDate(PLACES *activity,int baseLevel,FILE *fp_in,char *readBuf,int bufsize)
{
    char *buf;
    char *cptr;
    int level;
    int tagType;
#if 0
1 BIRT
2 DATE  3 APR 1946
2 PLAC Yuma,,AZ
#endif
    while(readGedLine(fp_in,readBuf,bufsize))
    {
        buf = readBuf;
        if(!(level = get_level_nbr(&buf,1)))
            return; /* Done with this individuals data */
        if(level <= baseLevel)
            return; /* done */
        if(level == -1)
            return;
        if(level == baseLevel + 1)
        {
            tagType = getTagType(buf);
            switch(tagType)
            {
                case TAGTYPE_DATE:
                       cptr = skip_over_tag(readBuf);
                       strncpy(activity->date,cptr,DATE_STR_LEN);
                break;
                case TAGTYPE_PLAC:
                       cptr = skip_over_tag(readBuf);
                       strncpy(activity->place,cptr,PLACE_WORD_LEN);
                break;
                case TAGTYPE_UNDEFINED:
                    break;
            }
        }
    }
    activity->place[PLACE_WORD_LEN -1] = 0;
    activity->date[DATE_STR_LEN -1] = 0;
}

void get_indiv_rec(RECORD_PTR rin,INDIV_REC *indrec)
{
    long location;
    char readBuf[256];
    char tagValBuf[MAX_TAG_NAME_LEN];
    char *buf;
    char *cptr;
    char c;
    int level;
    int tagType;
    int gotOne;

    memset(indrec,0,sizeof(INDIV_REC));
    indrec->sex = 'D'; /* deleted, the sign of failure */
    if(rin_max == 0)
    {
        fprintf(stderr,"\nApplication err.  Must fill in rin_max. Aborting\n");
        exit(-1);
    }
    if((rin == 0) || (rin > rin_max))
    {
        memset(indrec,0,sizeof(INDIV_REC));
        return;
    }
    if(!find_rin_entry((unsigned int) rin, &location))
    {
        fprintf(stderr,"Error searching for rin %d\n",rin);
        exit(0);
    }
    fseek(fp_ReadHashedGedcomFile,(long)location,SEEK_SET);
    readGedLine(fp_ReadHashedGedcomFile,readBuf,sizeof(readBuf));
    indrec->rin = rin;
#if 0
/*
0 @I5@ INDI
1 NAME John Henry /SMITH/
1 TITL III
1 SEX M
1 BIRT
2 DATE  3 APR 1946
2 PLAC Yuma,,AZ
1 REFN          2
1 NOTE Hobbies: programming,Raquetball
1 FAMS @F1@
1 FAMC @F3@
*/
#endif
        gotOne = 0;
        while(1)
        {
        if(!gotOne)
            if(!readGedLine(fp_ReadHashedGedcomFile,readBuf,sizeof(readBuf)))
            return;
        gotOne = 0;
        buf = readBuf;
        if(!(level = get_level_nbr(&buf,1)))
            return; /* Done with this individuals data */
        if(level == -1)
            return;
        if(level == 1)
        {
            tagType = getTagType(buf);
            switch(tagType)
            {
                case TAGTYPE_NAME:
                    name2INDIV_REC(readBuf, indrec);
                    break;
                case TAGTYPE_SEX:
                       cptr = skip_over_tag(readBuf);
                       indrec->sex = *cptr;
                    break;
                case TAGTYPE_BIRT:
                    getPlaceDate(&indrec->birth,1,fp_ReadHashedGedcomFile,
                        readBuf,sizeof(readBuf));
                    gotOne = 1;
                    break;
                case TAGTYPE_DEAT:
                    getPlaceDate(&indrec->death,1,fp_ReadHashedGedcomFile,
                        readBuf,sizeof(readBuf));
                    gotOne = 1;
                    break;
                case TAGTYPE_BURI:
                    getPlaceDate(&indrec->burial,1,fp_ReadHashedGedcomFile,
                        readBuf,sizeof(readBuf));
                    gotOne = 1;
                    break;
                case TAGTYPE_CHR:
                    getPlaceDate(&indrec->christ,1,fp_ReadHashedGedcomFile,
                        readBuf,sizeof(readBuf));
                    gotOne = 1;
                    break;
                case TAGTYPE_FAMS:
                    extract_tag_value(readBuf, tagValBuf);
                    indrec->ownmarriage = indi2rin(tagValBuf);
                    if(indrec->ownmarriage == 0)
                    {
                        fprintf(stderr,"Unable to find family record %s\n",
                            readBuf);
                    }
                    break;
                case TAGTYPE_FAMC:
                    extract_tag_value(readBuf, tagValBuf);
                    indrec->parentmarriage = indi2rin(tagValBuf);
                    if(indrec->parentmarriage == 0)
                    {
                        fprintf(stderr,"Unable to find family record %s\n",
                            readBuf);
                    }
                    break;
                case TAGTYPE_UNDEFINED:
                    break;
            }
        }
    }
#if 0
  char indraw[92];

  if(rin_max == 0)
  {
    fprintf(stderr,"\nApplication err.  Must fill in rin_max. Aborting\n");
    exit(-1);
  }
  if((rin == 0) || (rin > rin_max))
  {
    memset(indrec,0,sizeof(INDIV_REC));
    return;
  }
  get_paf_rec(rin,&paf_indiv,indraw);
  memcpy(&indrec->names[0],&indraw[0],10);
  indrec->sex = indraw[10];
  memcpy(&indrec->birthdate,&indraw[11],48);
  memcpy(indrec->ldsinfo,&indraw[59],15); /* Just save LDS, but don't parse */
  memcpy(&indrec->oldersib,&indraw[74],18);
  indrec->rin = rin;
#endif
}

void get_marr_rec(RECORD_PTR mrin,MARR_REC *marec)
{
    long location;
    char readBuf[256];
    char tagValBuf[MAX_TAG_NAME_LEN];
    char *buf;
    int level;
    int tagType;
    int gotOne;

    memset(marec,0,sizeof(MARR_REC));
    if(max_mrin == 0)
    {
        fprintf(stderr,"\nApplication err.  Must fill in max_mrin. Aborting\n");
        exit(-1);
    }
    if((mrin == 0) || (mrin > max_mrin))
    {
        memset(marec,0,sizeof(MARR_REC));
        return;
    }
    if(!find_mRin_entry((unsigned int) mrin, &location))
    {
        fprintf(stderr,"Error searching for mrin %d\n",mrin);
        exit(0);
    }

    marec->mrin = mrin;

    fseek(fp_ReadHashedGedcomFile,(long)location,SEEK_SET);
#if 0
/*
0 @F280@ FAM
1 HUSB @I567@
1 WIFE @I568@
1 CHIL @I75@
*/
#endif
    /* read the tag line */
    readGedLine(fp_ReadHashedGedcomFile,readBuf,sizeof(readBuf));

        gotOne = 0;
        while(1)
        {
        if(!gotOne)
            if(!readGedLine(fp_ReadHashedGedcomFile,readBuf,sizeof(readBuf)))
            return;
        gotOne = 0;
        buf = readBuf;
        if(!(level = get_level_nbr(&buf,1)))
            return; /* Done with this individuals data */
        if(level == -1)
            return;
        if(level == 1)
        {
            tagType = getTagType(buf);
            switch(tagType)
            {
                case TAGTYPE_MARR:
                    getPlaceDate(&marec->married,1,fp_ReadHashedGedcomFile,
                        readBuf,sizeof(readBuf));
                    gotOne = 1;
                    break;
                case TAGTYPE_HUSB:
                    extract_tag_value(readBuf, tagValBuf);
                    marec->husband = indi2rin(tagValBuf);
                    if(marec->husband == 0)
                    {
                        fprintf(stderr,"Unable to find husband %s\n",readBuf);
                    }
                    break;
                case TAGTYPE_WIFE:
                    extract_tag_value(readBuf, tagValBuf);
                    marec->wife = indi2rin(tagValBuf);
                    if(marec->wife == 0)
                    {
                        fprintf(stderr,"Unable to find wife %s\n",readBuf);
                    }
                    break;
                case TAGTYPE_CHIL:
                    extract_tag_value(readBuf, tagValBuf);
                    marec->child = indi2rin(tagValBuf);
                    if(marec->child == 0)
                    {
                        fprintf(stderr,"Unable to find child %s\n",readBuf);
                    }
                    break;
            }
        }
    }
#if 0
  char marraw[28];

  get_paf_rec(mrin,&paf_marr,marraw);
  memcpy(&marec->husband,&marraw[0],6);
  memcpy(&marec->marriagedate,&marraw[6],17);
  memcpy(&marec->mnext_husband,&marraw[23],4);
  marec->divorce = marraw[27];
  marec->mrin = mrin;
#endif
}
void get_note_rec(RECORD_PTR nbp,NOTE_REC *noterec)
{
#if 0
  char notearray[256];

  get_paf_rec(nbp,&paf_notes,notearray);
  memcpy(&noterec->next_pad,notearray,2);
  memcpy(noterec->notelines,&notearray[2],254);
#endif
}

    
