/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#ifdef SUNOS4_1_2
#include <unistd.h>
#endif
#include "gedsubs.h"
#include "gedutils.h"
#include "portable.h"

#define EOS             '\0'
extern char *monthnames[];
extern RIN     rin_max;

char *
delete_trailing_blanks(char *buf)
{
    char *ptr;
    ptr = buf + strlen(buf) - 1;
    while((*ptr == ' ') &&(ptr != buf))
      *ptr-- = 0;
    return(buf);
}
/* returns the address of the current date in the form mm/dd/yy */
char *
get_cur_date()
{
/* date to go on chart.  Format like : 16 Nov 1993  */
    static char buf[40]; /* where date will be built */
    char tmpbuf[40];
    struct tm *timeptr;
    time_t elapsed_seconds;
    time (&elapsed_seconds);
    timeptr = localtime(&elapsed_seconds);
    sprintf(&buf[0],"%d ",timeptr->tm_mday);
    strcat(buf,monthnames[timeptr->tm_mon+1]);
    sprintf(tmpbuf," 19%2d",timeptr->tm_year);
    strcat(buf,tmpbuf);
#if 0
    s
    sprintf(&buf[0],"%2d/%2d/%2d",
    	timeptr->tm_mon+1,
		timeptr->tm_mday,
		timeptr->tm_year);
    for(i=1;i < strlen(buf);i++)
    {
	if(buf[i] == ' ')
	    buf[i] = '0';
    }
#endif
    return(&buf[0]);
}


/* Returns NULL if no more data */
char *
readLine(char *buf,int maxlen,FILE *fp_in)
{
    int slen;
    char *status;
    buf[0] = buf[1] = EOS;
    if((status = fgets(buf, maxlen, fp_in)) == NULL)
        return status;
    if(buf[(slen=strlen(buf)) -1] == '\n')
        buf[slen -1 ] = EOS;
    buf[maxlen -1] = EOS;
    if(slen == 1)
        buf[slen] = EOS;
    return status;
}

/* Like strcpy except it doesn't copy the null terminator.  Overlays dest */
/* with src, up to a max size of n.  Watches dest, and if a terminating 0 */
/* is detected, it will terminate the string with 0 */
void
strnOvr(char *dest, char *src, int max_size)
{
    int i;
    int len;
    int term_it;

    term_it = 0;
    len = strlen(src);
    if(len > max_size) len = max_size;
    for(i=0;i < len;i++)
    {
        if(*dest == 0)
            term_it = 1;
        *dest++ = *src++;
    }
    if(term_it)
        *dest = 0;
}

void 
printnotes(int position,RECORD_PTR rin,FILE *fp_out)
{
#if 0
    INDIV_REC       indiv;
    NOTE_REC        pad;
    int             i;
    char            s1[100];

    get_indiv_rec(rin, &indiv);

    /* Only proceed if individual is not deleted and has a note. */
    if (indiv.sex != 'D' && (indiv.notepad != 0))
    {
        convert_name(s1, indiv.names);
        fprintf(fp_out,"\n%s [%d]",s1,position + 1);
        /* First note will insert newline */
        {
            int             llen;
            int             newline;
            RECORD_PTR      npad;
            llen = 0;
            newline = 0;
            npad = indiv.notepad;
            while (npad != 0)   /* Chain through notepads */
            {
                /* printf("[%d]",npad); */
                get_note_rec(npad, &pad);
                npad = pad.next_pad;
                for (i = 0; i < 254; i++) {
                    if (pad.notelines[i] == 0)
                        if (llen == 0) {
                            llen = 1;
                        } 
                        else if (llen == 1) {
                            llen = 0;
                            break;    /* Zero length line terminates notes */
                        } 
                        else {
                            llen = 0;
                            fprintf(fp_out,"\n  ");
                        }
                    else if (llen == 0 && pad.notelines[i] == 1) {
                        fprintf(fp_out,"\n  ");   /* Start of new note */
                        newline = 1;
                        llen++;
                    } 
                    else {
                        fprintf(fp_out,"%c",pad.notelines[i]);
                        llen++;
                    }
                }
            }
            if (!newline)
                fprintf(fp_out,"\n");   /* Just in case note was blank */
        }
    }
#endif /* USE_GEDCOM */
    return;
}

char *used_rin_file="rin_file.$$$";


void
erase_used_rin_file()
{
#ifdef unix
    unlink(used_rin_file);
#else
    remove(used_rin_file);
#endif /* DOS */
}

void
init_used_rin_file()
{
    FILE *fp_out;
	USED_RIN tmp;
    unsigned int i;

    fp_out = fopen(used_rin_file,FOPEN_WB);

	tmp.chart_nbr = 0;
	tmp.from_position = 0;
	tmp.used = 0;

    if(fp_out == NULL)
    {
        printf("Unable to add to tmp file %s\n",used_rin_file);
        exit(-1);
    }
	for(i=0;i < rin_max ;i++)
	{
        fwrite((char *)&tmp,sizeof(struct USED_RIN),1,fp_out);
	}
    fclose(fp_out);
}

int
read_from_used_rin_file(USED_RIN *rin,RECORD_PTR person)
{
    FILE *fp_out;
    int status;
    fp_out = fopen(used_rin_file,FOPEN_RB);
    if(fp_out == NULL)
    {
        printf("Unable to open tmp file %s\n",used_rin_file);
        exit(-1);
    }
    status = fseek(fp_out,person * sizeof(struct USED_RIN),SEEK_SET);
    if(!status)
    {
        status = fread(rin,sizeof(struct USED_RIN),1,fp_out);
        fclose(fp_out);
#ifdef DEBUG_PROG
        printf("Reading person %d used flag = %d in chart %d position %d\n",
             person,rin->used,rin->chart_nbr,rin->from_position);
#endif /* DEBUG_PROG */
        return 1;
    }
    else
    {
        printf("Seek error on %s, rin # %d\n",used_rin_file,person);
    }
    fclose(fp_out);
    return 0;
}

int
write_to_used_rin_file(USED_RIN *rin,RECORD_PTR person)
{
    FILE *fp_out;
    int status;
#ifdef DEBUG_PROG
                printf("Marking person %d used flag = %d chart %d position %d\n",
                person,rin->used,rin->chart_nbr,rin->from_position);
#endif /* DEBUG_PROG */
    fp_out = fopen(used_rin_file,FOPEN_RBP);
    if(fp_out == NULL)
    {
        printf("Unable to open tmp file %s\n",used_rin_file);
        exit(-1);
    }
    status = fseek(fp_out,person * sizeof(struct USED_RIN),SEEK_SET);
    if(!status)
    {
        fwrite(rin,sizeof(struct USED_RIN),1,fp_out);
        fclose(fp_out);
        return 1;
    }
    fclose(fp_out);
    return 0;
}

char *cont_file="contfile.$$$";
int nbr_cont_file_entries = 0;
void
delete_file(char *name)
{
#ifdef unix
    unlink(name);
#else
    remove(name);
#endif /* DOS */
}

void
delete_cont_file()
{
#ifdef unix
    unlink(cont_file);
#else
    remove(cont_file);
#endif /* DOS */
}

void
erase_cont_file()
{
    FILE *fp_cont;
    nbr_cont_file_entries = 0;
    fp_cont = fopen(cont_file,FOPEN_WB);
    if(fp_cont == NULL)
    {
        printf("Unable to open tmp file %s\n",cont_file);
        exit(-1);
    }
    fclose(fp_cont);
}

void
add_to_cont_file(CHART_START *cont)
{
    FILE *fp_cont;
    fp_cont = fopen(cont_file,FOPEN_AB);
    nbr_cont_file_entries++;
    if(fp_cont == NULL)
    {
        printf("Unable to add to tmp file %s\n",cont_file);
        exit(-1);
    }
    fwrite(cont,sizeof(struct CHART_START),1,fp_cont);
    fclose(fp_cont);
}

/* returns 0 if no entry found */
int
read_from_cont_file(CHART_START *cont,int chart_nbr)
{
    FILE *fp_cont;
    int status;

    if(chart_nbr > (nbr_cont_file_entries -1))
      return 0;
    fp_cont = fopen(cont_file,FOPEN_RB);
    if(fp_cont == NULL)
    {
        printf("Unable read chart continuation file %s\n",cont_file);
        exit(-1);
    }
    status = fseek(fp_cont,chart_nbr * sizeof(struct CHART_START),SEEK_SET);
    if(!status)
    {
        fread(cont,sizeof(struct CHART_START),1,fp_cont);
        fclose(fp_cont);
        return 1;
    }
    fclose(fp_cont);
    return 0;
}

int
add_int_array(int *array, int fields)
{
    int i;
    int sum;
    sum = 0;
    for(i = 0;i < fields;i++)
        sum += array[i];
    return(sum);
}

/* Shorten words as necessary to fit them in buffer. */
int abreviateIfNecessary(char *buf, char *source,int maxLen,
    int *offsets, int *lens,int maxFields,
    char seperator, char *sepStr)
{
  int sepStrLen;
  char *cptr;
  int lastOffset;
  int fields;
  int alternate;
    strcpy(buf,source);
    delete_trailing_blanks(buf);
    cptr = &buf[0];
    lastOffset = 0;

    sepStrLen = strlen(sepStr);
    for(fields=0;(*cptr != 0) && (fields < maxFields);fields++)
    {
            if((cptr = strchr(cptr,seperator)) != NULL)
            {
                offsets[fields] = lastOffset;
                lens[fields] = cptr - &buf[lastOffset];
                cptr++;
                lastOffset = cptr - &buf[0];
            }
            else
            {
                offsets[fields] = lastOffset;
                cptr = &buf[0] + strlen(buf) + 1;
                lens[fields] = cptr - &buf[lastOffset];
                lastOffset = cptr - &buf[0];
                cptr--; /* leave pointing to zero terminator */
            }
    }

    alternate = 0;
    while((add_int_array(lens,fields) + ((fields -1) * sepStrLen)) > maxLen)
    {
        /* would like to leave first and last untouched as much as possible */
        if(fields > 2)
        {
            alternate++;
            /* if the middle fields have room to shrink */
            if(add_int_array(&lens[1],fields - 2) > ((fields - 2) * 1))
            {
                /* shrink middle fields */
                alternate = alternate % (fields - 2);
                if(lens[alternate + 1] > 1)
                {
                    lens[alternate + 1]--;
                }
            }
            else
            {
                /* shrink all fields */
                alternate = alternate % fields;
                if(lens[alternate] > 1)
                {
                    lens[alternate]--;
                }
            }
        }
        else
        {

                alternate = alternate % fields;
                if(lens[alternate] > 1)
                {
                    lens[alternate]--;
                }
        }
    }
    return(fields);
}

char *convert_name(char *s, NAMES *ptrs,int maxLen)
{
  int             extraSpace;
  int             lens[3];
  s[0] = 0;
  
    s[0] = 0;
    lens[0] = strlen(ptrs->lastName);
    lens[1] = strlen(ptrs->middleName);
    lens[2] = strlen(ptrs->firstName);
    extraSpace = 0;
    if(ptrs->middleName[0])
        extraSpace++;
    if(ptrs->lastName[0])
        extraSpace++;

    while((lens[0] + lens[1] + lens[2] + extraSpace + 1) > maxLen)
    {
        if(lens[1] > 1)
        {
            lens[1]--;
        }
        else
            if(lens[2] > 1)
            {
                lens[2] --;
            }
            else
             lens[0]--;
    }
    if(ptrs->firstName[0])
    {
        strncat(s,ptrs->firstName,lens[2]);
    }
    if(ptrs->middleName[0])
    {
        strcat(s," ");
        strncat(s,ptrs->middleName,lens[1]);
    }
    if(ptrs->lastName[0])
    {
        strcat(s," ");
        strncat(s,ptrs->lastName,lens[0]);
    }
  return(s);
}

/* Just puts last name first. */
char *convert_name2(char *s, NAMES *ptrs,int maxLen)
{
  int extraSpace;
  int             lens[3];
  s[0] = 0;
  
    s[0] = 0;
    lens[0] = strlen(ptrs->lastName);
    lens[1] = strlen(ptrs->middleName);
    lens[2] = strlen(ptrs->firstName);
    extraSpace = 0;
    if(ptrs->middleName[0])
        extraSpace++;
    if(ptrs->lastName[0])
        extraSpace++;

    while((lens[0] + lens[1] + lens[2] + extraSpace + 1) > maxLen)
    {
        if(lens[1] > 1)
        {
            lens[1]--;
        }
        else
            if(lens[2] > 1)
            {
                lens[2] --;
            }
            else
             lens[0]--;
    }
    if(ptrs->lastName[0])
    {
        strncat(s,ptrs->lastName,lens[0]);
    }
    if(ptrs->firstName[0])
    {
        strcat(s," ");
        strncat(s,ptrs->firstName,lens[2]);
    }
    if(ptrs->middleName[0])
    {
        strcat(s," ");
        strncat(s,ptrs->middleName,lens[1]);
    }
  return(s);
}


char *convert_place(char *s, PLACES *ptrs,int maxLen)
{
#if 0
  s[0] = 0;
  if(ptrs->place[0])
    strcpy(s,ptrs->place);
  return(s);
#else
  char buf[256];
  int  lens[10];
  int offsets[10];
  int fields;
  char seperator;
  char *sepStr;
  int i;
  sepStr = ",";
  seperator = ',';
  fields = abreviateIfNecessary(buf, ptrs->place,maxLen,
      offsets, lens,10,seperator, sepStr);
  /* put them out in backward order */
  s[0] = 0;
  for(i= 0;i < fields;i++)
  {
      if(i != 0)
          strcat(s,sepStr);
      strncat(s,&buf[offsets[i]],lens[i]);
  }
  return(s);
#endif
}


char *convert_place_reversed(char *s, PLACES *ptrs,int maxLen)
{
  char buf[256];
  int  lens[10];
  int offsets[10];
  int fields;
  char seperator;
  char *sepStr;
  int i;
  sepStr = ",";
  seperator = ',';
  fields = abreviateIfNecessary(buf, ptrs->place,maxLen,
      offsets, lens,10,seperator, sepStr);
  /* put them out in backward order */
  s[0] = 0;
  for(i=fields -1;i >= 0;i--)
  {
      if(i != (fields -1))
          strcat(s,sepStr);
      strncat(s,&buf[offsets[i]],lens[i]);
  }
  return(s);
}
#if 0
char *monthnames[]={"NOMONTH","January","February","March","April",
		      "May","June","July","August","September",
		      "October","November","December","SEENOTES",
		      "SUBMITTED","CLEARED","UNCLEARED","COMPLETED",
		      "CANCELLED","STILLBORN","INFANT","DNS/CAN",
		      "DNS","BIC","CHILD"};
#else
char *monthnames[]={"NOMONTH","Jan","Feb","Mar","Apr",
              "May","Jun","Jul","Aug","Sep",
              "Oct","Nov","Dec","SEENOTES",
		      "SUBMITTED","CLEARED","UNCLEARED","COMPLETED",
		      "CANCELLED","STILLBORN","INFANT","DNS/CAN",
		      "DNS","BIC","CHILD"};
#endif

char *convert_date(char *s,PLACES *packed_date)
{
  s[0] = 0;
  if(packed_date->date[0])
    strcpy(s,packed_date->date);
  return(s);
}

char *convert_date_reversed(char *s,PLACES *packed_date)
{
#if 0
  s[0] = 0;
  if(packed_date->date[0])
    strcpy(s,packed_date->date);
  return(s);
#endif
    char buf[256];
    int len;
    int i;
  s[0] = 0;
  if(packed_date->date[0])
  {
    strcpy(buf,packed_date->date);
    len = strlen(buf);
    for(i=len - 1;i >= 0;i--)
    {
        if(buf[i] == ' ')
        {
            buf[i] = 0;
            strcat(s,&buf[i + 1]);
            strcat(s," ");
        }
    }
    if(buf[0])
        strcat(s,buf);
    }
  return(s);
}

