/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef SUNOS4_1_2
#include <unistd.h>
#endif
#include "doswld.h"
#include "gedhash.h"
#include "ramfile.h"

extern getopt(int nargc,char **nargv,char *ostr);
extern char *optarg;
extern int optind;

#define STDERROR stdout

void
usage(char *progName)
{
    printf("\n%s: Usage:\n");
    printf("%s [-d] gedcomfile\n",progName);
    printf(" -d deletes index files created for the gedcom file\n");
}

void
main(int argc, char **argv)
{
    int i;
    int c;
    int arg;
    FILE *fp_gedcom;
    char readBuf[256];
    char readBuf2[256];
    char readBuf3[256];
    RIN rin;
    unsigned int tested;
    int deleteIndexFiles;
    long results;
	char *cptr1;
	char *cptr2;
    long curLoc;
    long tagLoc;

    disableRamFile = 1;
    dosWildCard(&argc,&argv);  /* DOS wild card expander. Just returns if unix */
    deleteIndexFiles = 0;

#if 0
    heapdump(__FILE__,__LINE__);
#endif
    alloc_hash_buffers();
#if 0
    heapdump(__FILE__,__LINE__);
#endif

    if(argc < 2)
    {
        usage(argv[0]);
        exit(-1);
    }
    while ((c = getopt(argc, argv, "d")) != EOF)
        switch(c)
        {
                case 'd':
                deleteIndexFiles = 1;
                break;
                default:
                     usage(argv[0]);
                     exit(-1);
                     break;
        }
    if((optind + 1) > argc)
    {
        usage(argv[0]);
        exit(-1);
    }

    for(arg = optind;arg < argc;arg++)
    {
        printf("Testing %s\n",argv[arg]);
    hash_gedcom_file(argv[arg]);
    init_use_hash(argv[arg]);
    fp_gedcom = fp_ReadHashedGedcomFile;
    fprintf(stderr,"Testing all indexes\n");
    tested = 0;
    while(find_next_level_zero(fp_gedcom, &readBuf[0]))
    {
        tested++;
        curLoc = ftell(fp_gedcom);
        cptr1 = strchr(readBuf,'@');
        cptr1++;
        cptr2 = strchr(cptr1,'@');
        if(cptr2 != NULL)
            *cptr2 = 0;
        if(!find_hash_entry(cptr1,&tagLoc))
        {
            fprintf(STDERROR,"Cannot find %s\n",cptr1);
            results = (hash(cptr1,strlen(cptr1)) % MAX_GED_HASH);
            fprintf(STDERROR,"Hash number was %ld\n",results);
            dumpRamFile2Disk("gedhash.dmp");
            exit(0);
        }
        fflush(fp_gedcom);
        fseek(fp_gedcom,tagLoc,SEEK_SET);
        readBuf2[0] = 0;
        readGedLine(fp_gedcom,readBuf2,sizeof(readBuf2));
        cptr1 = strchr(readBuf2,'@');
        cptr1++;
        cptr2 = strchr(cptr1,'@');
        if(cptr2 != NULL)
            *cptr2 = 0;
        if(strcmp(readBuf,readBuf2))
        {
            fprintf(STDERROR,"Error finding >%s<\n",cptr1);
            fprintf(STDERROR,"Expected\n%s\nActual\n%s\n",readBuf,readBuf2);
            fprintf(STDERROR,"Actual line length was %d\n",gedLineLength);
            fprintf(STDERROR,"tagLoc = 0x%lx\n",tagLoc);
            fprintf(STDERROR,"actual loc 0x%lx\n",cur_ged_line_start);
            exit(0);
        }
        else
        {
 /*           fprintf(stderr,"\r%s",cptr1); */
        }
        fseek(fp_gedcom,curLoc,SEEK_SET);
    }

    fprintf(STDERROR,"Testing rin file\n");
    for(rin = 1;rin < max_rin;rin++)
    {
        if(!find_rin_entry(rin, &tagLoc))
        {
            fprintf(STDERROR,"Error finding rin %u\n",rin);
            exit(0);
        }
        fseek(fp_gedcom,tagLoc,SEEK_SET);
        readBuf2[0] = 0;
        readGedLine(fp_gedcom,readBuf2,sizeof(readBuf2));
        strcpy(readBuf3,readBuf2);
        cptr1 = strchr(readBuf2,'@');
        cptr1++;
        cptr2 = strchr(cptr1,'@');
        if(cptr2 != NULL)
            *cptr2 = 0;
        if(!find_hash_entry(cptr1,&tagLoc))
        {
            fprintf(STDERROR,"Error 1 crossreference rin %d to hash\n",rin);
            fprintf(STDERROR,"Cannot find %s\n",cptr1);
            results = (hash(cptr1,strlen(cptr1)) % MAX_GED_HASH);
            fprintf(STDERROR,"Hash number was %ld\n",results);
            dumpRamFile2Disk("gedhash.dmp");
            exit(0);
        }
        fseek(fp_gedcom,tagLoc,SEEK_SET);
        readBuf[0] = 0;
        readGedLine(fp_gedcom,readBuf,sizeof(readBuf2));
        if(strcmp(readBuf,readBuf3))
        {
            fprintf(STDERROR,"Error 2 crossreference rin %d to hash\n",rin);
            fprintf(STDERROR,"Error finding >%s<\n",cptr1);
            fprintf(STDERROR,"Expected\n%s\nActual\n%s\n",readBuf3,readBuf);
            fprintf(STDERROR,"Actual line length was %d\n",gedLineLength);
            fprintf(STDERROR,"tagLoc = 0x%lx\n",tagLoc);
            fprintf(STDERROR,"actual loc 0x%lx\n",cur_ged_line_start);
            exit(0);
        }
    }
    fprintf(STDERROR,"Testing mrin file\n");
    for(rin = 1;rin < max_mrin;rin++)
    {
        if(!find_mRin_entry(rin, &tagLoc))
        {
            fprintf(STDERROR,"Error finding mRin %u\n",rin);
            exit(0);
        }
        fseek(fp_gedcom,tagLoc,SEEK_SET);
        readBuf2[0] = 0;
        readGedLine(fp_gedcom,readBuf2,sizeof(readBuf2));
        strcpy(readBuf3,readBuf2);
        cptr1 = strchr(readBuf2,'@');
        cptr1++;
        cptr2 = strchr(cptr1,'@');
        if(cptr2 != NULL)
            *cptr2 = 0;
        if(!find_hash_entry(cptr1,&tagLoc))
        {
            fprintf(STDERROR,"Error 1 crossreference mrin %d to hash\n",rin);
            fprintf(STDERROR,"Cannot find %s\n",cptr1);
            results = (hash(cptr1,strlen(cptr1)) % MAX_GED_HASH);
            fprintf(STDERROR,"Hash number was %ld\n",results);
            dumpRamFile2Disk("gedhash.dmp");
            exit(0);
        }
        fseek(fp_gedcom,tagLoc,SEEK_SET);
        readBuf[0] = 0;
        readGedLine(fp_gedcom,readBuf,sizeof(readBuf2));
        if(strcmp(readBuf,readBuf3))
        {
            fprintf(STDERROR,"Error 2 crossreference mrin %d to hash\n",rin);
            fprintf(STDERROR,"Error finding >%s<\n",cptr1);
            fprintf(STDERROR,"Expected\n%s\nActual\n%s\n",readBuf3,readBuf);
            fprintf(STDERROR,"Actual line length was %d\n",gedLineLength);
            fprintf(STDERROR,"tagLoc = 0x%lx\n",tagLoc);
            fprintf(STDERROR,"actual loc 0x%lx\n",cur_ged_line_start);
            exit(0);
        }
    }
    close_use_hash(argv[arg],deleteIndexFiles);
    fprintf(stdout,"Tested %d tags\n",tested);
    fprintf(stdout,"%d INDI tags\n",max_rin);
    fprintf(stdout,"%d FAM tags\n",max_mrin);
    }
    free_hash_buffers();
    deleteOverFlowFile();
}
