/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "pafsubs.h"
#include "pafutils.h"

#define EOS             '\0'
extern char *monthnames[];
extern RECORD_PTR      rin_max;

char *
delete_trailing_blanks(char *buf)
{
    char *ptr;
    ptr = buf + strlen(buf) - 1;
    while((*ptr == ' ') &&(ptr != buf))
      *ptr-- = 0;
    return(buf);
}
/* returns the address of the current date in the form mm/dd/yy */
char *
get_cur_date()
{
/* date to go on chart.  Format like : 16 Nov 1993  */
    static char buf[40]; /* where date will be built */
    char tmpbuf[40];
    struct tm *timeptr;
    time_t elapsed_seconds;
    time (&elapsed_seconds);
    timeptr = localtime(&elapsed_seconds);
    sprintf(&buf[0],"%d ",timeptr->tm_mday);
    strcat(buf,monthnames[timeptr->tm_mon+1]);
    sprintf(tmpbuf," 19%2d",timeptr->tm_year);
    strcat(buf,tmpbuf);
#if 0
    s
    sprintf(&buf[0],"%2d/%2d/%2d",
    	timeptr->tm_mon+1,
		timeptr->tm_mday,
		timeptr->tm_year);
    for(i=1;i < strlen(buf);i++)
    {
	if(buf[i] == ' ')
	    buf[i] = '0';
    }
#endif
    return(&buf[0]);
}


/* Returns NULL if no more data */
char *
readLine(char *buf,int maxlen,FILE *fp_in)
{
    int slen;
    char *status;
    buf[0] = buf[1] = EOS;
    if((status = fgets(buf, maxlen, fp_in)) == NULL)
        return status;
    if(buf[(slen=strlen(buf)) -1] == '\n')
        buf[slen -1 ] = EOS;
    buf[maxlen -1] = EOS;
    if(slen == 1)
        buf[slen] = EOS;
    return status;
}

/* Like strcpy except it doesn't copy the null terminator.  Overlays dest */
/* with src, up to a max size of n.  Watches dest, and if a terminating 0 */
/* is detected, it will terminate the string with 0 */
void
strnOvr(char *dest, char *src, int max_size)
{
    int i;
    int len;
    int term_it;

    term_it = 0;
    len = strlen(src);
    if(len > max_size) len = max_size;
    for(i=0;i < len;i++)
    {
        if(*dest == 0)
            term_it = 1;
        *dest++ = *src++;
    }
    if(term_it)
        *dest = 0;
}

void 
printnotes(int position,RECORD_PTR rin,FILE *fp_out)
{
#ifndef USE_GEDCOM
    INDIV_REC       indiv;
    NOTE_REC        pad;
    int             i;
    char            s1[100];

    get_indiv_rec(rin, &indiv);

    /* Only proceed if individual is not deleted and has a note. */
    if (indiv.sex != 'D' && (indiv.notepad != 0))
    {
        convert_name(s1, indiv.names);
        fprintf(fp_out,"\n%s [%d]",s1,position + 1);
        /* First note will insert newline */
        {
            int             llen;
            int             newline;
            RECORD_PTR      npad;
            llen = 0;
            newline = 0;
            npad = indiv.notepad;
            while (npad != 0)   /* Chain through notepads */
            {
                /* printf("[%d]",npad); */
                get_note_rec(npad, &pad);
                npad = pad.next_pad;
                for (i = 0; i < 254; i++) {
                    if (pad.notelines[i] == 0)
                        if (llen == 0) {
                            llen = 1;
                        } 
                        else if (llen == 1) {
                            llen = 0;
                            break;    /* Zero length line terminates notes */
                        } 
                        else {
                            llen = 0;
                            fprintf(fp_out,"\n  ");
                        }
                    else if (llen == 0 && pad.notelines[i] == 1) {
                        fprintf(fp_out,"\n  ");   /* Start of new note */
                        newline = 1;
                        llen++;
                    } 
                    else {
                        fprintf(fp_out,"%c",pad.notelines[i]);
                        llen++;
                    }
                }
            }
            if (!newline)
                fprintf(fp_out,"\n");   /* Just in case note was blank */
        }
    }
#endif /* USE_GEDCOM */
    return;
}

char *used_rin_file="rin_file.$$$";


void
erase_used_rin_file()
{
#ifdef unix
    unlink(used_rin_file);
#else
    remove(used_rin_file);
#endif /* DOS */
}

void
init_used_rin_file()
{
    FILE *fp_out;
	USED_RIN tmp;
	int i;

    fp_out = fopen(used_rin_file,"wb");

	tmp.chart_nbr = 0;
	tmp.from_position = 0;
	tmp.used = 0;

    if(fp_out == NULL)
    {
        printf("Unable to add to tmp file %s\n",used_rin_file);
        exit(-1);
    }
	for(i=0;i < rin_max ;i++)
	{
    	fwrite(&tmp,sizeof(struct USED_RIN),1,fp_out);
	}
    fclose(fp_out);
}

int
read_from_used_rin_file(USED_RIN *rin,RECORD_PTR person)
{
    FILE *fp_out;
    int status;
    fp_out = fopen(used_rin_file,"rb");
    if(fp_out == NULL)
    {
        printf("Unable to open tmp file %s\n",used_rin_file);
        exit(-1);
    }
    status = fseek(fp_out,person * sizeof(struct USED_RIN),SEEK_SET);
    if(!status)
    {
        status = fread(rin,sizeof(struct USED_RIN),1,fp_out);
        fclose(fp_out);
#ifdef DEBUG_PROG
        printf("Reading person %d used flag = %d in chart %d position %d\n",
             person,rin->used,rin->chart_nbr,rin->from_position);
#endif /* DEBUG_PROG */
        return 1;
    }
    else
    {
        printf("Seek error on %s, rin # %d\n",used_rin_file,person);
    }
    fclose(fp_out);
    return 0;
}

int
write_to_used_rin_file(USED_RIN *rin,RECORD_PTR person)
{
    FILE *fp_out;
    int status;
#ifdef DEBUG_PROG
                printf("Marking person %d used flag = %d chart %d position %d\n",
                person,rin->used,rin->chart_nbr,rin->from_position);
#endif /* DEBUG_PROG */
    fp_out = fopen(used_rin_file,"rb+");
    if(fp_out == NULL)
    {
        printf("Unable to open tmp file %s\n",used_rin_file);
        exit(-1);
    }
    status = fseek(fp_out,person * sizeof(struct USED_RIN),SEEK_SET);
    if(!status)
    {
        fwrite(rin,sizeof(struct USED_RIN),1,fp_out);
        fclose(fp_out);
        return 1;
    }
    fclose(fp_out);
    return 0;
}

char *cont_file="contfile.$$$";
int nbr_cont_file_entries = 0;

void
delete_file(char *name)
{
#ifdef unix
    unlink(name);
#else
    remove(name);
#endif /* DOS */
}

void
delete_cont_file()
{
#ifdef unix
    unlink(cont_file);
#else
    remove(cont_file);
#endif /* DOS */
}

void
erase_cont_file()
{
    FILE *fp_cont;
    nbr_cont_file_entries = 0;
    fp_cont = fopen(cont_file,"wb");
    if(fp_cont == NULL)
    {
        printf("Unable to open tmp file %s\n",cont_file);
        exit(-1);
    }
    fclose(fp_cont);
}

void
add_to_cont_file(CHART_START *cont)
{
    FILE *fp_cont;
    fp_cont = fopen(cont_file,"ab");
    nbr_cont_file_entries++;
    if(fp_cont == NULL)
    {
        printf("Unable to add to tmp file %s\n",cont_file);
        exit(-1);
    }
    fwrite(cont,sizeof(struct CHART_START),1,fp_cont);
    fclose(fp_cont);
}

/* returns 0 if no entry found */
int
read_from_cont_file(CHART_START *cont,int chart_nbr)
{
    FILE *fp_cont;
    int status;

    if(chart_nbr > (nbr_cont_file_entries -1))
      return 0;
    fp_cont = fopen(cont_file,"rb");
    if(fp_cont == NULL)
    {
        printf("Unable read chart continuation file %s\n",cont_file);
        exit(-1);
    }
    status = fseek(fp_cont,chart_nbr * sizeof(struct CHART_START),SEEK_SET);
    if(!status)
    {
        fread(cont,sizeof(struct CHART_START),1,fp_cont);
        fclose(fp_cont);
        return 1;
    }
    fclose(fp_cont);
    return 0;
}

char *convert_name(char *s, RECORD_PTR * ptrs)
{
  int             i;
  NAME_REC        namrec;
  s[0] = 0;
  
/*  {int flag;
   flag = 0;
   for(i=0; i<5; i++){
     if(ptrs[i] > 864) flag = 1;
   }
   if(flag) printf("Bad nam ids: %d %d %d %d %d\n",ptrs[0],ptrs[1],ptrs[2],ptrs[3]
	  ,ptrs[4]);
 }*/
  for (i = 1; i < 4; i++)
  { /* Print three given names */
    if (ptrs[i] != 0) {
      get_name_rec(ptrs[i], &namrec);
      strcat(s, namrec.name);
      strcat(s, " ");
    }
  }
  if (ptrs[0] != 0) {		/* Print last name */
    get_name_rec(ptrs[0], &namrec);
    strcat(s, namrec.name);
  }
  else
  {
    strcat(s, "????");
  }
  if (ptrs[4] != 0) {		/* Print title */
    get_name_rec(ptrs[4], &namrec);
    strcat(s, " ");
    strcat(s, namrec.name);
  }
  return(s);
}
/* just puts las name first. */
char *convert_name2(char *s, RECORD_PTR * ptrs)
{
  int             i;
  NAME_REC        namrec;
  char s1[256];
  s[0] = 0;
  s1[0] = 0;
  
/*  {int flag;
   flag = 0;
   for(i=0; i<5; i++){
     if(ptrs[i] > 864) flag = 1;
   }
   if(flag) printf("Bad nam ids: %d %d %d %d %d\n",ptrs[0],ptrs[1],ptrs[2],ptrs[3]
	  ,ptrs[4]);
 }*/
  for (i = 1; i < 4; i++)
  { /* Print three given names */
    if (ptrs[i] != 0) {
      get_name_rec(ptrs[i], &namrec);
      strcat(s1, namrec.name);
      strcat(s1, " ");
    }
  }
  if (ptrs[0] != 0) {		/* Print last name */
    get_name_rec(ptrs[0], &namrec);
    strcat(s, namrec.name);
    strcat(s, " ");
    strcat(s,s1);
  }
  else
  {
    strcat(s, "????");
    strcat(s, " ");
    strcat(s,s1);
  }
  if (ptrs[4] != 0) {		/* Print title */
    get_name_rec(ptrs[4], &namrec);
    strcat(s, " ");
    strcat(s, namrec.name);
  }
  return(s);
}

char *convert_place(char *s, RECORD_PTR *ptrs)
{
#ifndef USE_GEDCOM
  int i;
  NAME_REC namrec;
  int empty;

  s[0] = 0;
  empty = TRUE;
  for(i=0; i<4; i++) {
    if(ptrs[i] != 0) {
      get_name_rec(ptrs[i], &namrec);
      strcat(s, namrec.name);
      empty = FALSE;
    }
    if(i<3) strcat(s,",");
  }
  i=strlen(s);
  while(s[--i] == ',') s[i] = 0;
/*  if(empty) s[0]=0;*/
#endif /* USE_GEDCOM */
  return(s);
}
#if 0
char *monthnames[]={"NOMONTH","January","February","March","April",
		      "May","June","July","August","September",
		      "October","November","December","SEENOTES",
		      "SUBMITTED","CLEARED","UNCLEARED","COMPLETED",
		      "CANCELLED","STILLBORN","INFANT","DNS/CAN",
		      "DNS","BIC","CHILD"};
#else
char *monthnames[]={"NOMONTH","Jan","Feb","Mar","Apr",
              "May","Jun","Jul","Aug","Sep",
              "Oct","Nov","Dec","SEENOTES",
		      "SUBMITTED","CLEARED","UNCLEARED","COMPLETED",
		      "CANCELLED","STILLBORN","INFANT","DNS/CAN",
		      "DNS","BIC","CHILD"};
#endif

char *convert_date(char *s,LONGDATE packed_date)
{
#ifndef USE_GEDCOM
  int year,month,day,dual;
  int i;
  int temp;
  
  if(packed_date == 0)
  {
    s[0] = 0;
  }
  else
  {
    temp = (int)(packed_date >> 16) & 0x3;
/*    switch((packed_date & 0x30000)>>16) */
    switch(temp)
      {
      case 0:
	strcpy(s,"BEF ");
	i = 4;
	break;
      case 1:
	strcpy(s,"ABT ");
	i = 4;
	break;
      case 2:
	s[0] = 0;
	i = 0;
	break;
      case 3:
 	strcpy(s,"AFT ");
	i = 4;
	break;
      }
    year = (int)(packed_date & 0xFF) << 4;
    year += (int)(packed_date >> 12) & 0xF;
    month = (int)((packed_date & 0xF00) >> 7);
    month += (int)((packed_date & 0x800000) >>23);
    day = (int)((packed_date & 0x7C0000) >> 18);
    if(month==0&&day==0)
    {
      if(year!=0)
      {
    if(day==0)
    {
	  sprintf(s+i,"%d",year);
    }
    else
    {
      sprintf(s+i,"%2d %d",day,year);
	}
      }
    }
    else
    {
      if(day==0)
      {
	sprintf(s+i,"%s %d",monthnames[month],year);
      }
      else
      {
    sprintf(s+i,"%2d %s %d",day, monthnames[month],year);
      }
    }
    if((dual=(int)((packed_date & 0xFF000000)>>24)) != 0)
    {
      i = strlen(s);
      sprintf(s+i,"/%d",year+dual);
    }
  }
#endif /* USE_GEDCOM */
  return(s);
}

