/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef SUNOS4_1_2
#include <unistd.h>
#endif
#include "doswld.h"
#include "gedhash.h"
#include "ramfile.h"
#include "gedsubs.h"
#include "gedutils.h"
#include "printer.h"
#include "fsort.h"

#define STDERROR stdout

#ifndef sys_errlist
extern char *sys_errlist[];
#endif

#ifdef NEED_GETOPT
extern getopt(int nargc,char **nargv,char *ostr);
extern char *optarg;
extern int optind;
#endif
int showMarriages = 1;
int showBirths = 1;
int showDeaths = 1;
int showIfNoPlace = 1;
int showIfNoDate = 1;
int linesPerPage = 55;
int sortType = 0;
char chart_date[80];

char *tmp_index_file = "index.$$$";
FILE *fp_tmpNdx;

#ifndef max
#define	max(a, b) 		((a) < (b) ? (b) : (a))
#endif
#ifndef min
#define	min(a, b) 		((a) > (b) ? (b) : (a))
#endif
#ifndef abs
#define	abs(x)			((x) >= 0 ? (x) : -(x))
#endif

void
usage(char *progname)
{
    printf("%s -[MBDdpal] gedcomFileName\n",progname);
    printf("Generates a report designed to maximize vital record research\n");
    printf("results.\n");
    printf("Generates a one line report for each birth, death, and marriage\n");
    printf("In State, County, City - Year Month Day order and sorts it.\n");
    printf("This has the effect of showing you all the people you should check\n");
    printf("when you access vital records and census records.\n");
    printf("Options:\n");
    printf("-a alternate sort style. Place, surname, then all else\n");
    printf("-M Don\'t show marriage records\n");
    printf("-B Don\'t show birth records\n");
    printf("-D Don\'t show death records\n");
    printf("-d Don\'t show lines with no date\n");
    printf("-p Don\'t show lines with no place\n");
    printf("-l val Lines per page.  Currently %d\n",linesPerPage);
    printf("-r Remove index files created for the gedcom file\n");

}


int
getSurnameLen(char FAR *s1, char FAR *s2)
{
    int i;
    int len1 = 0;
    int len2 = 0;
    /* search for first space or end of string */
    for(len1=0;(s1[len1] != 0) && (s1[len1] != ' '); len1++){;}

    for(len2=0;(s2[len2] != 0) && (s2[len2] != ' '); len2++){;}

    i = min(len1,len2);
    return(i + 1);
}

int
cmpNdxStrings(char FAR **s1,char FAR **s2)
{
    int status;
    int len;
    char FAR * ptr1;
    char FAR * ptr2;

    memcpy(&ptr1,s1,sizeof(char FAR *));
    memcpy(&ptr2,s2,sizeof(char FAR *));
#if 0
    status = STRNICMP((char FAR *)&ptr1[32],
        (char FAR *)&ptr2[32],100);
    if(status)
        return(status);
#endif
    switch (sortType)
    {
        case 0: /* sort on place date */
            return(STRICMP((char FAR *)ptr1,(char FAR *)ptr2));
        break;
        case 1: /* sort on place, only */
        status = STRNICMP((char FAR *)&ptr1[0],
            (char FAR *)&ptr2[0],34);
        if(status)
            return(status);

        /* now sort on just the surname */
        len = getSurnameLen(&ptr1[50], &ptr2[50]);
        status = STRNICMP((char FAR *)&ptr1[50],
            (char FAR *)&ptr2[50],len);
        if(status)
            return(status);

        /* This one gets the date and everything else */
        return(STRICMP((char FAR *)ptr1,(char FAR *)ptr2));
        break;
    }
    /* check everything else */
}
char *dashes = "------------------------------------------------------";
void
main(int argc, char **argv)
{
    int i;
    FILE *fp_gedcom;
    INDIV_REC       indiv;
    MARR_REC marriage_record;
    int deleteIndexFiles;
    char nameBuf[256];
    char placeBuf[256];
    char dateBuf[256];
    fsort_args sparms;
    int status;
    int c;
    int lineCount;
    int arg;
    unsigned int rin;
    unsigned int tested;

    showHashDetails = 0;
    disableRamFile = 1;
    deleteIndexFiles = 0;
    dosWildCard(&argc,&argv);  /* DOS wild card expander. Just returns if unix */

    strcpy(chart_date,get_cur_date());
    while ((c = getopt(argc, argv, "DBMdpl:ar")) != EOF)
        switch(c)
        {
                case 'D':
                    showDeaths = 0;
                break;
                case 'B':
                    showBirths = 0;
                break;
                case 'M':
                    showMarriages = 0;
                break;
                case 'd':
                    showIfNoDate = 0;
                break;
                case 'p':
                    showIfNoPlace = 0;
                break;
                case 'a':
                    sortType = 1;
                break;
                case 'l':
                    linesPerPage = atoi(optarg);
                break;
                case 'r':
                deleteIndexFiles = 1;
                break;
                default:
                     usage(argv[0]);
                     exit(-1);
                     break;
        }

    if((optind + 1) > argc)
    {
        usage(argv[0]);
        exit(-1);
    }

    fp_tmpNdx = fopen (tmp_index_file,"w");
    if(!fp_tmpNdx)
    {
        fprintf(stderr,"Unable to open temporary index file. Aborting\n");
        exit(-1);
    }
    alloc_hash_buffers();
    for(i = optind;i < argc;i++)
    {
    fprintf(stderr,"Indexing %s file\n",argv[i]);
    hash_gedcom_file(argv[i]);
    init_use_hash(argv[i]);
    rin_max = max_rin; /* init my stuff from hash_gedcom_file */
    fp_gedcom = fp_ReadHashedGedcomFile;
    tested = 0;
    rin = 0;
    while(rin < rin_max)
    {
        rin++;
        get_indiv_rec(rin, &indiv);
        convert_name2(nameBuf, &indiv.names,24);
        if(showBirths)
        {
        convert_date_reversed(dateBuf,&indiv.birth);
        convert_place_reversed(placeBuf,&indiv.birth,35);
        delete_trailing_blanks(placeBuf);
        delete_trailing_blanks(dateBuf);
        if((showIfNoPlace || placeBuf[0]) && (showIfNoDate || dateBuf[0]))
            fprintf(fp_tmpNdx,"%-34.34s %-12.12s B %-24.24s %d\n",
                placeBuf,dateBuf,nameBuf,rin);
        }
        if(showDeaths)
        {
        convert_date_reversed(dateBuf,&indiv.death);
        convert_place_reversed(placeBuf,&indiv.death,35);
        delete_trailing_blanks(placeBuf);
        delete_trailing_blanks(dateBuf);
        if((showIfNoPlace || placeBuf[0]) && (showIfNoDate || dateBuf[0]))
           fprintf(fp_tmpNdx,"%-34.34s %-12.12s D %-24.24s %d\n",
                placeBuf,dateBuf,nameBuf,rin);
        }
        if(showMarriages)
        {
        get_marr_rec(indiv.ownmarriage,&marriage_record);
        convert_date_reversed(dateBuf,&marriage_record.married);
        convert_place_reversed(placeBuf,&marriage_record.married,35);
        delete_trailing_blanks(placeBuf);
        delete_trailing_blanks(dateBuf);
        if((showIfNoPlace || placeBuf[0]) && (showIfNoDate || dateBuf[0]))
            fprintf(fp_tmpNdx,"%-34.34s %-12.12s M %-24.24s %d\n",
                placeBuf,dateBuf,nameBuf,rin);
        }
        fprintf(stderr,"\rWorking on %32.32s",nameBuf);
    }
    close_use_hash(argv[i],deleteIndexFiles);
    fprintf(stderr,"\nprocessed %d rins\n",rin);
    }
    free_hash_buffers();
    deleteOverFlowFile();
    fclose(fp_tmpNdx);
    fprintf(stderr,"Sorting index file\n");
    set_sort_file_defaults(&sparms);
    strcpy(sparms.inFileName,tmp_index_file);  /* File to sort */
    sprintf(sparms.outFileName,"index.$$$");
    sparms.deleteDuplicateRecords = 1; /* Delete Duplicate Records */
    sparms.compareRecord = cmpNdxStrings;
    delete_tmp_files = 1;       /* 0 = Leave tmp files for debugging */
    if(status = sort_file(&sparms))   /* do sort */
    {
        fprintf(stderr,"Sort failed.\n");
        fprintf(stderr,"error %d = %s\n",status,sys_errlist[status]);
        exit(-1);
    }

    lineCount = 4;
    fp_tmpNdx = fopen("index.$$$","r");
    /* create header line.  Figure how much to padd left margin to center it.*/
    sprintf(nameBuf,"Place Date Report %s",chart_date);
    c = strlen(nameBuf);
    for(i=0;i < (39 - (c/2));i++)
        fprintf(stdout," ");
    fprintf(stdout,"%s\n",nameBuf);
    fprintf(stdout,
    "The Event letter between the Date and Name field are B M or D, which stand\n");
    fprintf(stdout,"for Born, Married or Died.\n");
    fprintf(stdout,"%-34.34s %-12.12s E %-24.24s %-4.4s\n",
                "Place","Date","Name","Rin");
    fprintf(stdout,"%-34.34s %-12.12s - %-24.24s %-4.4s\n",
    dashes,dashes,dashes,dashes);
    while(readLine(placeBuf,sizeof(placeBuf),fp_tmpNdx))
    {
        lineCount++;
        if(lineCount >= linesPerPage)
        {
            form_feed(stdout);
            fprintf(stdout,"%-34.34s %-12.12s E %-24.24s %-4.4s\n",
                "Place","Date","Name","Rin");
            fprintf(stdout,"%-34.34s %-12.12s - %-24.24s %-4.4s\n",
                dashes,dashes,dashes,dashes);
            lineCount = 2;
        }
        fprintf(stdout,"%s\n",placeBuf);
    }
    form_feed(stdout);
    delete_file("index.$$$");
    delete_file(tmp_index_file);
}
