/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "gedrtns.h"
#include "path.h"
#include "printer.h"
#include "cfgfile.h"
#if 0
          The report is designed for compressed print (17 characters per
          line) with eight lines per inch.

          The codes used for the dot matrix printer are:
               ESC SI - to turn on compressed print
               ESC 0  - to turn on 1/8-inch line spacing
               ESC @  - to reset the printer to its normal settings

          The codes used for the laser printer are:
               ESC (8U ESC (s0p16.66h8.5v0s0b0T - to select the line
                                                  printer font
               ESC &l8D - to select 1/8-inch line spacing
               ESC E    - to reset the printer to its normal settings
#endif

#define ESCAPE 0x1b

int allowPrintFormatting = 1;
int allowFormFeeds = 1;
int printerType = PRINTER_TYPE_EPSON_GENERIC;
char formFeed[25];

/* pass in the paths to search. Typically a specified directory and the*/
/* DOS path, and it will search for the printer config file, and build */
/* the control strings needed to print the reports. */
int
printer_config(char *programName)
{
    char configFileName[256];
    if(PathString == NULL)
        getPathString();

    cleanup_file_name(programName,configFileName,1);
    strcat(configFileName,".cfg");
    if(!parse_file_4_variables(configFileName,"formFeed",sizeof(formFeed),
        formFeed,0))
        {
            formFeed[0] = 0xc;
            formFeed[1] = 0;
        }
}

int
out_to_ptr(FILE *fp_out,int len, char *str)
{
    int i;
    for(i=0;i < len;i++)
        fprintf(fp_out,"%c",*str++);
    return 1;
}

void
condensed_print(FILE *fp_out)
{
    char buf[80];
    if(allowPrintFormatting)
    {
    switch(printerType)
    {
        case PRINTER_TYPE_EPSON_GENERIC:
            buf[0] = 0xf;
            out_to_ptr(fp_out,1,&buf[0]);
            break;
        case PRINTER_TYPE_HP_GENERIC:
            /* select 6 lines per inch */
            sprintf(buf,"%c&l6D",ESCAPE);
            out_to_ptr(fp_out,strlen(buf),&buf[0]);
            /* select roman 8 font */
            sprintf(buf,"%c(8U",ESCAPE);
            out_to_ptr(fp_out,strlen(buf),&buf[0]);
            /* Fixed spacing */
            sprintf(buf,"%c(s0P",ESCAPE);
            out_to_ptr(fp_out,strlen(buf),&buf[0]);
#if 1
            /* 16.6 pitch */
            sprintf(buf,"%c&k2S",ESCAPE);
#else
            /* 12 pitch */
            sprintf(buf,"%c&k4S",ESCAPE);
#endif
            out_to_ptr(fp_out,strlen(buf),&buf[0]);
            break;
    }
    }
}

void
standard_print(FILE *fp_out)
{
    char buf[80];
    if(allowPrintFormatting)
    {
    switch(printerType)
    {
        case PRINTER_TYPE_EPSON_GENERIC:
            buf[0] = 0x12;
            out_to_ptr(fp_out,1,&buf[0]);
            break;
        case PRINTER_TYPE_HP_GENERIC:
            /* select 6 lines per inch */
            sprintf(buf,"%c&l6D",ESCAPE);
            out_to_ptr(fp_out,strlen(buf),&buf[0]);
            /* select roman 8 font */
            sprintf(buf,"%c(8U",ESCAPE);
            out_to_ptr(fp_out,strlen(buf),&buf[0]);
            /* Fixed spacing */
            sprintf(buf,"%c(s0P",ESCAPE);
            out_to_ptr(fp_out,strlen(buf),&buf[0]);
            /* 10.0 pitch */
            sprintf(buf,"%c&k0S",ESCAPE);
            out_to_ptr(fp_out,strlen(buf),&buf[0]);
            break;
    }
    }
}

void
form_feed(FILE *fp_out)
{
    char buf[10];
    if(allowFormFeeds)
    {
    buf[0] = 0xc;
    out_to_ptr(fp_out,1,&buf[0]);
    }
}
void
show_allowed_printer_types()
{
    printf("Allowed printer types are:\n");
    printf("0 = Epson\n");
    printf("1 = HP\n");
}
void
show_printer_type()
{
    printf("Current printer type is ");
    switch(printerType)
    {
        case PRINTER_TYPE_EPSON_GENERIC:
            printf("Epson\n");
            break;
        case PRINTER_TYPE_HP_GENERIC:
            printf("HP\n");
            break;
    }
}
