/*
 * Copyright (C) 1995, John H. Smith.  All rights reserved.
 *
 * This file is part of ged2htm.
 *  
 * These programs are free software; you can redistribute them and/or modify
 * them under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * The GNU General Public License can be found in the file LICENSE.
 * 
 * The author of ged2htm is John Smith.  He may be reached on internet
 * at the address johns@FileNet.com, or via mail at 28032 Singleleaf,
 * Mission Viejo, Ca. 92692.
 * 
 * ------------
 * 
 *This file contains subroutines used by the various ged2htm programs.
 *
 * Last modified:
 *    4/10/95
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#ifdef SUNOS4_1_2
#include <unistd.h>
#endif
#include "fsort.h"
#include "portable.h"
#define NO_RAM_FILE_MACROS
#include "ramfile.h"

#define TRY_IT 1

/* a rather simple replacement for fopen, fclose, fread, fwrite, fseek, */
/* ftell that uses far memory as much as possible, and overflows into */
/* a disk file if far memory is exausted. */

int disableRamFile = 0;

#ifndef max
#define	max(a, b) 		((a) < (b) ? (b) : (a))
#endif
#ifndef min
#define	min(a, b) 		((a) > (b) ? (b) : (a))
#endif
#ifndef abs
#define	abs(x)			((x) >= 0 ? (x) : -(x))
#endif

#define TEST_RAMFILE 0

#if TEST_RAMFILE
    size_t ram_buffer_size;
    unsigned char FAR *Pointers1;
    unsigned char FAR *Pointers2;
    char FAR *sanity;

void
sanity_check()
{
    sanity = (char FAR *)Pointers1;
    if(sanity[ram_buffer_size -1] != 0x55)
    {
        fprintf(stderr,"Pointers1 corrupted\n");
        exit(-1);
    }
    sanity = (char FAR *)Pointers2;
    if(sanity[ram_buffer_size -1] != 0x55)
    {
        fprintf(stderr,"Pointers2 corrupted\n");
        exit(-1);
    }
}
#endif /* TEST_RAMFILE */

#define FILE_DEBUG 0
#if FILE_DEBUG
static long realFileOffset = 0;
#endif

int nbrRamFarBuffers;
char FAR *ramFarBuffers[MAX_RAM_BUFFERS];
unsigned int ramFarBufferSizes[MAX_RAM_BUFFERS];

static long ramBufMaxOffset[MAX_RAM_BUFFERS];
static long ramBufMinOffset[MAX_RAM_BUFFERS];
static long maxRamBufferOffset;

static long file_position;
static long file_size = 0;
static int file_open = 0;
static int intoOverflow = 0;
static int curRamBuffer;
static long offsetThisBuf;
static int ramFileOpenType; /* 0 for read, 1 for write */

static char registeredFileName[FILENAME_MAX];
static FILE *fp_registered;
static char *openedType;

static char *overFlowFile = "ramfile.ovf";
static FILE *fp_over;


#if TEST_RAMFILE
static int show_stuff = 0;
static void
show_changing_stuff(int line)
{
    int i;
    i = line; /* to shut up compiler */
#if 1
    if(show_stuff)
    {
    printf("Line %d\n",line);
    printf("file_position 0x%lx\n",file_position);
    printf("file_size 0x%lx\n",file_size);
    printf("intoOverflow %d\n",intoOverflow);
    printf("curRamBuffer %d\n",curRamBuffer);
    printf("offsetThisBuf 0x%lx\n",offsetThisBuf);
#if FILE_DEBUG
    printf("Overflow file offset 0x%lx\n",realFileOffset);
#endif
    for(i=0;i < 5;i++)
    {
    printf("%c ",ramFarBuffers[curRamBuffer][offsetThisBuf + (i - 2)]);
    }
    printf("\n");
    fflush(stdout);
    }
#endif
}

static void
show_static_stuff()
{
#if 1
    int i;
    for(i = 0;i < nbrRamFarBuffers;i++)
    {
    printf("ramBufMaxOffset[%d] 0x%lx\n",i,(long)ramBufMaxOffset[i]);
    printf("ramBufMinOffset[%d] 0x%lx\n",i,(long)ramBufMinOffset[i]);
    printf("ramFarBufferSizes[%d] 0x%lx\n",i,(long)ramFarBufferSizes[i]);
    }
    fflush(stdout);
#endif
}
#endif /*  TEST_RAMFILE */

void
calcBufferOffset4seeks()
{
    int i;
    intoOverflow = 0;
    for(i=0;i < nbrRamFarBuffers;i++)
    {
        if((file_position >= ramBufMinOffset[i]) &&
           (file_position <= ramBufMaxOffset[i]))
        {
            curRamBuffer = i;
            offsetThisBuf = file_position - ramBufMinOffset[i];
#if 0
            if(i)
              offsetThisBuf++;
#endif
            return;
        }
    }
    intoOverflow = 1;
}

FILE * diskFile2RamFile(char *fileName,char *type)
{
#if TRY_IT
    FILE *fp_read;
    struct stat sbuf;
    int status;
    long fileSize;
    long sizeCopied;
    char buf[512];
    /* got to read the file from disk into ramfile */
    if(stat(fileName, &sbuf) == 0)
    {
        fileSize = sbuf.st_size;
    }
    else
        return(NULL); /* File not there probably */

    if((fp_read = fopen(fileName,FOPEN_RB)) == NULL)
    {
        /* normal file not there error, just return the NULL */
        return(fp_read);
    }
    fp_over = fopen(overFlowFile,FOPEN_WB);
    if(fp_over == NULL)
    {
        fprintf(stderr,"Cant open overflow file %s\n",overFlowFile);
        exit(-1);
    }

    sizeCopied = 0;
    while((sizeCopied + sizeof(buf)) < fileSize)
    {
        status = fread(buf, sizeof(buf), 1, fp_read);
        if(status == 0)
        {
            fprintf(stderr,"Error reading %s\n",fileName);
            exit(-1);
        }
        ram_fwrite(buf, sizeof(buf), 1, (FILE *)fp_registered);
        sizeCopied += sizeof(buf);
    }
    /* get the last odd bytes */
    status = fread(buf, (int)(fileSize - sizeCopied), 1, fp_read);
    if(status == 0)
    {
        fprintf(stderr,"Error reading last bytes of %s\n",fileName);
        exit(-1);
    }
    ram_fwrite(buf, 1, 1, (FILE *)fp_registered);
    fclose(fp_read);
    fclose(fp_over);
#endif
}

void
registerNameWithRamfile(char *fileName)
{
    registeredFileName[0] = 0;
    if(!disableRamFile)
        strcpy(registeredFileName,fileName);
}

FILE *ram_fopen(char *fileName, char *type)
{
    int i;
    long off;
    FILE *fp_returned;

    if(strcmp(registeredFileName,fileName))
    {
        return(fopen(fileName,type));
    }
    if(file_open)
    {
        fprintf(stderr,"RamFile already open\n");
        exit(-1);
    }

    openedType = type;

    off = 0;
    for(i=0;i < nbrRamFarBuffers;i++)
    {
        ramBufMinOffset[i] = off;
        off += (ramFarBufferSizes[i] - 1);
        ramBufMaxOffset[i] = off;
        maxRamBufferOffset = off;
        off++;
    }

    ramFileOpenType = 0;
    if(strchr(type,'r'))
    ramFileOpenType = 1;
    if(strchr(type,'w'))
    {
    ramFileOpenType = 2;
    file_size = 0;
    }
    if(strchr(type,'+'))
     ramFileOpenType = 3;
    file_position = 0;
    curRamBuffer = 0;
    offsetThisBuf = 0;
    file_open = 1;
    intoOverflow = 0;
#if TRY_IT
    if(type[0] == 'r')
    {
        diskFile2RamFile(fileName,type);
        file_position = 0;
        curRamBuffer = 0;
        offsetThisBuf = 0;
        file_open = 1;
        intoOverflow = 0;
    }
#endif
    fp_over = fopen(overFlowFile,type);
    if(fp_over == NULL)
    {
        fprintf(stderr,"Cant open overflow file %s\n",overFlowFile);
        exit(-1);
    }
    fp_returned = (FILE *)&fp_registered;
    fp_registered = fp_returned;
    return((FILE *)fp_returned);
}

int ram_fclose(FILE *fp)
{

    if(fp_registered != fp)
    {
        return(fclose(fp));
    }
    if(!file_open)
    {
        fprintf(stderr,"RamFile Was not already open\n");
        exit(-1);
    }
    file_size = max(file_position,file_size);
    file_open = 0;
    fclose(fp_over);
    return(0);
}

size_t
overWrite(void *ptr, size_t size, size_t nitems, FILE *fp)
{
    long did;
    fwrite(ptr,size,nitems,fp_over);
    did = size;
    did *= nitems;
    file_position += did;
    file_size = max(file_position,file_size);
#if FILE_DEBUG
    realFileOffset = ftell(fp_over);
#endif
    return(size);
}

size_t
overRead(void *ptr, size_t size, size_t nitems, FILE *fp)
{
    long did;
    fread(ptr,size,nitems,fp_over);
    did = size;
    did *= nitems;
    file_position += did;
    file_size = max(file_position,file_size);
#if FILE_DEBUG
    realFileOffset = ftell(fp_over);
#endif
    return(size);
}

/* returns size, or 0 for eof and error */
size_t
ram_fwrite(void *ptr, size_t size, size_t nitems, FILE *fp)
{
    size_t length;
    size_t thisLength;
    size_t i;
    char FAR *dptr;
    void *sptr;

    if(fp_registered != fp)
    {
        return(fwrite(ptr, size, nitems, fp));
    }
    for(i=0;i < nitems;i++)
    {
        length = size;
        sptr = ptr;
        while(length)
        {
            if(intoOverflow)
            {
                overWrite(sptr,length,1,fp_over);
                length = 0;
            }
            else
            {
        if((ramBufMaxOffset[curRamBuffer] +1) >= ((long)length + file_position))
        {
            dptr = ramFarBuffers[curRamBuffer] + offsetThisBuf;
            MEMMOVE((char FAR *)dptr,(char FAR *)sptr,(size_t) length);
            offsetThisBuf += (long)length;
            file_position += (long)length;
            length = 0;
        }
        else
        {
            /* we have to split it up */
            thisLength = (size_t)((ramBufMaxOffset[curRamBuffer] +1) - file_position);
            if(thisLength) /* Might be zero length */
            {
            dptr = ramFarBuffers[curRamBuffer] + offsetThisBuf;
            MEMMOVE((char FAR *)dptr,(char FAR *)sptr,(size_t) thisLength);
            file_position += (long)thisLength;
            length -= thisLength;
            offsetThisBuf += (long)thisLength;
            (char *)sptr += thisLength;
            } /* zero length */
            curRamBuffer++;
            offsetThisBuf = 0;
            if(curRamBuffer >= nbrRamFarBuffers)
                intoOverflow = 1;
        }
        }
        } /* length */
    }/* Nbr items */
    file_size = max(file_position,file_size);
    return(size);
}

/* returns size, or 0 for eof and error */
size_t
ram_fread(void *ptr, size_t size, size_t nitems, FILE *fp)
{
    size_t length;
    size_t thisLength;
    long bytesRead;
    size_t i;
    char FAR *dptr;
    void *sptr;
    bytesRead = 0;

    if(fp_registered != fp)
    {
        return(fread(ptr, size, nitems, fp));
    }
    for(i=0;i < nitems;i++)
    {
        length = size;
        sptr = ptr;
        if((file_position + length) > file_size)
            length = (size_t)(file_size - file_position);
        while(length)
        {
            if(intoOverflow)
            {
                bytesRead += overRead(sptr,length,1,fp_over);
                length = 0;
            }
            else
            {
        if((ramBufMaxOffset[curRamBuffer] +1) >= ((long)length + file_position))
        {
            dptr = ramFarBuffers[curRamBuffer] + offsetThisBuf;
            MEMMOVE((char FAR *)sptr,(char FAR *)dptr,(size_t) length);
            offsetThisBuf += (long)length;
            file_position += (long)length;
            bytesRead += length;
            length = 0;
        }
        else
        {
            /* we have to split it up */
            thisLength = (size_t)((ramBufMaxOffset[curRamBuffer] +1) - file_position);
            if(thisLength) /* Might be zero length */
            {
            dptr = ramFarBuffers[curRamBuffer] + offsetThisBuf;
            MEMMOVE((char FAR *)sptr,(char FAR *)dptr,(size_t) thisLength);
            offsetThisBuf += (long)thisLength;
            file_position += (long)thisLength;
            length -= thisLength;
            (char *)sptr += thisLength;
            bytesRead += thisLength;
            }
            curRamBuffer++;
            offsetThisBuf = 0;
            if(curRamBuffer >= nbrRamFarBuffers)
                intoOverflow = 1;
        }
        }
        }
    }
    return((size_t)bytesRead / size);
}

int
overSeek(FILE *fp,long offset, int ptrname)
{
    switch(ptrname)
    {
        case SEEK_SET:
                fseek(fp_over,(offset - (maxRamBufferOffset + 1)),ptrname);
           break;
        case SEEK_CUR:
                fseek(fp_over,(offset),SEEK_CUR);
           break;
        case SEEK_END:
                fseek(fp_over,(offset),SEEK_END);
           break;
    }
#if FILE_DEBUG
    realFileOffset = ftell(fp_over);
#endif
    return(0);
}
/* returns 0 if ok, or -1 for errors */
int
ram_fseek(FILE *fp, long offset, int ptrname)
{
    if(fp_registered != fp)
    {
        return(fseek(fp, offset, ptrname));
    }
    switch(ptrname)
    {
        case SEEK_SET:
           file_position = offset;
           if(file_position > file_size)
              file_position = file_size;
           if(file_position < 0)
             file_position = 0;
            calcBufferOffset4seeks();
            if(intoOverflow)
                overSeek(fp,file_position, SEEK_SET);
           break;
        case SEEK_CUR:
           file_position = file_position + offset;
           if(file_position > file_size)
              file_position = file_size;
           if(file_position < 0)
             file_position = 0;
            calcBufferOffset4seeks();
            if(intoOverflow)
                overSeek(fp,file_position, SEEK_SET);
           break;
        case SEEK_END:
           file_position = (file_size) + offset;
           if(file_position < 0)
             file_position = 0;
           if(file_position > (file_size))
             file_position = (file_size);
            calcBufferOffset4seeks();
            if(intoOverflow)
                overSeek(fp,file_position, SEEK_SET);
           break;
    }
    return(0);
}

long ram_ftell(FILE *fp)
{
    if(fp_registered != fp)
    {
        return(ftell(fp));
    }
    return(file_position);
}

void
deleteOverFlowFile()
{
    int status;
#ifndef DOS
    status = unlink(overFlowFile);
#else
    status = remove(overFlowFile);
#endif /* DOS */
}

void
dumpRamFile2Disk(char *dumpFileName)
{
    FILE *fp_dump;
    FILE *fp_ram;
    char saveFileName[FILENAME_MAX];
    char buf[10];
    int wasOpen;
    long position;
    FILE *oldRegistered;

    if(disableRamFile)
        return;
    wasOpen = 0;
    if(file_open)
    {
        wasOpen = 1;
        oldRegistered = fp_registered;
        position = file_position;
        ram_fclose(fp_registered);
    }
    strcpy(saveFileName,registeredFileName);
    registerNameWithRamfile("ramfile.ts2");
    fp_ram = ram_fopen("ramfile.ts2", FOPEN_RB);
    fp_dump = fopen(dumpFileName,FOPEN_WB);
    if(fp_dump == NULL)
    {
        fprintf(stderr,"Unable to open ramfile dump file %s\n",dumpFileName);
        exit(0);
    }
    while(ram_fread(&buf[0],1,1,fp_ram))
    {
       fwrite(&buf[0],1,1,fp_dump);
    }
    fclose(fp_dump);
    ram_fclose(fp_ram);
    /* try and leave it as it was */
    if(wasOpen)
    {
        registerNameWithRamfile(saveFileName);
        fp_ram = ram_fopen(saveFileName,openedType);
        ram_fseek(fp_ram,position,SEEK_SET);
    }
}

#if TEST_RAMFILE
void
testRamFile()
{
    FILE *fp_test;
    FILE *fp_ram;
    int i;
    int c;
    int count;
    int byte;
    long seekOffset;
    long curLoc1;
    long curLoc2;
    long biggest;
    char buf[256];
    char buf2[256];
    int readWriteSize;
    int testFileSize;

    fprintf(stdout,"\n");
    sanity_check();
    registerNameWithRamfile("ramfile.ts2");
    fp_ram = ram_fopen("ramfile.ts2", FOPEN_WB);
    show_static_stuff();
    ram_fclose(fp_ram);
    for(testFileSize = 10;testFileSize < 20;testFileSize++)
    {
        for(readWriteSize = 1;readWriteSize < 99;readWriteSize++)
        {
            fprintf(stdout,"testFileSize %d, readWriteSize %d\n",
                testFileSize,readWriteSize);
            fprintf(stderr,"testFileSize %d, readWriteSize %d\n",
                testFileSize,readWriteSize);
    /* Create a test file on hard disk */
    printf("Creating test file\n");
    fp_test = ram_fopen("ramfile.ts1",FOPEN_WB);
    for(i=0;i < testFileSize;i++)
      for(c = '0';c <= '9';c++)
        fwrite(&c,1,1,fp_test);
    fprintf(stdout,"test file written size %ld\n",ftell(fp_test));
    ram_fclose(fp_test);

    /* because we are reading with large chunks, we won't get the whole file*/
    /* see how big we will get */
    fp_test = ram_fopen("ramfile.ts1",FOPEN_RB);
    while(ram_fread(&buf[0],readWriteSize,1,fp_test))
        biggest = ftell(fp_test);
    printf("Biggest I could read with these sizes was %ld\n",biggest);
    fclose(fp_test);

    sanity_check();
    /* read disk file into ram file */
    printf("\nRead disk file into ram file\n");
    fp_ram = ram_fopen("ramfile.ts2", FOPEN_WB);
    fp_test = fopen("ramfile.ts1",FOPEN_WB);
    while(ram_fread(&buf[0],readWriteSize,1,fp_test))
      ram_fwrite(&buf[0],readWriteSize,1,fp_ram);
    fprintf(stdout,"Disk file size %ld\n",ftell(fp_test));
    fprintf(stdout,"Ram file size %ld\n",ram_ftell(fp_ram));
    fclose(fp_test);
    ram_fclose(fp_ram);

    sanity_check();
    /* read ram file into disk file */
    printf("\nRead ram file into disk file\n");
    fp_test = fopen("ramfile.ts3",FOPEN_WB);
    fp_ram = ram_fopen("ramfile.ts2", FOPEN_RB);
    count = 0;
    while(ram_fread(&buf[0],readWriteSize,1,fp_ram))
    {
       count++;
       fwrite(&buf[0],readWriteSize,1,fp_test);
       show_changing_stuff(__LINE__);
#if 0
        printf("Count %d\n",count);
        fprintf(stdout,"Disk file size %ld\n",ftell(fp_test));
        fprintf(stdout,"Ram file size %ld\n",ram_ftell(fp_ram));
#endif
    }
    fprintf(stdout,"Disk file size %ld\n",ftell(fp_test));
    fprintf(stdout,"Ram file size %ld\n",ram_ftell(fp_ram));

    fclose(fp_test);
    ram_fclose(fp_ram);
    sanity_check();

    printf("\nCompare two disk files\n");
    fp_test = fopen("ramfile.ts1",FOPEN_RB);
    fp_ram = fopen("ramfile.ts3",FOPEN_RB);
    byte = 0;
    while(fread(&buf[0],readWriteSize,1,fp_test))
    {
      if(!fread(&buf2[0],readWriteSize,1,fp_ram))
      {
        fprintf(stdout,"ram file to small.\n");
        show_stuff = 1;
        show_static_stuff();
        show_changing_stuff(__LINE__);
        fprintf(stdout,"Ramfile shows size as %ld\n",file_size);
        fprintf(stdout,"Line %d\n",__LINE__);
        exit(0);
      }

      if(memcmp(&buf[0], &buf2[0],readWriteSize))
      {
        fprintf(stdout,"\nData miss compare.  Line %d Disk data %c ram data %c\n",
         __LINE__,buf[0], buf2[0]);
        show_stuff = 1;
        show_static_stuff();
        show_changing_stuff(__LINE__);
        exit(0);
      }
      byte++;
    }
    fclose(fp_test);
    fclose(fp_ram);
#if 1
    printf("\nSeek test forward\n");
    fp_test = fopen("ramfile.ts3",FOPEN_RB);
    fp_ram = ram_fopen("ramfile.ts2", FOPEN_RB);
    count = 0;
    show_stuff = 0;
    show_changing_stuff(__LINE__);
    while(fread(&buf[0],readWriteSize,1,fp_test))
    {
      if(count == 62)
        count = count; /* for breakpoints */
#if 0
    fprintf(stdout,"testFileSize %d, readWriteSize %d count %d\n",
                testFileSize,readWriteSize,count);
    if((testFileSize == 11) && (readWriteSize == 1) && (count == 126))
        count = count;   /* for breakpoints */
#endif
    show_changing_stuff(__LINE__);
      ram_fread(&buf2[0],readWriteSize,1,fp_ram);
    show_changing_stuff(__LINE__);
      if((curLoc1 = ftell(fp_test)) != (curLoc2 = ram_ftell(fp_ram)))
      {
        printf("Ftell error after seeks line %d  Disk loc 0x%lx, ram loc 0x%lx count %d\n",
        __LINE__,curLoc1, curLoc2,count);
        show_stuff = 1;
        show_static_stuff();
        show_changing_stuff(__LINE__);
        exit(0);
      }
#if 0
        fprintf(stderr,"\r%3.3d 0x%8.8lx",readWriteSize,curLoc1);
#endif
      if(memcmp(&buf[0], &buf2[0],readWriteSize))
      {
        fprintf(stdout,"\nData miss compare after seeks.\n");
        fprintf(stdout,"Line %d disk data %c ram data %c count %d\n",
         __LINE__,buf[0], buf2[0],count);
        curLoc1 = ftell(fp_test);
        curLoc2 = ram_ftell(fp_ram);
        printf("locs 0x%lx 0x%lx\n",curLoc1,curLoc2);
        show_stuff = 1;
        show_static_stuff();
        show_changing_stuff(__LINE__);
        exit(0);
      }
      else
      {
        /* printf("Good buf[0] %c\n",buf[0]); */
    show_changing_stuff(__LINE__);
      }
    seekOffset = (long)readWriteSize + ftell(fp_test);
    fseek(fp_test,seekOffset,SEEK_SET);
    ram_fseek(fp_ram,seekOffset,SEEK_SET);
    if((curLoc1 = ftell(fp_test)) != (curLoc2 = ram_ftell(fp_ram)))
    {
        if(!(curLoc1 > biggest))
        {  /* Ignore, ram seek won't go past file size */
        printf("Ftell error after seeks. Line %d Disk loc 0x%lx, ram loc 0x%lx count %d\n",
        __LINE__,curLoc1, curLoc2,count);
        printf("Biggest = 0x%lx\n",biggest);
        show_stuff = 1;
        show_static_stuff();
        show_changing_stuff(__LINE__);
        exit(0);
        }
    }
    show_changing_stuff(__LINE__);
    count++;
    }
    fclose(fp_test);
    ram_fclose(fp_ram);
    show_stuff = 0;
#endif
#if 1
    sanity_check();
    printf("\nSeek test cur Backwards\n");
    fp_test = fopen("ramfile.ts3",FOPEN_RB);
    fp_ram = ram_fopen("ramfile.ts2", FOPEN_RB);
    count = 0;
    show_stuff = 0;
    show_changing_stuff(__LINE__);
    fseek(fp_test,0L,SEEK_END);
    ram_fseek(fp_ram,0L,SEEK_END);
    show_changing_stuff(__LINE__);
    while(1)
    {
    count++;
#if 0
    fprintf(stdout,"testFileSize %d, readWriteSize %d count %d\n",
                testFileSize,readWriteSize,count);
    if((testFileSize == 1) && (readWriteSize == 11) && (count == 1))
        count = count;   /* for breakpoints */
#endif
    sanity_check();
    if((curLoc1 = ftell(fp_test)) != (curLoc2 = ram_ftell(fp_ram)))
    {
        printf("\nFtell error after seeks. Line %d Disk loc 0x%lx, ram loc 0x%lx count %d\n",
        __LINE__,curLoc1, curLoc2,count);
        show_stuff = 1;
        show_static_stuff();
        show_changing_stuff(__LINE__);
        exit(0);
    }
    sanity_check();
/*    fprintf(stderr,"\rLoc 0x%lx",curLoc1); */
    if(fseek(fp_test,-readWriteSize,SEEK_CUR)) /* error */
        break; /* out of while */
    ram_fseek(fp_ram,-readWriteSize,SEEK_CUR);

    sanity_check();
    show_changing_stuff(__LINE__);
    if((curLoc1 = ftell(fp_test)) != (curLoc2 = ram_ftell(fp_ram)))
    {
        printf("\nFtell error after seeks.  Line %d Disk loc 0x%lx, ram loc 0x%lx count %d\n",
        __LINE__,curLoc1, curLoc2,count);
        show_stuff = 1;
        show_static_stuff();
        show_changing_stuff(__LINE__);
        exit(0);
    }

      fread(&buf[0],1,1,fp_test);
      ram_fread(&buf2[0],1,1,fp_ram);
    sanity_check();
      if(memcmp(&buf[0], &buf2[0],1))
      {
        fprintf(stdout,"\nData miss compare after seeks.\n");
        fprintf(stdout,"Line %d disk data %c ram data %c count %d\n",
         __LINE__,buf[0], buf2[0],count);
        curLoc1 = ftell(fp_test);
        curLoc2 = ram_ftell(fp_ram);
        printf("locs 0x%lx 0x%lx\n",curLoc1,curLoc2);
        show_stuff = 1;
        show_static_stuff();
        show_changing_stuff(__LINE__);
        exit(0);
      }
    show_changing_stuff(__LINE__);
    sanity_check();
      fseek(fp_test,-1,SEEK_CUR); /* undo the read */
      ram_fseek(fp_ram,-1,SEEK_CUR);
    sanity_check();
    show_changing_stuff(__LINE__);
      if((curLoc1 = ftell(fp_test)) == 0)
        break; /* out of while loop */
    }
    fclose(fp_test);
    ram_fclose(fp_ram);
    show_stuff = 0;
#endif
    sanity_check();





#if 1
    printf("\nSeek CUR forward\n");


    fp_test = fopen("ramfile.ts3",FOPEN_RB);
    fp_ram = ram_fopen("ramfile.ts2", FOPEN_RB);
    count = 0;
    show_stuff = 0;
    show_changing_stuff(__LINE__);
    while(fread(&buf[0],readWriteSize,1,fp_test))
    {
#if 0
    fprintf(stdout,"Line %d testFileSize %d, readWriteSize %d count %d\n",
                __LINE__,testFileSize,readWriteSize,count);
    if((testFileSize == 26) && (readWriteSize == 1) && (count == 125))
    {
        show_stuff = 1;
        count = count;   /* for breakpoints */
    }
#endif
    show_changing_stuff(__LINE__);
      ram_fread(&buf2[0],readWriteSize,1,fp_ram);
    show_changing_stuff(__LINE__);
      if((curLoc1 = ftell(fp_test)) != (curLoc2 = ram_ftell(fp_ram)))
      {
        printf("Ftell error after seeks  Line %d Disk loc 0x%lx, ram loc 0x%lx count %d\n",
        __LINE__,curLoc1, curLoc2,count);
        show_stuff = 1;
        show_static_stuff();
        show_changing_stuff(__LINE__);
        exit(0);
      }
#if 0
        fprintf(stderr,"\r%3.3d 0x%8.8lx",readWriteSize,curLoc1);
#endif
      if(memcmp(&buf[0], &buf2[0],readWriteSize))
      {
        fprintf(stdout,"\nData miss compare after seeks.\n");
        fprintf(stdout,"Line %d disk data %c ram data %c count %d\n",
         __LINE__,buf[0], buf2[0],count);
        curLoc1 = ftell(fp_test);
        curLoc2 = ram_ftell(fp_ram);
        printf("locs 0x%lx 0x%lx\n",curLoc1,curLoc2);

        show_stuff = 1;
        show_static_stuff();
        show_changing_stuff(__LINE__);
        exit(0);
      }
      else
      {
        /* printf("Good buf[0] %c\n",buf[0]); */
    show_changing_stuff(__LINE__);
      }
#if 0
    seekOffset = (long)readWriteSize + ftell(fp_test);
#else
    seekOffset = readWriteSize;
#endif
    fseek(fp_test,seekOffset,SEEK_CUR);
    ram_fseek(fp_ram,seekOffset,SEEK_CUR);
    if((curLoc1 = ftell(fp_test)) != (curLoc2 = ram_ftell(fp_ram)))
    {
        if(!(curLoc1 > biggest))
        {  /* Ignore, ram seek won't go past file size */
        printf("Ftell error after seeks.  Line %d Disk loc 0x%lx, ram loc 0x%lx count %d\n",
        __LINE__,curLoc1, curLoc2,count);
        printf("Biggest = 0x%lx\n",biggest);
        show_stuff = 1;
        show_static_stuff();
        show_changing_stuff(__LINE__);
        exit(0);
        }
    }
    show_changing_stuff(__LINE__);
    count++;
    }
    fclose(fp_test);
    ram_fclose(fp_ram);
    show_stuff = 0;
#endif
        }
    }
    fprintf(stdout,"\nTest complete\n");
}

void
main()
{

    ram_buffer_size = 51;

    Pointers1 = (unsigned char  FAR *)MALLOC(ram_buffer_size);
    Pointers2 = (unsigned char  FAR *)MALLOC(ram_buffer_size);
    if((Pointers1 == NULL) || (Pointers2 == NULL))
    {
            fprintf(stderr,"Unable to allocate Pointers\n;");
            exit(-1);
    }
    sanity = (char FAR *)Pointers1;
    sanity[ram_buffer_size -1] = 0x55;
    sanity = (char FAR *)Pointers2;
    sanity[ram_buffer_size -1] = 0x55;
    nbrRamFarBuffers = 2;
    ramFarBuffers[0]= (char FAR *)Pointers1;
    ramFarBuffers[1]= (char FAR *)Pointers2;
    ramFarBufferSizes[0] = ram_buffer_size - 2;
    ramFarBufferSizes[1] = ram_buffer_size - 2;

#if 1
    testRamFile();
#endif

    sanity_check();
    FREE(Pointers1);
    FREE(Pointers2);
}

#endif /* TEST_RAMFILE */
